@@ -114,6 +114,7 @@ EX bool wrongMode(char flags) {
     if(geometry != gNormal) return true;
     if(disksize) return true;
     }
+  if(ineligible_starting_land && !flags) return true;
 
   if(shmup::on != (flags == rg::shmup || flags == rg::racing)) return true;
   if(racing::on != (flags == rg::racing)) return true;
@@ -125,15 +126,17 @@ EX bool wrongMode(char flags) {
   if(yendor::on) return true;
   if(peace::on) return true;
   if(tactic::on) return true;
-  if(dpgen::in) return true;
 #if CAP_TOUR
   if(tour::on) return true;
 #endif
   eLandStructure dls = lsNiceWalls;
-  if(flags == rg::special_geometry || flags == rg::racing || flags == rg::princess)
+  if(flags == rg::special_geometry || flags == rg::princess)
     dls = lsSingle;
   if(flags == rg::chaos)
     dls = lsChaos;
+  /* in the official racing achievements, the tracks are saved maps anyway */
+  if(flags == rg::racing)
+    dls = land_structure;
 
   if(land_structure != dls) return true;
   if(numplayers() > 1 && !multi::friendly_fire) return true;
@@ -175,7 +178,7 @@ EX void achievement_log(const char* s, char flags) {
 #if CAP_SAVE
   remove_emergency_save();
 
-  FILE *f = fopen(scorefile, "at");
+  FILE *f = fopen(scorefile.c_str(), "at");
   if(!f) return;
   
   int t = (int) (time(NULL) - timerstart);
