@@ -454,7 +454,7 @@ EX void killMutantIvy(cell *c, eMonster who) {
   changes.ccell(c);
   removeIvy(c);
   for(int i=0; i<c->type; i++)
-    if(c->move(i)->mondir == c->c.spin(i) && (isMutantIvy(c->move(i)) || c->move(i)->monst == moFriendlyIvy))
+    if(c->move(i) && c->move(i)->mondir == c->c.spin(i) && (isMutantIvy(c->move(i)) || c->move(i)->monst == moFriendlyIvy))
       kills[c->move(i)->monst]++, killMutantIvy(c->move(i), who);
   if(c->land == laClearing) clearing::imput(c);
   }
@@ -1202,6 +1202,8 @@ EX void killHardcorePlayer(int id, flagtype flags) {
     }
   }
 
+EX bool suicidal;
+
 EX void killThePlayer(eMonster m, int id, flagtype flags) {
   if(markOrb(itOrbShield)) return;
   if(shmup::on) {
@@ -1231,6 +1233,7 @@ EX void killThePlayer(eMonster m, int id, flagtype flags) {
   else {
 //  printf("confused!\n");
     addMessage(XLAT("%The1 is confused!", m));
+    changes.value_set(suicidal, true);
     }
   }
 
@@ -1283,7 +1286,8 @@ EX void stabbingAttack(movei mi, eMonster who, int bonuskill IS(0)) {
     
     bool stabthere = false, away = true;
     if(logical_adjacent(mt, who, c)) stabthere = true, away = false;
-  
+    if(inmirror(c)) c = mirror::reflect(c).at;
+
     if(stabthere && c->wall == waExplosiveBarrel && markOrb(itOrbThorns))
       explodeBarrel(c);
     
