@@ -6,6 +6,10 @@
  */
 
 #include "hyper.h"
+#ifdef FONTCONFIG
+#include <fontconfig/fontconfig.h>
+#endif
+
 namespace hr {
 
 #if HDR
@@ -51,6 +55,7 @@ struct display_data {
   vector<radarpoint> radarpoints;
   vector<radarline> radarlines;
   transmatrix radar_transform;
+  transmatrix radar_transform_post;
 
   ld eyewidth();
   bool stereo_active();
@@ -172,6 +177,8 @@ EX SDL_Renderer *s_renderer, *s_software_renderer;
 #endif
 EX SDL_Texture *s_texture;
 EX SDL_Window *s_window;
+EX SDL_GLContext s_context;
+EX bool s_have_context;
 #endif
 
 EX color_t qpixel_pixel_outside;
@@ -213,12 +220,49 @@ EX void present_screen() {
 
 #if CAP_SDLTTF
 
-EX string fontpath = ISWEB ? "sans-serif" : HYPERFONTPATH "DejaVuSans-Bold.ttf";
+#define DEFAULT_FONT "DejaVuSans-Bold.ttf"
+
+#ifdef FONTCONFIG
+/** if this is non-empty, find the font using fontconfig */
+EX string font_to_find = DEFAULT_FONT;
+#endif
+
+/** actual font path */
+EX string fontpath = ISWEB ? "sans-serif" : string(HYPERFONTPATH) + DEFAULT_FONT;
+
+const string& findfont() {
+  #ifdef FONTCONFIG
+  if(font_to_find == "") return fontpath;
+  FcPattern   *pat;
+  FcResult	result;
+  if (!FcInit()) {
+    return fontpath;
+  }
+  pat = FcNameParse((FcChar8 *)font_to_find.c_str());
+  FcConfigSubstitute(0, pat, FcMatchPattern);
+  FcDefaultSubstitute(pat);
+
+  FcPattern   *match;
+  match = FcFontMatch(0, pat, &result);
+  if (match) {
+    FcChar8 *file;
+    if (FcPatternGetString(match, FC_FILE, 0, &file) == FcResultMatch) {
+      fontpath = (const char *)file;
+    }
+    FcPatternDestroy(match);
+  }
+  FcPatternDestroy(pat);
+  FcFini();
+  font_to_find = "";
+  if(debugflags & DF_INIT) println(hlog, "fontpath is: ", fontpath);
+  #endif
+  return fontpath;
+  }
 
 void loadfont(int siz) {
   fix_font_size(siz);
   if(!font[siz]) {
-    font[siz] = TTF_OpenFont(fontpath.c_str(), siz);
+    font[siz] = TTF_OpenFont(findfont().c_str(), siz);
     // Destination set by ./configure (in the GitHub repository)
     #ifdef FONTDESTDIR
     if (font[siz] == NULL) {
@@ -913,12 +957,19 @@ EX color_t colormix(color_t a, color_t b, color_t c) {
   return a;
   }
 
+/* color difference for 24-bit colors, from 0 to 255*3 */
+EX int color_diff(color_t a, color_t b) {
+  int res = 0;
+  for(int i=0; i<3; i++) res += abs(part(a, i) - part(b, i));
+  return res;
+  }
+
 EX int rhypot(int a, int b) { return (int) sqrt(a*a - b*b); }
 
 EX ld realradius() {
   ld vradius = current_display->radius;
   if(sphere) {
-    if(sphereflipped()) 
+    if(flip_sphere())
       vradius /= sqrt(pconf.alpha*pconf.alpha - 1);
     else
       vradius = 1e12; // use the following
@@ -1015,7 +1066,7 @@ EX void drawCircle(int x, int y, int size, color_t color, color_t fillcolor IS(0
     if(pts > 1500) pts = 1500;
     if(ISMOBILE && pts > 72) pts = 72;
     for(int r=0; r<pts; r++) {
-      float rr = (M_PI * 2 * r) / pts;
+      float rr = (TAU * r) / pts;
       glcoords.push_back(glhr::makevertex(x + size * sin(rr), y + size * pconf.stretch * cos(rr), 0));
       }
     current_display->set_all(0, lband_shift);
@@ -1148,6 +1199,9 @@ EX void close_renderer() {
 EX void close_window() {
   #if CAP_SDL2
   close_renderer();
+  if(s_have_context) {
+    SDL_GL_DeleteContext(s_context), s_have_context = false;
+    }
   if(s_window) SDL_DestroyWindow(s_window), s_window = nullptr;
   #endif
   }
@@ -1277,8 +1331,13 @@ EX void setvideomode() {
 
   auto create_win = [&] {
     #if CAP_SDL2
-    if(s_window && current_window_flags != (flags | sizeflag))
+    if(s_window && current_window_flags != (flags | sizeflag)) {
+      if(s_have_context) {
+        SDL_GL_DeleteContext(s_context), s_have_context = false;
+        glhr::glew = false;
+        }
       SDL_DestroyWindow(s_window), s_window = nullptr;
+      }
     if(s_window)
       SDL_SetWindowSize(s_window, vid.xres, vid.yres);
     else
@@ -1315,6 +1374,7 @@ EX void setvideomode() {
     }
   
   #if CAP_SDL2
+  if(s_renderer) SDL_DestroyRenderer(s_renderer), s_renderer = nullptr;
   s_renderer = SDL_CreateRenderer(s_window, -1, vid.current_vsync ? SDL_RENDERER_PRESENTVSYNC : 0);
   SDL_GetRendererOutputSize(s_renderer, &vid.xres, &vid.yres);
   
@@ -1341,6 +1401,12 @@ EX void setvideomode() {
       glDisable(GL_MULTISAMPLE_ARB);
       }
   
+    #if CAP_SDL2
+    if(s_have_context) SDL_GL_DeleteContext(s_context), s_have_context = false;
+    if(!s_have_context) s_context = SDL_GL_CreateContext(s_window);
+    s_have_context = true; glhr::glew = false;
+    #endif
+
     glViewport(0, 0, vid.xres, vid.yres);
     glhr::init();
     resetGL();
@@ -1472,7 +1538,7 @@ EX void initialize_all() {
 #if CAP_ARCM
   arcm::current.parse();
 #endif
-  if(hybri) geometry = hybrid::underlying;
+  if(mhybrid) geometry = hybrid::underlying;
 
 #if CAP_COMMANDLINE
   arg::read(2);
