@@ -153,7 +153,7 @@ EX int default_levs() {
   if(S3 >= OINF)
     return 1;
   #if MAXMDIM >= 4
-  if(reg3::in_rule()) return 0;
+  if(reg3::in_hrmap_rule_or_subrule()) return 0;
   #endif
   return S3-3;
   }
@@ -184,7 +184,7 @@ map<heptagon*, short> altmap::quotient_relspins;
 auto qclear = addHook(hooks_clearmemory, 200, [] { altmap::quotient_relspins.clear(); });
 
 void hrmap::extend_altmap(heptagon *h, int levs, bool link_cdata) {
-  if(hybri) { PIU ( extend_altmap(h, levs, link_cdata) ); }
+  if(mhybrid) { PIU ( extend_altmap(h, levs, link_cdata) ); }
   if(!h->alt) return;
   preventbarriers(h->c7);
   if(h->c7) forCellEx(c2, h->c7) preventbarriers(c2);
@@ -230,7 +230,7 @@ EX int hrandom_adjacent(cellwalker cw) {
 
 EX heptagon *create_altmap(cell *c, int rad, hstate firststate, int special IS(0)) {
 
-  if(hybri) {
+  if(mhybrid) {
     if(hybrid::under_class() == gcSphere) return NULL;
     c = hybrid::get_where(c).first;
     return PIU ( create_altmap(c, rad, firststate, special) );
@@ -240,7 +240,7 @@ EX heptagon *create_altmap(cell *c, int rad, hstate firststate, int special IS(0
   int gdir = -1;
   for(int i=0; i<c->type; i++) {
     #if MAXMDIM >= 4
-    if(!reg3::in_rule()) {
+    if(!reg3::in_hrmap_rule_or_subrule()) {
     #else
     if(true) {
     #endif
@@ -252,7 +252,7 @@ EX heptagon *create_altmap(cell *c, int rad, hstate firststate, int special IS(0
       }
     }
   #if MAXMDIM >= 4
-  if(reg3::in_rule() && c->master->distance == 0) gdir = 0;
+  if(reg3::in_hrmap_rule_or_subrule() && c->master->distance == 0) gdir = 0;
   #endif
   if(gdir < 0) return NULL;
   
@@ -296,7 +296,7 @@ EX heptagon *create_altmap(cell *c, int rad, hstate firststate, int special IS(0
   if(!currentmap->link_alt(h, alt, firststate, p.last.spin)) {
     return nullptr;
     }
-  if(hybri) hybrid::altmap_heights[alt] = hybrid::get_where(centerover).second;
+  if(mhybrid) hybrid::altmap_heights[alt] = hybrid::get_where(centerover).second;
   alt->alt = alt;
   h->alt = alt;
   alt->cdata = (cdata*) h;
@@ -814,10 +814,10 @@ EX void buildEquidistant(cell *c) {
     ls::nice_walls() ? true :
     false;
   
-  if(c->landparam > 30 && b == laOcean && !generatingEquidistant && !hybri && hrand(10) < 5 && chance) 
+  if(c->landparam > 30 && b == laOcean && !generatingEquidistant && !mhybrid && hrand(10) < 5 && chance)
     buildAnotherEquidistant(c);
 
-  if(c->landparam > HAUNTED_RADIUS+5 && b == laGraveyard && !generatingEquidistant && !hybri && hrand(100) < (PURE?25:5) && items[itBone] >= U10 && chance) 
+  if(c->landparam > HAUNTED_RADIUS+5 && b == laGraveyard && !generatingEquidistant && !mhybrid && hrand(100) < (PURE?25:5) && items[itBone] >= U10 && chance)
     buildAnotherEquidistant(c);
   }
 
@@ -921,7 +921,7 @@ EX void setLandSphere(cell *c) {
     int y = getHemisphere(c, 2);
     elementalXY(c, x, y, (c->type != 6 || GOLDBERG));
     }
-  if(!(euclid && quotient))
+  if(!(euclid && (quotient || disksize)))
   if(specialland == laCrossroads || specialland == laCrossroads2 || specialland == laCrossroads3 || specialland == laTerracotta) {
     int x = getHemisphere(c, 1);
     if(x == 0 && specialland == laTerracotta)
@@ -959,7 +959,7 @@ EX void clear_euland(eLand first) {
   }
 
 bool valid_wall_at(int c) {
-  if(nonisotropic || hybri) return true;
+  if(nonisotropic || mhybrid) return true;
   return short(c) % 3 == 0;
   }
   
@@ -972,7 +972,7 @@ EX eLand switchable(eLand nearland, eLand farland, int c) {
   else if(ls::no_walls()) {
     if((dual::state && nearland == laCrossroads4) || hrand(15) == 0)
       return getNewLand(nearland);
-    if(nearland == laCrossroads4 && (nonisotropic || hybri))
+    if(nearland == laCrossroads4 && (nonisotropic || mhybrid))
       return getNewLand(nearland);
     return nearland;
     }
@@ -1369,9 +1369,13 @@ EX int wallchance(cell *c, bool deepOcean) {
 
 /** \brief should we generate the horocycles in the current geometry? */
 EX bool horo_ok() {  
-  if(INVERSE) return false;
+  if(INVERSE) return false;  
+  if(hat::in()) return false;
   if(currentmap->strict_tree_rules()) return true;
-  return hyperbolic && !bt::in() && !arcm::in() && !kite::in() && !experimental && !hybri && !arb::in() && !quotient;
+  #if MAXMDIM >= 4
+  if(reg3::in_hrmap_h3() && !PURE) return false;
+  #endif
+  return mhyperbolic && !bt::in() && !arcm::in() && !kite::in() && !experimental && !mhybrid && !arb::in() && !quotient;
   }
 
 /** \brief should we either generate the horocycles in the current geometry, or have them exist via eubinary? */
@@ -1486,7 +1490,7 @@ EX bool good_for_wall(cell *c) {
 
 EX bool walls_not_implemented() {
   // if(WDIM == 3 && !PURE) return true;
-  if(sphere || quotient || nonisotropic || (kite::in() && !bt::in()) || experimental) return true;
+  if(sphere || quotient || nonisotropic || aperiodic || experimental) return true;
   return WDIM == 3 && (cgflags & qIDEAL);
   }
 
@@ -1495,7 +1499,7 @@ EX bool old_nice_walls() {
   }
 
 EX bool nice_walls_available() {
-  if(hybri) return PIU(nice_walls_available());
+  if(mhybrid) return PIU(nice_walls_available());
   if(fake::in()) return FPIU(nice_walls_available());
   return WDIM == 2;
   }
@@ -1519,7 +1523,7 @@ EX void build_walls(cell *c, cell *from) {
   
   // buildgreatwalls
   
-  if(hybri) return;  /* Great Walls generated via the underlying geometry */
+  if(mhybrid) return;  /* Great Walls generated via the underlying geometry */
   
   if(walls_not_implemented()) return; // walls not implemented here  
 
@@ -1669,7 +1673,7 @@ EX void build_horocycles(cell *c, cell *from) {
 
   if(ls::any_order() && bearsCamelot(c->land) && can_start_horo(c) && !bt::in() && 
     #if MAXMDIM >= 4
-    !(hyperbolic && WDIM == 3 && !reg3::in_rule()) && 
+    !(hyperbolic && WDIM == 3 && !reg3::in_hrmap_rule_or_subrule()) &&
     #endif
     (quickfind(laCamelot) || peace::on || (hrand(I2000) < (c->land == laCrossroads4 || ls::no_walls() ? 800 : 200) && horo_ok() &&
     items[itEmerald] >= U5))) 
@@ -1859,7 +1863,7 @@ EX void buildCamelot(cell *c) {
 EX int masterAlt(cell *c) {
   if(eubinary) return celldistAlt(c);
   #if MAXMDIM >= 4
-  if(WDIM == 3 && hyperbolic && !reg3::in_rule()) return reg3::altdist(c->master);
+  if(WDIM == 3 && hyperbolic && !reg3::in_hrmap_rule_or_subrule()) return reg3::altdist(c->master);
   #endif
   return c->master->alt->distance;
   }
@@ -1919,7 +1923,7 @@ EX void gen_temple(cell *c) {
       auto d = hybrid::get_where(c);
       if(!PIU(pseudohept(d.first))) c->wall = waColumn;
       }
-    else if(hybri) {
+    else if(mhybrid) {
       auto d = hybrid::get_where(c);
       if(d.first->wall == waColumn || (d.second&1)) c->wall = waColumn;
       }
