@@ -51,7 +51,7 @@ public:
   virtual transmatrix spin_to(cell *c, int d, ld bonus=0);
   virtual transmatrix spin_from(cell *c, int d, ld bonus=0);
   
-  virtual double spacedist(cell *c, int i) { return hdist0(tC0(adj(c, i))); }
+  virtual double spacedist(cell *c, int i);
   
   virtual bool strict_tree_rules() { return false; }
 
@@ -74,6 +74,14 @@ public:
 
   /** \brief the sequence of heptagon movement direction to get from c->master to c->move(i)->master; implemented only for reg3 */
   virtual const vector<int>& get_move_seq(cell *c, int i);
+
+  /** generate a new map that is disconnected from what we already have, disconnected from the map we have so far */
+  virtual cell* gen_extra_origin(int fv) { throw hr_exception("gen_extra_origin not supported on this map"); }
+
+  transmatrix adjmod(cell *c, int i) { return adj(c, gmod(i, c->type)); }
+  transmatrix adjmod(heptagon *h, int i) { return adj(h, gmod(i, h->type)); }
+  transmatrix iadjmod(cell *c, int i) { return iadj(c, gmod(i, c->type)); }
+  transmatrix iadjmod(heptagon *h, int i) { return iadj(h, gmod(i, h->type)); }
   };
 
 /** hrmaps which are based on regular non-Euclidean 2D tilings, possibly quotient  
@@ -95,6 +103,7 @@ struct hrmap_standard : hrmap {
   virtual hyperpoint get_corner(cell *c, int cid, ld cf) override;
   virtual transmatrix master_relative(cell *c, bool get_inverse) override;
   virtual bool link_alt(heptagon *h, heptagon *alt, hstate firststate, int dir) override;
+  virtual void on_dim_change();
   };
 
 void clearfrom(heptagon*);
@@ -115,6 +124,13 @@ struct hrmap_hyperbolic : hrmap_standard {
   };
 #endif
 
+void hrmap_standard::on_dim_change() {
+  for(auto& p: gp::gp_swapped) swapmatrix(gp::gp_adj[p]);
+  gp::gp_swapped.clear();
+  }
+
+double hrmap::spacedist(cell *c, int i) { return hdist(tile_center(), adj(c, i) * tile_center()); }
+
 heptagon *hrmap::create_step(heptagon *h, int direction) {
   throw hr_exception("create_step called unexpectedly");
   return NULL;
@@ -166,7 +182,7 @@ const vector<int>& hrmap::get_move_seq(cell *c, int i) {
 transmatrix hrmap::spin_to(cell *c, int d, ld bonus) {
   ld sa = spin_angle(c, d);
   if(sa != SPIN_NOT_AVAILABLE) { return spin(bonus + sa); }
-  transmatrix T = rspintox(tC0(adj(c, d)));
+  transmatrix T = lrspintox(tC0(adj(c, d)));
   if(WDIM == 3) return T * cspin(2, 0, bonus);
   return T * spin(bonus);
   }
@@ -174,7 +190,7 @@ transmatrix hrmap::spin_to(cell *c, int d, ld bonus) {
 transmatrix hrmap::spin_from(cell *c, int d, ld bonus) {
   ld sa = spin_angle(c, d);
   if(sa != SPIN_NOT_AVAILABLE) { return spin(bonus - sa); }
-  transmatrix T = spintox(tC0(adj(c, d)));
+  transmatrix T = lspintox(tC0(adj(c, d)));
   if(WDIM == 3) return T * cspin(2, 0, bonus);
   return T * spin(bonus);
   }
@@ -184,7 +200,7 @@ transmatrix hrmap::adj(heptagon *h, int i) { return relative_matrix(h->cmove(i),
 vector<cell*>& hrmap::allcells() { 
   static vector<cell*> default_allcells;
   if(disksize) return all_disk_cells;
-  if(closed_manifold && !(cgflags & qHUGE_BOUNDED) && !(hybri && hybrid::csteps == 0)) {
+  if(closed_manifold && !(cgflags & qHUGE_BOUNDED) && !(mhybrid && hybrid::csteps == 0)) {
     celllister cl(gamestart(), 1000000, 1000000, NULL);
     default_allcells = cl.lst;
     return default_allcells;
@@ -229,7 +245,7 @@ EX vector<hrmap*> allmaps;
 
 EX hrmap *newAltMap(heptagon *o) { 
   #if MAXMDIM >= 4
-  if(reg3::in_rule())
+  if(reg3::in_hrmap_rule_or_subrule())
     return reg3::new_alt_map(o);
   #endif
   if(currentmap->strict_tree_rules())
@@ -397,16 +413,161 @@ EX bool is_in_disk(cell *c) {
   return *it == c;
   }
 
+bool sierpinski3(gp::loc g) {
+  int x = g.first;
+  int y = g.second;
+  set<pair<int, int>> visited;
+  while(true) {
+    if(visited.count({x,y})) return false;
+    visited.insert({x,y});
+    // if(x == -1 && y == -2) return false;
+    // if(x == -2 && y == -3) return false;
+    if(x == 0 && y == 0) return true;
+    if((x&1) == 1 && (y&1) == 1) return false;
+    x >>= 1;
+    y >>= 1;
+    // x--; y++;
+    tie(x, y) = make_pair(-x-y, x);
+    }
+  }
+
+bool sierpinski46(gp::loc g) {
+  int x = g.first;
+  int y = g.second;
+  set<pair<int, int>> visited;
+  if(S7 == 6)
+    x += 2785684, y += 289080;
+  else
+    x += 75239892, y += 7913772;
+  while(true) {
+    if(visited.count({x,y})) return false;
+    visited.insert({x,y});
+    if(x == 0 && y == 0) return true;
+    int dx = gmod(x, 3);
+    int dy = gmod(y, 3);
+    if(dx == 1 && dy == 1) return false;
+    if(S7 == 6 && dx == dy) return false;
+    x = (x-dx) / 3;
+    y = (y-dy) / 3;
+    }
+  }
+
+bool menger_sponge(euc::coord c) {
+  c[0] += 528120*9; c[1] += 438924*9; c[2] += 306712*9;
+  set<euc::coord> visited;
+  while(true) {
+    if(visited.count(c)) return false;
+    visited.insert(c);
+    if(c[0] == 0 && c[1] == 0 && c[2] == 0) return true;
+    int ones = 0;
+    for(int i=0; i<3; i++) {
+      int d = gmod(c[i], 3);
+      c[i] = (c[i] - d) / 3;
+      if(d == 1) ones++;
+      }
+    if(ones >= 2) return false;
+    }
+  }
+
+bool sierpinski_tet(euc::coord c) {
+  set<euc::coord> visited;
+  c[0] += 16 * (1+8+64+512);
+  c[1] += 16 * (1+8+64+512);
+  c[1] += 32 * (1+8+64+512);
+  c[2] += 32 * (1+8+64+512);
+  c[0] += 64 * (1+8+64+512);
+  c[1] += 64 * (1+8+64+512);
+  while(true) {
+    if(visited.count(c)) return false;
+    visited.insert(c);
+    if(c[0] == 0 && c[1] == 0 && c[2] == 0) return true;
+    int ones = 0;
+    for(int i=0; i<3; i++) {
+      int d = gmod(c[i], 2);
+      c[i] = (c[i] - d) / 2;
+      if(d == 1) ones++;
+      }
+    if(ones & 1) return false;
+    }
+  }
+
+EX bool is_in_fractal(cell *c) {
+  if(fake::in()) return FPIU(is_in_fractal(c));
+  if(mhybrid) { c = hybrid::get_where(c).first; return PIU(is_in_fractal(c)); }
+  switch(geometry) {
+    case gSierpinski3:
+      return sierpinski3(euc::full_coords2(c));
+    case gSierpinski4:
+    case gSixFlake:
+      return sierpinski46(euc::full_coords2(c));
+    case gMengerSponge:
+      return menger_sponge(euc::get_ispacemap()[c->master]);
+    case gSierpinskiTet: {
+      return sierpinski_tet(euc::get_ispacemap()[c->master]);
+      }
+    default:
+      return true;
+    }
+  }
+
+EX cell *fractal_rep(cell *c) {
+  switch(geometry) {
+    case gSierpinski3: {
+      auto co = euc::full_coords2(c);
+      co.first += 4;
+      co.first &= ~15;
+      co.first -= 4;
+      co.second += 2;
+      co.second &= ~15;
+      co.second -= 2;
+      if(co.first == -4 && co.second == -2) co.first = 0, co.second = 0;
+      return euc::get_at(euc::to_coord(co))->c7;
+      }
+    case gSierpinski4:
+    case gSixFlake: {
+      auto co = euc::full_coords2(c);
+      if(S7 == 6) co.first += 4;
+      co.first -= gmod(co.first, 9);
+      co.second -= gmod(co.second, 9);
+      return euc::get_at(euc::to_coord(co))->c7;
+      }
+    case gSierpinskiTet: {
+      auto co = euc::get_ispacemap()[c->master];
+      co[0] &=~7;
+      co[1] &=~7;
+      co[2] &=~7;
+      return euc::get_at(co)->c7;
+      }
+    case gMengerSponge: {
+      auto co = euc::get_ispacemap()[c->master];
+      for(int i=0; i<3; i++) co[i] = co[i] - gmod(co[i], 9);
+      return euc::get_at(co)->c7;
+      }
+    default:
+      throw hr_exception("unknown fractal");
+    }
+  }
+
 /** create a map in the current geometry */
 EX void initcells() {
   DEBB(DF_INIT, ("initcells"));
-  
+
+  if(embedded_plane) {
+    geom3::swap_direction = -1;
+    for(auto& m: cgi.heptmove) swapmatrix(m);
+    IPF(initcells());
+    geom3::swap_direction = +1;
+    for(auto& m: cgi.heptmove) swapmatrix(m);
+    currentmap->on_dim_change();
+    return;
+    }
+
   hrmap* res = callhandlers((hrmap*)nullptr, hooks_newmap);
   if(res) currentmap = res;
   #if CAP_SOLV
   else if(asonov::in()) currentmap = asonov::new_map();
   #endif
-  else if(nonisotropic || hybri) currentmap = nisot::new_map();
+  else if(nonisotropic || mhybrid) currentmap = nisot::new_map();
   else if(INVERSE) currentmap = gp::new_inverse();
   else if(fake::in()) currentmap = fake::new_map();
   #if CAP_CRYSTAL
@@ -418,6 +579,7 @@ EX void initcells() {
   else if(arcm::in()) currentmap = arcm::new_map();
   #endif
   else if(euc::in()) currentmap = euc::new_map();
+  else if(hat::in()) currentmap = hat::new_map();
   #if CAP_BT
   else if(kite::in()) currentmap = kite::new_map();
   #endif
@@ -454,8 +616,8 @@ EX void clearcell(cell *c) {
     c->move(t)->move(c->c.spin(t)) = NULL;
     }
   DEBB(DF_MEMORY, (format("DEL %p\n", hr::voidp(c))));
-  destroy_cell(c);
   gp::delete_mapped(c);
+  destroy_cell(c);
   }
 
 EX heptagon deletion_marker;
@@ -583,11 +745,12 @@ EX int compdist(int dx[]) {
 
 EX int celldist(cell *c) {
   if(experimental) return 0;
-  if(hybri) 
+  if(mhybrid)
     return hybrid::celldistance(c, currentmap->gamestart());
   if(nil && !quotient) return DISTANCE_UNKNOWN;
+  if(hat::in()) return clueless_celldistance(currentmap->gamestart(), c);
   if(euc::in()) return celldistance(currentmap->gamestart(), c);
-  if(sphere || bt::in() || WDIM == 3 || cryst || sn::in() || kite::in() || closed_manifold) return celldistance(currentmap->gamestart(), c);
+  if(sphere || bt::in() || WDIM == 3 || cryst || sn::in() || aperiodic || closed_manifold) return celldistance(currentmap->gamestart(), c);
   #if CAP_IRR
   if(IRREGULAR) return irr::celldist(c, false);
   #endif
@@ -614,7 +777,7 @@ static const int ALTDIST_ERROR = 90000;
 
 EX int celldistAlt(cell *c) {
   if(experimental) return 0;
-  if(hybri) { 
+  if(mhybrid) {
     if(in_s2xe()) return hybrid::get_where(c).second;
     auto w = hybrid::get_where(c); 
     int d = c->master->alt && c->master->alt->alt ? hybrid::altmap_heights[c->master->alt->alt] : 0;
@@ -635,7 +798,7 @@ EX int celldistAlt(cell *c) {
     }
   #if MAXMDIM >= 4
   if(euc::in()) return euc::dist_alt(c);
-  if(hyperbolic && WDIM == 3 && !reg3::in_rule())
+  if(hyperbolic && WDIM == 3 && !reg3::in_hrmap_rule_or_subrule())
     return reg3::altdist(c->master);
   #endif
   if(!c->master->alt) return 0;
@@ -673,8 +836,8 @@ EX int updir(heptagon *h) {
   if(bt::in()) return bt::updir();
   #endif
   #if MAXMDIM >= 4
-  if(WDIM == 3 && reg3::in_rule()) {
-    for(int i=0; i<S7; i++) if(h->move(i) && h->move(i)->distance < h->distance) 
+  if(WDIM == 3 && reg3::in_hrmap_rule_or_subrule()) {
+    for(int i=0; i<h->type; i++) if(h->move(i) && h->move(i)->distance < h->distance) 
       return i;
     return -1;
     }
@@ -687,7 +850,7 @@ EX int updir(heptagon *h) {
 EX int updir_alt(heptagon *h) {
   if(euclid || !h->alt) return -1;
   #if MAXMDIM >= 4
-  if(WDIM == 3 && reg3::in_rule()) {
+  if(WDIM == 3 && reg3::in_hrmap_rule_or_subrule()) {
     for(int i=0; i<S7; i++) if(h->move(i) && h->move(i)->alt && h->move(i)->alt->distance < h->alt->distance) 
       return i;
     return -1;
@@ -828,7 +991,8 @@ EX bool randpatternMajority(cell *c, int ival, int iterations) {
 cdata orig_cdata;
 
 EX bool geometry_supports_cdata() {
-  if(hybri) return PIU(geometry_supports_cdata());
+  if(mhybrid) return PIU(geometry_supports_cdata());
+  if(embedded_plane) return IPF( geometry_supports_cdata() );
   return among(geometry, gEuclid, gEuclidSquare, gNormal, gOctagon, g45, g46, g47, gBinaryTiling) || (arcm::in() && !sphere) || (currentmap && currentmap->strict_tree_rules());
   }
 
@@ -922,7 +1086,8 @@ cdata *getHeptagonCdata_legacy(heptagon *h) {
 
 
 cdata *getHeptagonCdata(heptagon *h) {
-  if(hybri) return PIU ( getHeptagonCdata(h) );
+  if(mhybrid) return PIU ( getHeptagonCdata(h) );
+  if(embedded_plane) return IPF( getHeptagonCdata(h) );
   if(geometry == gNormal && BITRUNCATED) return getHeptagonCdata_legacy(h);
   if(h->cdata) return h->cdata;
 
@@ -1068,7 +1233,7 @@ int ld_to_int(ld x) {
 #if CAP_ARCM
 EX gp::loc pseudocoords(cell *c) {
   transmatrix T = arcm::archimedean_gmatrix[c->master].second;
-  return {ld_to_int(T[0][LDIM]), ld_to_int((spin(60*degree) * T)[0][LDIM])};
+  return {ld_to_int(T[0][LDIM]), ld_to_int((spin(60._deg) * T)[0][LDIM])};
   }
 
 EX cdata *arcmCdata(cell *c) {
@@ -1087,8 +1252,9 @@ EX cdata *arcmCdata(cell *c) {
 
 EX int getCdata(cell *c, int j) {
   if(fake::in()) return FPIU(getCdata(c, j));
+  if(embedded_plane) return IPF(getCdata(c, j));
   if(experimental) return 0;
-  if(hybri) { c = hybrid::get_where(c).first; return PIU(getBits(c)); }
+  if(mhybrid) { c = hybrid::get_where(c).first; return PIU(getBits(c)); }
   else if(INVERSE) {
     cell *c1 = gp::get_mapped(c);
     return UIU(getCdata(c1, j));
@@ -1113,8 +1279,9 @@ EX int getCdata(cell *c, int j) {
 
 EX int getBits(cell *c) {
   if(fake::in()) return FPIU(getBits(c));
+  if(embedded_plane) return IPF(getBits(c));
   if(experimental) return 0;
-  if(hybri) { c = hybrid::get_where(c).first; return PIU(getBits(c)); }
+  if(mhybrid) { c = hybrid::get_where(c).first; return PIU(getBits(c)); }
   else if(INVERSE) {
     cell *c1 = gp::get_mapped(c);
     return UIU(getBits(c1));
@@ -1260,9 +1427,11 @@ EX int clueless_celldistance(cell *c1, cell *c2) {
 
 EX int celldistance(cell *c1, cell *c2) {
 
+  if(embedded_plane) return IPF(celldistance(c1, c2));
+
   if(fake::in()) return FPIU(celldistance(c1, c2));
 
-  if(hybri) return hybrid::celldistance(c1, c2);
+  if(mhybrid) return hybrid::celldistance(c1, c2);
   
   #if CAP_FIELD
   if(geometry == gFieldQuotient && (PURE || BITRUNCATED)) {
@@ -1281,7 +1450,7 @@ EX int celldistance(cell *c1, cell *c2) {
     return euc::cyldist(euc2_coordinates(c1), euc2_coordinates(c2));
     }
 
-  if(arcm::in() || quotient || sn::in() || (kite::in() && euclid) || experimental || sl2 || nil || arb::in()) 
+  if(arcm::in() || quotient || sn::in() || (aperiodic && euclid) || experimental || sl2 || nil || arb::in()) 
     return clueless_celldistance(c1, c2);
    
    if(S3 >= OINF) return inforder::celldistance(c1, c2);
@@ -1305,6 +1474,8 @@ EX int celldistance(cell *c1, cell *c2) {
     /* TODO */
     }
 
+  if(euclid) return clueless_celldistance(c1, c2);
+
   return hyperbolic_celldistance(c1, c2);
   }
 
@@ -1362,7 +1533,7 @@ EX vector<cell*> build_shortest_path(cell *c1, cell *c2) {
   }
 
 EX void clearCellMemory() {
-  #if MAXMDIM >= 4 && CAP_RAY
+  #if CAP_PORTALS
   if(intra::in) {
     intra::erase_all_maps();
     return;
@@ -1412,6 +1583,7 @@ EX array<map<cell*, vector<adj_data>>, 2> adj_memo;
 
 EX bool geometry_has_alt_mine_rule() {
   if(S3 >= OINF) return false;
+  if(aperiodic) return true;
   if(WDIM == 2) return valence() > 3;
   if(WDIM == 3) return !among(geometry, gHoroHex, gCell5, gBitrunc3, gCell8, gECell8, gCell120, gECell120);
   return true;
@@ -1434,7 +1606,7 @@ EX vector<adj_data> adj_minefield_cells_full(cell *c) {
     cellwalker cw1 = cw;
     do {
       res.emplace_back(adj_data{cw.at, cw.mirrored, T});
-      T = T * currentmap->adj(c, cw.spin);
+      T = T * currentmap->adj(cw.at, cw.spin);
       cw += wstep;
       cw++;
       if(cw.cpeek() == c) cw++;
@@ -1472,11 +1644,16 @@ EX vector<cell*> adj_minefield_cells(cell *c) {
   vector<cell*> res;
   auto ori = adj_minefield_cells_full(c);
   for(auto p: ori) res.push_back(p.c);
+  if(hat::in()) {
+    // reduce repetitions
+    sort(res.begin(), res.end());
+    res.erase(std::unique(res.begin(), res.end()), res.end());
+    }
   return res;
   }
 
 EX vector<int> reverse_directions(cell *c, int dir) {
-  if(PURE && !(kite::in() && WDIM == 2)) return reverse_directions(c->master, dir);
+  if(PURE && !(aperiodic && WDIM == 2)) return reverse_directions(c->master, dir);
   int d = c->degree();
   if(d & 1)
     return { gmod(dir + c->type/2, c->type), gmod(dir + (c->type+1)/2, c->type) };
@@ -1528,7 +1705,7 @@ EX vector<int> reverse_directions(heptagon *c, int dir) {
   }
 
 EX bool standard_tiling() {
-  return !arcm::in() && !kite::in() && !bt::in() && !arb::in() && !nonisotropic && !hybri;
+  return !arcm::in() && !aperiodic && !bt::in() && !arb::in() && (WDIM == 2 || !nonisotropic) && !mhybrid;
   }
 
 EX int valence() {
