@@ -750,9 +750,12 @@ enum eGeometry {
   gInfOrderMixed, gSpace436, gFake,
   gSpace345, gSpace353, gSpace354, gSpace355,
   gHalfBring,
+  gAperiodicHat,
+  gSierpinski3, gSierpinski4, gSixFlake, gMengerSponge, gSierpinskiTet,
+  gAperiodicSpectre,
   gGUARD};
 
-enum eGeometryClass { gcHyperbolic, gcEuclid, gcSphere, gcSolNIH, gcNil, gcProduct, gcSL2 };
+enum eGeometryClass { gcHyperbolic, gcEuclid, gcSphere, gcSol, gcNIH, gcSolN, gcNil, gcProduct, gcSL2 };
 
 enum class eVariation { bitruncated, pure, goldberg, irregular, dual, untruncated, warped, unrectified, subcubes, coxeter, dual_subcubes, bch, bch_oct };
 
@@ -826,8 +829,14 @@ static const flagtype qSTRETCHABLE     = Flag(27);
 
 static const flagtype qCAT             = Flag(28);
 
+static const flagtype qAPERIODIC       = Flag(29);
+static const flagtype qHAT             = Flag(30);
+
+static const flagtype qFRACTAL         = Flag(31);
+
 // note: dnext assumes that x&7 equals 7
 static const int SEE_ALL = 50;
+// note: check_football_colorability in arbitrile.cpp assumes OINF is divisible by 3
 static const int OINF = 123;
 
 extern eGeometry geometry;
@@ -859,9 +868,14 @@ EX geometryinfo1 giEuclid3 = { gcEuclid,     3, 3, 4, {1,1, 1,0 } };
 EX geometryinfo1 giHyperb3 = { gcHyperbolic, 3, 3, 4, {1,1, 1,-1} };
 EX geometryinfo1 giSphere3 = { gcSphere,     3, 3, 4, {1,1, 1,+1} };
 
-EX geometryinfo1 giSolNIH  = { gcSolNIH,     3, 3, 4, {1,1, 1,0 } };
+EX geometryinfo1 giSol     = { gcSol,        3, 3, 4, {1,1, 1,0 } };
+EX geometryinfo1 giNIH     = { gcNIH,        3, 3, 4, {1,1, 1,0 } };
+EX geometryinfo1 giSolN    = { gcSolN,       3, 3, 4, {1,1, 1,0 } };
+
 EX geometryinfo1 giNil     = { gcNil,        3, 3, 4, {1,1, 1,0 } };
-EX geometryinfo1 giProduct = { gcSL2,        3, 3, 4, {1,1, 1,0 } /* will be filled in product::configure() */ };
+EX geometryinfo1 giProductH= { gcProduct,    3, 3, 3, {1,1,-1,0 } };
+EX geometryinfo1 giProductS= { gcProduct,    3, 3, 3, {1,1, 1,0 } };
+EX geometryinfo1 giProduct = { gcProduct,    3, 3, 3, {1,1, 1,0 } /* will be filled in product::configure() */ };
 EX geometryinfo1 giSL2     = { gcSL2,        3, 3, 4, {1,1,-1,-1} };
 
 EX modecode_t no_code = 0x1;
@@ -918,20 +932,20 @@ EX vector<geometryinfo> ginf = {
   {"{4,3,5}","field",   "{4,3,5} field quotient space",               "f435",     6, 5, qsSMALLBF | qDEPRECATED, giHyperb3, {{SEE_ALL, SEE_ALL}}, eVariation::pure},
   {"{5,3,4}","field",   "{5,3,4} field quotient space",               "f435",    12, 4, qsSMALLBF | qDEPRECATED, giHyperb3, {{SEE_ALL, SEE_ALL}}, eVariation::pure},
   {"binary4","none",    "standard binary tiling",                     "binary4",  5, 3, qBINARY,   giHyperb2, {{7, 5}}, eVariation::pure},
-  {"sol",    "none",    "Solv geometry",                              "sol",      8, 3, qBINARY|qSOL,   giSolNIH, {{7, 5}}, eVariation::pure},
-  {"kd2",    "none",    "kite-and-dart",                              "kd2",      4, 3, qKITE,     giEuclid2, {{7, 7}}, eVariation::pure},
+  {"sol",    "none",    "Solv geometry",                              "sol",      8, 3, qBINARY|qSOL,   giSol, {{7, 5}}, eVariation::pure},
+  {"kd2",    "none",    "kite-and-dart",                              "kd2",      4, 3, qKITE | qAPERIODIC,     giEuclid2, {{7, 7}}, eVariation::pure},
   {"kd3",    "none",    "kite-and-dart on horospheres",               "kd3",     12, 3, qsBP,      giHyperb3, {{7, 3}}, eVariation::pure},
   {"nil",    "none",    "Nil geometry",                               "nil",      6, 3, qOPTQ,     giNil,     {{7, 5}}, eVariation::pure},
   {"product","none",    "product space",                              "product",  7, 3, qHYBRID,   giProduct, {{7, 3}}, eVariation::pure},
   {"twisted","none",    "rotation space",                             "twisted",  7, 3, qHYBRID,   giSL2,     {{6, 4}}, eVariation::pure},
   {"ternary","none",    "standard ternary tiling",                    "ternary",  6, 3, qBINARY,   giHyperb2, {{6, 4}}, eVariation::pure},
-  {"3x2",    "none",    "stretched hyperbolic",                       "3:2",     11, 3, qBINARY|qNIH,   giSolNIH,  {{6, 3}}, eVariation::pure},
-  {"3x1/2",  "none",    "stretched Solv",                             "3:1/2",    9, 3, (qBINARY|qSOL|qNIH),   giSolNIH,  {{7, 3}}, eVariation::pure},
+  {"3x2",    "none",    "stretched hyperbolic",                       "3:2",     11, 3, qBINARY|qNIH,   giNIH,  {{6, 3}}, eVariation::pure},
+  {"3x1/2",  "none",    "stretched Solv",                             "3:1/2",    9, 3, (qBINARY|qSOL|qNIH),   giSolN,  {{7, 3}}, eVariation::pure},
   {"{3,oo}", "none",    "{3,∞} (infinite triangles)",                 "oox3",     3, OINF, qIDEAL,  giHyperb2, {{7, 7}}, eVariation::pure},
   {"{3,3,6}","none",    "{3,3,6} hyperbolic honeycomb",               "336",      4, 6, qIDEAL,    giHyperb3, {{7, 2}}, eVariation::pure},
   {"{3,4,4}","none",    "{3,4,4} hyperbolic honeycomb",               "344",      8, 4, qIDEAL,    giHyperb3, {{7, 2}}, eVariation::pure},
   {"{3,4,4}","Crystal", "4D crystal in H3",                           "Cryst3" ,  8, 4, qIDEAL | qANYQ | qCRYSTAL, giHyperb3, {{7, 3}}, eVariation::pure},
-  {"cat",    "cat",     "Arnold's cat mapping torus",                 "cat",     12, 3, qBINARY | qSOL | qsBQ | qOPTQ | qCAT, giSolNIH, {{6, 4}}, eVariation::pure},
+  {"cat",    "cat",     "Arnold's cat mapping torus",                 "cat",     12, 3, qBINARY | qSOL | qsBQ | qOPTQ | qCAT, giSol, {{6, 4}}, eVariation::pure},
   {"file",   "none",    "load from file",                             "file",     7, 3, 0,  giEuclid2, {{7, 5}}, eVariation::pure},
   {"{4,oo}", "none",    "{4,∞} (infinite squares)",                   "oox4",     4, OINF, qIDEAL,  giHyperb2, {{5, 5}}, eVariation::pure},
   {"{5,3,4}","Crystal", "6D crystal in H3",                           "Cryst6" , 12, 4, qANYQ | qCRYSTAL, giHyperb3, {{7, 3}}, eVariation::pure},
@@ -948,13 +962,20 @@ EX vector<geometryinfo> ginf = {
   {"{3,5,4}","none",    "{3,5,4} hyperbolic honeycomb",               "354",     20, 5, qIDEAL | qULTRA,    giHyperb3, {{7, 2}}, eVariation::pure},
   {"{3,5,5}","none",    "{3,5,5} hyperbolic honeycomb",               "355",     20, 5, qIDEAL | qULTRA,    giHyperb3, {{7, 2}}, eVariation::pure},
   {"{5,4}", "pBring",   "projective Bring's Surface",                 "pBring",   5, 4, qsSMALLN,   giHyperb2, {{6, 4}}, eVariation::bitruncated},
+  {"hat",    "none",    "aperiodic hat",                              "hat",     14, 3, qAPERIODIC | qHAT,  giEuclid2, {{7, 7}}, eVariation::pure},
+  {"triangle","none",   "Sierpiński triangle",                        "S3",       6, 3, qFRACTAL,  giEuclid2, {{10, 10}}, eVariation::pure},
+  {"carpet", "none",    "Sierpiński carpet",                          "S4",       4, 4, qFRACTAL,  giEuclid2, {{10, 10}}, eVariation::pure},
+  {"6-flake","none",    "6-flake fractal",                            "S6",       6, 3, qFRACTAL,  giEuclid2, {{10, 10}}, eVariation::pure},
+  {"{4,3,4}","none",    "Menger sponge",                              "S8",       6, 4, qFRACTAL,  giEuclid3, {{10, 10}}, eVariation::pure},
+  {"rh{4,3,4}","none",  "Sierpiński tetrahedron",                     "S4b",     12, 3, qFRACTAL,  giEuclid3, {{10, 10}}, eVariation::pure},
+  {"spectre","none",    "aperiodic spectre",                          "spectre", 14, 3, qAPERIODIC | qHAT,  giEuclid2, {{7, 7}}, eVariation::pure},
   };
   // bits: 9, 10, 15, 16, (reserved for later) 17, 18
 
 #if HDR
 namespace mf {
   static const flagtype azimuthal = 1;
-  static const flagtype cylindrical = 2;
+  static const flagtype cylindrical = 2; /* usually you want 'band' */
   static const flagtype equiarea = 4;
   static const flagtype equidistant = 8;
   static const flagtype conformal = 16;
@@ -962,15 +983,22 @@ namespace mf {
   static const flagtype space = 64;
   static const flagtype hyper_only = 128;
   static const flagtype hyper_or_torus = 256;
-  static const flagtype pseudocylindrical = 512; /* includes cylindrical */
+  static const flagtype pseudocylindrical = 512; /* includes cylindrical; usually you want 'band' or 'pseudoband' */
   static const flagtype equivolume = 1024;
   static const flagtype twopoint = 2048;
   static const flagtype uses_bandshift = 4096;
   static const flagtype broken = 8192; /* in spherical case, these are broken along the meridian 180 deg */
   static const flagtype technical = 16384; /* don't display in the list */
+  static const flagtype product_special = (1<<15);
+  static const flagtype axial = (1<<16);
+  static const flagtype perspective = (1<<17);
+  static const flagtype orientation = (1<<18);
+  static const flagtype transition = (1<<19);
+  static const flagtype werner = (1<<20);
+  static const flagtype horocyclic = (1<<21);
   
-  static const flagtype band = (cylindrical | pseudocylindrical | uses_bandshift);
-  static const flagtype pseudoband = (pseudocylindrical | uses_bandshift);
+  static const flagtype band = (cylindrical | pseudocylindrical | uses_bandshift | orientation);
+  static const flagtype pseudoband = (pseudocylindrical | uses_bandshift | orientation);
   }
   
 struct modelinfo {
@@ -979,13 +1007,6 @@ struct modelinfo {
   const char *name_spherical;  
   
   flagtype flags;
-
-  int is_azimuthal;
-  int is_band;
-  int is_equiarea;
-  int is_equidistant;
-  int is_conformal;
-  const char* name;
   };
 
 enum eModel : int {
@@ -1006,9 +1027,11 @@ enum eModel : int {
   mdHorocyclic, mdQuadrant, mdAxial, mdAntiAxial,
   // 32..38
   mdWerner, mdAitoff, mdHammer, mdLoximuthal, mdMiller, mdGallStereographic, mdWinkelTripel,
-  // 39..
-  mdPoorMan, mdPanini, mdRetroCraig, mdRetroLittrow, mdRetroHammer, mdThreePoint, mdLiePerspective, mdLieOrthogonal,
-  // 47..
+  // 39..48
+  mdPoorMan, mdPanini, mdRetroCraig, mdRetroLittrow, mdRetroHammer, mdThreePoint, mdLiePerspective, mdLieOrthogonal, mdRelPerspective, mdRelOrthogonal,
+  // 49..50
+  mdHorocyclicEqa, mdConformalSquare,
+  // 51..
   mdGUARD, mdPixel, mdHyperboloidFlat, mdPolynomial, mdManual
   };
 #endif
@@ -1017,66 +1040,68 @@ enum eModel : int {
 // (other bits are used for other information)
 
 #define X3(x) x, x, x
-#define DEFAULTS 0, 0, 0, 0, 0, nullptr
 
 /** list of available models (i.e., projections) */
 EX vector<modelinfo> mdinf = {
-  {"disk/Gans", "general perspective", "general perspective", mf::azimuthal | mf::conformal, DEFAULTS},
-  {"half-plane", "inversion", "stereographic projection [VR]", mf::conformal, DEFAULTS},
-  {"band", "band", "Mercator", mf::band | mf::conformal, DEFAULTS},
-  {X3("polygonal"), mf::conformal, DEFAULTS},
-  {X3("formula"), 0, DEFAULTS},
-  {X3("azimuthal equidistant"), mf::azimuthal | mf::equidistant | mf::euc_boring, DEFAULTS},
-  {X3("azimuthal equi-area"), mf::azimuthal | mf::equiarea | mf::euc_boring, DEFAULTS},
-  {X3("ball model"), mf::conformal | mf::azimuthal | mf::space, DEFAULTS},
-  {"Minkowski hyperboloid", "plane", "sphere", mf::conformal | mf::space | mf::euc_boring, DEFAULTS},
-  {"hemisphere", "sphere", "Minkowski hyperboloid", mf::conformal | mf::space, DEFAULTS},
-  {X3("band equidistant"), mf::band | mf::equidistant | mf::euc_boring, DEFAULTS},
-  {X3("band equi-area"), mf::band | mf::equiarea | mf::euc_boring, DEFAULTS},
-  {X3("sinusoidal"), mf::pseudoband | mf::equiarea | mf::euc_boring, DEFAULTS},
-  {X3("two-point equidistant"), mf::equidistant | mf::euc_boring | mf::twopoint, DEFAULTS},
-  {X3("fisheye"), 0, DEFAULTS},
-  {X3("Joukowsky transform"), mf::hyper_only | mf::conformal, DEFAULTS},
-  {X3("Joukowsky+inversion"), mf::hyper_only | mf::conformal, DEFAULTS},
-  {X3("rotated hyperboles"), mf::hyper_only, DEFAULTS},
-  {X3("spiral/ring"), mf::hyper_or_torus | mf::uses_bandshift, DEFAULTS},
-  {X3("native perspective"), 0, DEFAULTS},
-  {X3("azimuthal equi-volume"), mf::azimuthal | mf::equivolume | mf::euc_boring, DEFAULTS},
-  {X3("central inversion"), mf::azimuthal | mf::conformal, DEFAULTS},
-  {X3("two-point azimuthal"), mf::euc_boring | mf::twopoint, DEFAULTS},
-  {X3("two-point hybrid"), mf::euc_boring | mf::twopoint, DEFAULTS},
-  {X3("geodesic"), 0, DEFAULTS},
-  {X3("Mollweide"), mf::euc_boring | mf::pseudoband | mf::equiarea, DEFAULTS},
-  {X3("central cylindrical"), mf::euc_boring | mf::band, DEFAULTS},
-  {X3("Collignon"), mf::pseudoband | mf::equiarea, DEFAULTS},
-  {X3("horocyclic coordinates"), mf::euc_boring, DEFAULTS},
-  {X3("quadrant coordinates"), mf::euc_boring, DEFAULTS},
-  {X3("axial coordinates"), mf::euc_boring, DEFAULTS},
-  {X3("anti-axial coordinates"), mf::euc_boring, DEFAULTS},
-  {X3("Werner projection"), mf::euc_boring | mf::broken, DEFAULTS}, // keep distances from pole, and distances along parallels
-  {X3("Aitoff projection"), mf::euc_boring | mf::broken, DEFAULTS}, // halve longitudes, do azequid, double x
-  {X3("Hammer projection"), mf::euc_boring | mf::broken, DEFAULTS}, // halve longitudes, do azequia, double x
-  {X3("loximuthal projection"), mf::euc_boring | mf::broken, DEFAULTS}, // map loxodromes azimuthally and equidistantly
-  {X3("Miller projection"), mf::euc_boring | mf::band, DEFAULTS}, // scale latitude 4/5 -> Mercator -> 5/4
-  {X3("Gall stereographic"), mf::euc_boring | mf::band, DEFAULTS}, // like central cylindrical but stereographic
-  {X3("Winkel tripel"), mf::euc_boring | mf::broken, DEFAULTS}, // mean of equirec and Aitoff
-  {X3("Poor man's square"), mf::euc_boring, DEFAULTS}, // 
-  {X3("Panini projection"), mf::euc_boring, DEFAULTS}, // 
-  {X3("Craig retroazimuthal"), mf::euc_boring | mf::broken, DEFAULTS}, // retroazimuthal cylindrical
-  {X3("Littrow retroazimuthal"), mf::euc_boring | mf::broken, DEFAULTS}, // retroazimuthal conformal
-  {X3("Hammer retroazimuthal"), mf::euc_boring, DEFAULTS}, // retroazimuthal equidistant
-  {X3("three-point equidistant"), mf::euc_boring, DEFAULTS},
-  {X3("Lie perspective"), mf::euc_boring, DEFAULTS},
-  {X3("Lie orthogonal"), mf::euc_boring, DEFAULTS},
-  {X3("guard"), mf::technical, DEFAULTS},
-  {X3("pixel"), mf::technical, DEFAULTS},
-  {X3("hypflat"), mf::technical, DEFAULTS},
-  {X3("polynomial"), mf::technical | mf::conformal, DEFAULTS},
-  {X3("manual"), mf::technical, DEFAULTS},
+  {"disk/Gans", "general perspective", "general perspective", mf::azimuthal | mf::conformal},
+  {"half-plane", "inversion", "stereographic projection [VR]", mf::conformal | mf::orientation | mf::horocyclic},
+  {"band", "band", "Mercator", mf::band | mf::conformal | mf::transition},
+  {X3("polygonal"), mf::conformal | mf::orientation},
+  {X3("formula"), 0},
+  {X3("azimuthal equidistant"), mf::azimuthal | mf::equidistant | mf::euc_boring | mf::product_special},
+  {X3("azimuthal equi-area"), mf::azimuthal | mf::equiarea | mf::euc_boring},
+  {X3("ball model"), mf::conformal | mf::azimuthal | mf::space},
+  {"Minkowski hyperboloid", "plane", "sphere", mf::conformal | mf::space | mf::euc_boring},
+  {"hemisphere", "sphere", "Minkowski hyperboloid", mf::conformal | mf::space},
+  {X3("band equidistant"), mf::band | mf::equidistant | mf::euc_boring},
+  {X3("band equi-area"), mf::band | mf::equiarea | mf::euc_boring},
+  {X3("sinusoidal"), mf::pseudoband | mf::equiarea | mf::euc_boring},
+  {X3("two-point equidistant"), mf::equidistant | mf::euc_boring | mf::twopoint | mf::orientation},
+  {X3("fisheye"), 0},
+  {X3("Joukowsky transform"), mf::hyper_only | mf::conformal | mf::transition | mf::orientation},
+  {X3("Joukowsky+inversion"), mf::hyper_only | mf::conformal | mf::transition | mf::orientation},
+  {X3("rotated hyperboles"), mf::hyper_only | mf::orientation},
+  {X3("spiral/ring"), mf::hyper_or_torus | mf::uses_bandshift | mf::orientation},
+  {X3("native perspective"), mf::perspective | mf::product_special},
+  {X3("azimuthal equi-volume"), mf::azimuthal | mf::equivolume | mf::euc_boring},
+  {X3("central inversion"), mf::azimuthal | mf::conformal},
+  {X3("two-point azimuthal"), mf::euc_boring | mf::azimuthal | mf::twopoint | mf::orientation},
+  {X3("two-point hybrid"), mf::euc_boring | mf::azimuthal | mf::equidistant | mf::twopoint | mf::orientation},
+  {X3("geodesic"), mf::perspective | mf::product_special},
+  {X3("Mollweide"), mf::euc_boring | mf::pseudoband | mf::equiarea},
+  {X3("central cylindrical"), mf::euc_boring | mf::band},
+  {X3("Collignon"), mf::pseudoband | mf::equiarea},
+  {X3("horocyclic coordinates"), mf::euc_boring | mf::orientation | mf::horocyclic},
+  {X3("quadrant coordinates"), mf::euc_boring | mf::orientation},
+  {X3("axial coordinates"), mf::euc_boring | mf::transition | mf::orientation},
+  {X3("anti-axial coordinates"), mf::euc_boring | mf::orientation},
+  {X3("Werner projection"), mf::euc_boring | mf::broken | mf::werner | mf::orientation}, // keep distances from pole, and distances along parallels
+  {X3("Aitoff projection"), mf::euc_boring | mf::broken | mf::orientation}, // halve longitudes, do azequid, double x
+  {X3("Hammer projection"), mf::euc_boring | mf::broken | mf::orientation}, // halve longitudes, do azequia, double x
+  {X3("loximuthal projection"), mf::euc_boring | mf::broken | mf::orientation}, // map loxodromes azimuthally and equidistantly
+  {X3("Miller projection"), mf::euc_boring | mf::band}, // scale latitude 4/5 -> Mercator -> 5/4
+  {X3("Gall stereographic"), mf::euc_boring | mf::band}, // like central cylindrical but stereographic
+  {X3("Winkel tripel"), mf::euc_boring | mf::broken | mf::orientation}, // mean of equirec and Aitoff
+  {X3("Poor man's square"), mf::euc_boring | mf::orientation}, // https://archive.bridgesmathart.org/2018/bridges2018-59.html
+  {X3("Panini projection"), mf::euc_boring | mf::orientation},
+  {X3("Craig retroazimuthal"), mf::euc_boring | mf::broken | mf::pseudoband}, // retroazimuthal cylindrical
+  {X3("Littrow retroazimuthal"), mf::euc_boring | mf::broken | mf::pseudoband}, // retroazimuthal conformal
+  {X3("Hammer retroazimuthal"), mf::euc_boring | mf::pseudoband}, // retroazimuthal equidistant
+  {X3("three-point equidistant"), mf::euc_boring | mf::equidistant | mf::orientation | mf::product_special | mf::twopoint},
+  {X3("Lie perspective"), mf::euc_boring | mf::perspective},
+  {X3("Lie orthogonal"), mf::euc_boring},
+  {X3("relativistic perspective"), mf::euc_boring | mf::perspective},
+  {X3("relativistic orthogonal"), mf::euc_boring},
+  {X3("horocyclic equal-area"), mf::euc_boring | mf::equiarea | mf::orientation | mf::horocyclic},
+  {X3("conformal square"), mf::orientation | mf::broken | mf::transition},
+  {X3("guard"), mf::technical},
+  {X3("pixel"), mf::technical},
+  {X3("hypflat"), mf::technical},
+  {X3("polynomial"), mf::technical | mf::conformal | mf::orientation},
+  {X3("manual"), mf::technical},
   };
 
 #undef X3
-#undef DEFAULTS
 
 #if HDR
 static inline bool orbProtection(eItem it) { return false; } // not implemented
