@@ -148,7 +148,7 @@ EX namespace brownian {
   
   EX void apply_futures(cell *c) {
     if(futures.count(c)) {
-      auto m = move(futures[c]);
+      auto m = std::move(futures[c]);
       futures.erase(c);
       for(auto p: m)
         recurse(p.first, p.second);
@@ -526,7 +526,7 @@ EX void count_status() {
   for(cell *c: currentmap->allcells()) if(among(c->wall, waMineMine, waMineUnknown) && mine::marked_mine(c)) kills[moTameBomberbird]++;
   if(last && !kills[moBomberbird]) {
     mine::victory_time = getgametime();
-    pushScreen(showMission);
+    showMissionScreen();
     }
   }
 
@@ -1239,7 +1239,7 @@ EX namespace dice {
         }
       }
     
-    shiftmatrix V1 = V * ddspin(c, dir) * spin(M_PI);
+    shiftmatrix V1 = V * ddspin(c, dir, M_PI);
     if(dd.mirrored) V1 = V1 * MirrorY;
     
     // loop:
@@ -1260,8 +1260,8 @@ EX namespace dice {
     
     if(1) {
       dynamicval<eGeometry> g(geometry, gSphere);
-      ld alpha = 360 * degree / dw->order;
-      ld beta = 180 * degree / dw->facesides;
+      ld alpha = TAU / dw->order;
+      ld beta = M_PI / dw->facesides;
       inradius  = edge_of_triangle_with_angles(alpha, beta, beta);
       outradius = edge_of_triangle_with_angles(beta, alpha, beta);
       }
@@ -1273,11 +1273,13 @@ EX namespace dice {
       dynamicval<eGeometry> g(geometry, gSphere);
       hyperpoint de = direct_exp(log_shift);
       S = rgpushxto0(de);
+      #if MAXMDIM >= 4
       if(GDIM == 3) {
-        for(int i=0; i<4; i++) swap(S[i][2], S[i][3]);
-        for(int i=0; i<4; i++) swap(S[2][i], S[3][i]);
+        for(int i=0; i<MAXMDIM; i++) swap(S[i][2], S[i][3]);
+        for(int i=0; i<MAXMDIM; i++) swap(S[2][i], S[3][i]);
         }
-      for(int i=0; i<4; i++) S[i][1] *= -1;
+      #endif
+      for(int i=0; i<MAXMDIM; i++) S[i][1] *= -1;
       }
     
     add_to_queue(S, val);
@@ -1320,7 +1322,7 @@ EX namespace dice {
       
       for(int d=0; d<si; d++) {
         dynamicval<eGeometry> g(geometry, highdim);
-        add_to_queue(T * cspin(0, 1, 2*M_PI*d/si) * cspin(2, 0, inradius) * cspin(0, 1, M_PI-2*M_PI*dw->spins[ws][d]/si), dw->sides[ws][d]);
+        add_to_queue(T * cspin(0, 1, TAU*d/si) * cspin(2, 0, inradius) * cspin(0, 1, M_PI-TAU*dw->spins[ws][d]/si), dw->sides[ws][d]);
         }
       
       if(1) {
@@ -1348,22 +1350,28 @@ EX namespace dice {
       auto sphere_to_space = [&] (hyperpoint h) {
         if(fpp) return h;
         if(osphere) {
-          h[2] = 1 - h[2]; h[3] = 0;
+          h[2] = 1 - h[2];
+          #if MAXMDIM > 3
+          h[3] = 0;
+          #endif
           return h;
           }
         if(oeuclid) { h[2] = 1-h[2]; return h; }
         ld z = asin_auto(h[2]);
         h = zpush(-z) * h;
-        h[2] = h[3]; h[3] = 0;
+        #if MAXMDIM > 3
+        h[2] = h[3];
+        h[3] = 0;
+        #endif
         dynamicval<eGeometry> g(geometry, orig);
-        return zshift(h, geom3::scale_at_lev(z));
+        return orthogonal_move(h, z);
         };
 
       for(int d=0; d<=si; d++) {
         hyperpoint h, hs;
         if(1) {
           dynamicval<eGeometry> g(geometry, highdim);
-          h = zpush(base_to_base) * T * cspin(0, 1, 2*M_PI*(d+.5)/si) * cspin(2, 0, outradius) * zpush0(dieradius);
+          h = zpush(base_to_base) * T * cspin(0, 1, TAU*(d+.5)/si) * cspin(2, 0, outradius) * zpush0(dieradius);
           if(d < si) face[d] = h;
           hs = sphere_to_space(h);
           }
@@ -1393,7 +1401,7 @@ EX namespace dice {
         cy = face[1] - (face[3] + face[4]) * .4;
         }
       
-      queuecurve(V1, 0xFFFFFFFF, color & 0xFFFFFF9F, prio);
+      queuecurve(V1, (poly_outline == OUTLINE_NONE) ? 0xFFFFFFFF : poly_outline, color & 0xFFFFFF9F, prio);
       
       #if !CAP_EXTFONT
       if(!vid.usingGL) continue;
@@ -1461,5 +1469,9 @@ EX namespace dice {
   EX bool swap_forbidden(cell *a, cell *b) { return false; }
   EX void chaos_swap(cellwalker wa, cellwalker wb) {}
 EX }
+
+EX namespace mine {
+  EX bool in_minesweeper() { return false; }
+EX }
 #endif
 }
