@@ -39,6 +39,7 @@ static const int POLY_FAT = (1<<26);            // fatten this model in WRL expo
 static const int POLY_SHADE_TEXTURE = (1<<27);  // texture has 'z' coordinate for shading
 static const int POLY_ONE_LEVEL = (1<<28);      // only one level of the universal cover in SL(2,R)
 static const int POLY_APEIROGONAL = (1<<29);    // only vertices indexed up to she are drawn as the boundary
+static const int POLY_NO_FOG = (1<<30);         // disable fog for this
 
 /** \brief A graphical element that can be drawn. Objects are not drawn immediately but rather queued.
  *
@@ -167,35 +168,40 @@ EX int texts_merged;
 EX int shapes_merged;
 
 #if MINIMIZE_GL_CALLS
-color_t triangle_color, line_color;
+PPR lprio;
 ld m_shift;
-vector<glvertex> triangle_vertices;
-vector<glvertex> line_vertices;
+vector<glhr::colored_vertex> triangle_vertices;
+vector<glhr::colored_vertex> line_vertices;
 #endif
 
 EX void glflush() {
   DEBBI(DF_GRAPH, ("glflush"));
   #if MINIMIZE_GL_CALLS
-  current_display->set_all(0, m_shift);
   if(isize(triangle_vertices)) {
-    // printf("%08X %08X | %d shapes, %d/%d vertices\n", triangle_color, line_color, shapes_merged, isize(triangle_vertices), isize(line_vertices));
-    if(triangle_color) {
+    // printf("%3d | %d shapes, %d/%d vertices\n", lprio, shapes_merged, isize(triangle_vertices), isize(line_vertices));
+    current_display->next_shader_flags = GF_VARCOLOR;
+    current_display->set_all(0, m_shift);
+    if(true) {
       glhr::be_nontextured();
       glapplymatrix(Id);
       glhr::current_vertices = NULL;
-      glhr::vertices(triangle_vertices);
-      glhr::color2(triangle_color);
+      glhr::prepare(triangle_vertices);
+      glhr::color2(0xFFFFFFFF);
       glDrawArrays(GL_TRIANGLES, 0, isize(triangle_vertices));
       }
     triangle_vertices.clear();
+    if(isize(line_vertices)) goto jump;
     }
   if(isize(line_vertices)) {
-    if(line_color) {
+    current_display->next_shader_flags = GF_VARCOLOR;
+    current_display->set_all(0, m_shift);
+    jump:
+    if(true) {
       glhr::be_nontextured();
       glapplymatrix(Id);
       glhr::current_vertices = NULL;
-      glhr::vertices(line_vertices);
-      glhr::color2(line_color);
+      glhr::prepare(line_vertices);
+      glhr::color2(0xFFFFFFFF);
       glDrawArrays(GL_LINES, 0, isize(line_vertices));
       }
     line_vertices.clear();
@@ -438,10 +444,10 @@ void addpoint(const shiftpoint& H) {
     if(sphere && pmodel == mdSpiral) {
       if(isize(glcoords)) {
         hyperpoint Hscr1;
-        shiftpoint H1 = H; H1.shift += 2 * M_PI;
+        shiftpoint H1 = H; H1.shift += TAU;
         applymodel(H1, Hscr1);
         if(hypot_d(2, Hlast-Hscr1) < hypot_d(2, Hlast-Hscr)) { Hscr = Hscr1; }
-        H1.shift -= 4 * M_PI;
+        H1.shift -= 2 * TAU;
         applymodel(H1, Hscr1);
         if(hypot_d(2, Hlast-Hscr1) < hypot_d(2, Hlast-Hscr)) { Hscr = Hscr1; }
         }
@@ -481,7 +487,11 @@ bool behind3(shiftpoint h) {
   if(pmodel == mdGeodesic) 
     return lp_apply(inverse_exp(h))[2] < 0;
   if(pmodel == mdLiePerspective)
-    return lp_apply(lie_log(unshift(h)))[2] < 0;
+    return lp_apply(lie_log(h))[2] < 0;
+  #if MAXMDIM >= 4
+  if(pmodel == mdRelPerspective)
+    return lp_apply(rel_log(h, false))[2] < 0;
+  #endif
   return h[2] < 0;
   }
 
@@ -499,7 +509,15 @@ void addpoly(const shiftmatrix& V, const vector<glvertex> &tab, int ofs, int cnt
     }
   tofix.clear(); knowgood = false;
   if(in_perspective()) {
-    if(poly_flags & POLY_TRIANGLES) {
+    if(get_shader_flags() & SF_SEMIDIRECT) {
+      dynamicval<bool> d(computing_semidirect, true);
+      for(int i=ofs; i<ofs+cnt; i++) {
+        hyperpoint Hscr;
+        applymodel(V * glhr::gltopoint(tab[i]), Hscr);
+        add1(Hscr);
+        }
+      }
+    else if(poly_flags & POLY_TRIANGLES) {
       for(int i=ofs; i<ofs+cnt; i+=3) {
         shiftpoint h0 = V * glhr::gltopoint(tab[i]);
         shiftpoint h1 = V * glhr::gltopoint(tab[i+1]);
@@ -647,25 +665,30 @@ void dqi_poly::gldraw() {
   
 #if MINIMIZE_GL_CALLS  
   if(current_display->stereo_active() == 0 && !tinf && (color == 0 || ((flags & (POLY_VCONVEX | POLY_CCONVEX)) && !(flags & (POLY_INVERSE | POLY_FORCE_INVERTED))))) {
-    if(color != triangle_color || outline != line_color || texts_merged || m_shift != V.shift) {
+    if(lprio != prio || texts_merged || m_shift != V.shift) {
       glflush();
-      triangle_color = color;
-      line_color = outline;
+      lprio = prio;
       m_shift = V.shift;
       }
     shapes_merged++;
 
     if((flags & POLY_CCONVEX) && !(flags & POLY_VCONVEX)) {
-      vector<glvertex> v2(cnt+1);
-      for(int i=0; i<cnt+1; i++) v2[i] = glhr::pointtogl( V.T * glhr::gltopoint( v[offset+i-1] ) );
+      vector<glhr::colored_vertex> v2(cnt+1);
+      for(int i=0; i<cnt+1; i++) v2[i] = glhr::colored_vertex( V.T * glhr::gltopoint( v[offset+i-1] ), color);
       if(color) for(int i=0; i<cnt; i++) triangle_vertices.push_back(v2[0]), triangle_vertices.push_back(v2[i]), triangle_vertices.push_back(v2[i+1]);
-      for(int i=1; i<cnt; i++) line_vertices.push_back(v2[i]), line_vertices.push_back(v2[i+1]);
+      if(outline) {
+        for(auto& v: v2) v.set_color(outline);
+        for(int i=1; i<cnt; i++) line_vertices.push_back(v2[i]), line_vertices.push_back(v2[i+1]);
+        }
       }
     else {
-      vector<glvertex> v2(cnt);
-      for(int i=0; i<cnt; i++) v2[i] = glhr::pointtogl( V.T * glhr::gltopoint( v[offset+i] ) );
+      vector<glhr::colored_vertex> v2(cnt);
+      for(int i=0; i<cnt; i++) v2[i] = glhr::colored_vertex( V.T * glhr::gltopoint( v[offset+i] ), color);
       if(color) for(int i=2; i<cnt-1; i++) triangle_vertices.push_back(v2[0]), triangle_vertices.push_back(v2[i-1]), triangle_vertices.push_back(v2[i]);
-      for(int i=1; i<cnt; i++) line_vertices.push_back(v2[i-1]), line_vertices.push_back(v2[i]);
+      if(outline) {
+        for(auto& v: v2) v.set_color(outline);
+        for(int i=1; i<cnt; i++) line_vertices.push_back(v2[i-1]), line_vertices.push_back(v2[i]);
+        }
       }
     return;
     }
@@ -679,6 +702,7 @@ void dqi_poly::gldraw() {
     else
       glhr::be_textured();
     if(flags & POLY_SHADE_TEXTURE) current_display->next_shader_flags |= GF_TEXTURE_SHADED;
+    if(flags & POLY_NO_FOG) current_display->next_shader_flags |= GF_NO_FOG;
     glBindTexture(GL_TEXTURE_2D, tinf->texture_id);
     if(isize(tinf->colors))
       glhr::vertices_texture_color(v, tinf->tvertices, tinf->colors, offset, offset_texture);
@@ -688,6 +712,7 @@ void dqi_poly::gldraw() {
     }
   else { 
     glhr::be_nontextured();
+    if(flags & POLY_NO_FOG) current_display->next_shader_flags |= GF_NO_FOG;
     glhr::vertices(v);
     }
   
@@ -700,8 +725,13 @@ void dqi_poly::gldraw() {
       current_display->set_all(ed, sl2 ? 0 : V.shift);
       glhr::set_index_sl(V.shift + M_PI * min_slr * hybrid::csteps / cgi.psl_steps);
       }
+    else if(sl2) {
+      current_display->set_all(ed, 0);
+      glhr::set_index_sl(V.shift);
+      }
     else {
-      current_display->set_all(ed, V.shift);
+      current_display->set_all(ed, sl2 ? 0 : V.shift);
+      glhr::set_index_sl(V.shift + M_PI * min_slr * hybrid::csteps / cgi.psl_steps);
       }
     bool draw = color;
 
@@ -717,7 +747,7 @@ void dqi_poly::gldraw() {
         glhr::color2(color, (flags & POLY_INTENSE) ? 2 : 1);
         glhr::set_depthtest(model_needs_depth() && prio < PPR::SUPERLINE);
         glhr::set_depthwrite(model_needs_depth() && prio != PPR::TRANSPARENT_SHADOW && prio != PPR::EUCLIDEAN_SKY);
-        glhr::set_fogbase(prio == PPR::SKY ? 1.0 + (euclid ? 20 : 5 / sightranges[geometry]) : 1.0);
+        glhr::set_fogbase(prio == PPR::SKY ? 1.0 + ((abs(cgi.SKY - cgi.LOWSKY)) / sightranges[geometry]) : 1.0);
         glDrawArrays(GL_TRIANGLES, ioffset, cnt);
         }
       else {
@@ -728,7 +758,7 @@ void dqi_poly::gldraw() {
         glStencilOp( GL_INVERT, GL_INVERT, GL_INVERT);
         glStencilFunc( GL_ALWAYS, 0x1, 0x1 );
         glhr::color2(0xFFFFFFFF);
-        glDrawArrays(tinf ? GL_TRIANGLES : GL_TRIANGLE_FAN, offset, cnt);
+        glDrawArrays(tinf ? GL_TRIANGLES : GL_TRIANGLE_FAN, ioffset, cnt);
         
         current_display->set_mask(ed);
         glhr::color2(color);
@@ -756,7 +786,7 @@ void dqi_poly::gldraw() {
         else { 
           glStencilOp( GL_ZERO, GL_ZERO, GL_ZERO);
           glStencilFunc( GL_EQUAL, 1, 1);
-          glDrawArrays(tinf ? GL_TRIANGLES : GL_TRIANGLE_FAN, offset, cnt);
+          glDrawArrays(tinf ? GL_TRIANGLES : GL_TRIANGLE_FAN, ioffset, cnt);
           }
         
         glDisable(GL_STENCIL_TEST);
@@ -799,7 +829,13 @@ void dqi_poly::gldraw() {
 #endif
 
 EX ld scale_at(const shiftmatrix& T) {
-  if(GDIM == 3 && pmodel == mdPerspective) return 1 / abs((tC0(unshift(T)))[2]);
+  if(GDIM == 3 && pmodel == mdPerspective) {
+    ld z = (tC0(unshift(T)))[2];
+    if(z == 0) return 1;
+    z = 1 / abs(z);
+    if(z > 10) return 10;
+    return z;
+    }
   if(sol) return 1;
   hyperpoint h1, h2, h3;
   applymodel(tC0(T), h1);
@@ -831,6 +867,17 @@ EX ld linewidthat(const shiftpoint& h) {
       return dfc;
       }
     }
+  else if(hyperbolic && pmodel == mdRelPerspective) {
+    if(abs(h[3]) < 1e-6) return 1;
+    return 1 / (1 + abs(h[3]));
+    }
+  else if(sl2 && pmodel == mdRelPerspective) {
+    if(abs(h[2]) < 1e-6) return 1;
+    return 1 / (1 + abs(h[2]));
+    }
+  else if(hyperbolic && GDIM == 3 && pmodel == mdPerspective && pconf.alpha == 0 && h[3] < 0.99) {
+    return 1;
+    }
   else if(perfect_linewidth >= (inHighQual ? 1 : 2)) {
     hyperpoint h0 = h.h / zlevel(h.h);
     shiftmatrix T = shiftless(rgpushxto0(h0), h.shift);
@@ -1081,7 +1128,7 @@ EX namespace s2xe {
     }
   
   void add2(pt h, int gen) {
-    glcoords.push_back(glhr::pointtogl(point31(sin(h[0]) * (h[1] + 2 * M_PI * gen), cos(h[0]) * (h[1] + 2 * M_PI * gen), h[2])));
+    glcoords.push_back(glhr::pointtogl(point31(sin(h[0]) * (h[1] + TAU * gen), cos(h[0]) * (h[1] + TAU * gen), h[2])));
     stinf.tvertices.push_back(glhr::makevertex(h[3], h[4], 0));
     }
 
@@ -1098,13 +1145,13 @@ EX namespace s2xe {
   bool to_right(const pt& h2, const pt& h1) {
     ld x2 = h2[0];
     ld x1 = h1[0];
-    if(x2 < x1) x2 += 2 * M_PI;
+    if(x2 < x1) x2 += TAU;
     return x2 >= x2 && x2 <= x1 + M_PI;
     }
   
   EX int qrings = 32;
   
-  ld seg() { return 2 * M_PI / qrings; }
+  ld seg() { return TAU / qrings; }
   
   void add_ortho_triangle(pt bl, pt tl, pt br, pt tr) {
     
@@ -1137,12 +1184,12 @@ EX namespace s2xe {
     }
   
   void add_ordered_triangle(array<pt, 3> v) {
-    if(v[1][0] < v[0][0]) v[1][0] += 2 * M_PI;
-    if(v[2][0] < v[1][0]) v[2][0] += 2 * M_PI;
+    if(v[1][0] < v[0][0]) v[1][0] += TAU;
+    if(v[2][0] < v[1][0]) v[2][0] += TAU;
     if(v[2][0] - v[0][0] < 1e-6) return;
     ld x = (v[1][0] - v[0][0]) / (v[2][0] - v[0][0]);
     
-    if(v[2][0] < v[0][0] + M_PI / 4 && maxy < M_PI - M_PI/4 && sightranges[geometry] <= 5) {
+    if(v[2][0] < v[0][0] + 45._deg && maxy < 135._deg && sightranges[geometry] <= 5) {
       addall(v[0], v[1], v[2]);
       return;
       }
@@ -1168,16 +1215,16 @@ EX namespace s2xe {
     }
 
   void add_triangle_around(array<pt, 3> v) {
-    ld baseheight = (v[0][1] > M_PI/2) ? M_PI : 0;
+    ld baseheight = (v[0][1] > 90._deg) ? M_PI : 0;
     ld tu = (v[0][3] + v[1][3] + v[2][3]) / 3;
     ld tv = (v[0][4] + v[1][4] + v[2][4]) / 3;
     array<pt, 3> vhigh;
     for(int i=0; i<3; i++) { vhigh[i] = v[i]; vhigh[i][1] = baseheight; vhigh[i][3] = tu; vhigh[i][4] = tv; }
-    if(v[1][0] < v[0][0]) v[1][0] = v[1][0] + 2 * M_PI, vhigh[1][0] = vhigh[1][0] + 2 * M_PI;
+    if(v[1][0] < v[0][0]) v[1][0] = v[1][0] + TAU, vhigh[1][0] = vhigh[1][0] + TAU;
     add_ortho_triangle(v[0], vhigh[0], v[1], vhigh[1]);
-    if(v[2][0] < v[1][0]) v[2][0] = v[2][0] + 2 * M_PI, vhigh[2][0] = vhigh[2][0] + 2 * M_PI;
+    if(v[2][0] < v[1][0]) v[2][0] = v[2][0] + TAU, vhigh[2][0] = vhigh[2][0] + TAU;
     add_ortho_triangle(v[1], vhigh[1], v[2], vhigh[2]);
-    if(v[0][0] < v[2][0]) v[0][0] = v[0][0] + 2 * M_PI, vhigh[0][0] = vhigh[0][0] + 2 * M_PI;
+    if(v[0][0] < v[2][0]) v[0][0] = v[0][0] + TAU, vhigh[0][0] = vhigh[0][0] + TAU;
     add_ortho_triangle(v[2], vhigh[2], v[0], vhigh[0]);
     }
 
@@ -1188,12 +1235,12 @@ EX namespace s2xe {
     
     minz = min(abs(v[0][2]), max(abs(v[1][2]), abs(v[2][2])));
     auto& s = sightranges[geometry];
-    maxgen = sqrt(s * s - minz * minz) / (2 * M_PI) + 1;
+    maxgen = sqrt(s * s - minz * minz) / TAU + 1;
     
     maxy = max(v[0][1], max(v[1][1], v[2][1]));
     miny = min(v[0][1], min(v[1][1], v[2][1]));
     with_zero = true;
-    if(maxy < M_PI / 4) {
+    if(maxy < 45._deg) {
       add2(v[0], 0);
       add2(v[1], 0);
       add2(v[2], 0);
@@ -1283,7 +1330,7 @@ void draw_s2xe0(dqi_poly *p) {
   set_width(1);
   glcoords.clear();
 
-  int maxgen = sightranges[geometry] / (2 * M_PI) + 1;
+  int maxgen = sightranges[geometry] / TAU + 1;
   
   auto crossdot = [&] (const hyperpoint h1, const hyperpoint h2) { return make_pair(h1[0] * h2[1] - h1[1] * h2[0], h1[0] * h2[0] + h1[1] * h2[1]); };
   vector<point_data> pd;
@@ -1315,7 +1362,7 @@ void draw_s2xe0(dqi_poly *p) {
   for(int i=0; i<p->cnt; i++) {
     auto &c1 = pd[i];
     auto &c0 = pd[i==0?p->cnt-1 : i-1];
-    if(c1.distance > M_PI/2 && c0.distance > M_PI/2 && crossdot(c0.direction, c1.direction).second < 0) return;
+    if(c1.distance > 90._deg && c0.distance > 90._deg && crossdot(c0.direction, c1.direction).second < 0) return;
     if(c1.bad == 2) return;
     if(c1.bad == 1) no_gens = true;
     }
@@ -1327,12 +1374,12 @@ void draw_s2xe0(dqi_poly *p) {
       angles[i] = atan2(pd[i].direction[1], pd[i].direction[0]);
       }
     sort(angles.begin(), angles.end());
-    angles.push_back(angles[0] + 2 * M_PI);
+    angles.push_back(angles[0] + TAU);
     bool ok = false;
     for(int i=1; i<isize(angles); i++)
       if(angles[i] >= angles[i-1] + M_PI) ok = true;
     if(!ok) {
-      for(auto &c: pd) if(c.distance > M_PI/2) return;
+      for(auto &c: pd) if(c.distance > 90._deg) return;
       no_gens = true;
       }
     }
@@ -1342,7 +1389,7 @@ void draw_s2xe0(dqi_poly *p) {
   for(int gen=-g; gen<=g; gen++) {
     for(int i=0; i<p->cnt; i++) {
       auto& cur = pd[i];
-      ld d = cur.distance + 2 * M_PI * gen;
+      ld d = cur.distance + TAU * gen;
       hyperpoint h;
       h[0] = cur.direction[0] * d;
       h[1] = cur.direction[1] * d;
@@ -1409,8 +1456,6 @@ void draw_stretch(dqi_poly *p) {
       auto &lb = results[i+1];
       auto &lc = results[i+2];
       
-      int ia = 0, ib = 0, ic = 0;
-      
       for(auto& ha: la) for(auto& hb: lb) if(test(ha, hb))
         for(auto& hc: lc) if(test(ha, hc) && test(hb, hc)) {
         
@@ -1420,7 +1465,6 @@ void draw_stretch(dqi_poly *p) {
         if(p->tinf) 
           for(int j=0; j<3; j++)
             stinf.tvertices.push_back(p->tinf->tvertices[p->offset_texture+i+j]);
-        ia++; ib++; ic++;
         }
       }
     npoly.cnt = isize(glcoords);  
@@ -1532,7 +1576,7 @@ EX namespace ods {
 
       for(int j=0; j<3; j++) {
         hyperpoint o = p->V * glhr::gltopoint((*p->tab)[p->offset+i+j]);
-        if(nonisotropic || prod) {
+        if(nonisotropic || gproduct) {
           o = lp_apply(inverse_exp(o, iTable, false));
           o[3] = 1;
           dynamicval<eGeometry> g(geometry, gEuclid);
@@ -1545,9 +1589,9 @@ EX namespace ods {
       
       for(int j=0; j<6; j++) {
         // let Delta be from 0 to 2PI
-        if(h[j][2]<0) h[j][2] += 2 * M_PI;
+        if(h[j][2]<0) h[j][2] += TAU;
         // Theta is from -PI/2 to PI/2. Let it be from 0 to PI
-        h[j][1] += global_projection * M_PI/2;
+        h[j][1] += global_projection * 90._deg;
         h[j][3] = 1;
         }
 
@@ -1564,8 +1608,8 @@ EX namespace ods {
       cyclefix(h[4][0], h[3][0]);
       cyclefix(h[5][0], h[3][0]);
 
-      if(abs(h[1][1] - h[0][1]) > M_PI/2) goto next_i;
-      if(abs(h[2][1] - h[0][1]) > M_PI/2) goto next_i;
+      if(abs(h[1][1] - h[0][1]) > 90._deg) goto next_i;
+      if(abs(h[2][1] - h[0][1]) > 90._deg) goto next_i;
       
       if(h[0][0] < -M_PI || h[0][0] > M_PI) println(hlog, h[0][0]);
 
@@ -1574,7 +1618,7 @@ EX namespace ods {
         if(h[1][0] < -M_PI || h[2][0] < -M_PI) lst++;
         if(h[1][0] > +M_PI || h[2][0] > +M_PI) fst--;
         for(int x=fst; x<=lst; x++) for(int j=0; j<3; j++) {
-          glcoords.push_back(glhr::makevertex(h[j][0] + 2 * M_PI * x, h[j][1], h[j][2]));
+          glcoords.push_back(glhr::makevertex(h[j][0] + TAU * x, h[j][1], h[j][2]));
           if(npoly.tinf) stinf.tvertices.push_back(p->tinf->tvertices[p->offset_texture+i+j]);
           }
         }
@@ -1595,21 +1639,46 @@ EX namespace ods {
 bool broken_projection(dqi_poly& p0) {
   int broken_coord = models::get_broken_coord(pmodel);
   static bool in_broken = false;
+  bool both_broken = pmodel == mdConformalSquare;
+
+  transmatrix T = p0.V.T, IT = Id, FT = Id;
+  if(both_broken) FT = cspin(0, 1, 45._deg), T = FT * T, IT = cspin(0, 1, -45._deg);
+
+  ld zlow = 0;
+  if(both_broken) {
+    ld t = pconf.model_transition;
+    zlow = (1-t*t) / (1+t*t);
+    }
+  // x * mt / (1-z) <= 1
+  // sqrt(1-z*z) * mt / (1-z) <= 1
+  // sqrt(1-z*z) <= (1-z) / mt
+
   if(broken_coord && !in_broken) {
 
     int zcoord = broken_coord;
     int ycoord = 3 - zcoord;
+    int xcoord = 0;
+
+    zcoord = 2;
     
     vector<hyperpoint> all;
     for(int i=0; i<p0.cnt; i++) 
-      all.push_back(p0.V.T * glhr::gltopoint((*p0.tab)[p0.offset+i]));
+      all.push_back(T * glhr::gltopoint((*p0.tab)[p0.offset+i]));
     int fail = 0;
     int last_fail;
 
     for(auto& h: all) models::apply_orientation(h[0], h[1]);
-    
+
+    auto break_in_xz = [&] (hyperpoint a, hyperpoint b, int xcoord, int zcoord) {
+      return a[xcoord] * b[xcoord] <= 0 && (a[xcoord] * (b[zcoord]+zlow) - b[xcoord] * (a[zcoord]+zlow)) * (a[xcoord] - b[xcoord]) < 0;
+      };
+ 
     auto break_in = [&] (hyperpoint a, hyperpoint b) {
-      return a[0] * b[0] <= 0 && (a[0] * b[zcoord] - b[0] * a[zcoord]) * (a[0] - b[0]) < 0;
+      if(both_broken) {
+        for(int xc=0; xc<2; xc++) {if(break_in_xz(a, b, xc, zcoord)) { xcoord = xc; ycoord = 1-xc; return true; } }
+        return false;
+        }
+      return break_in_xz(a, b, xcoord, zcoord);
       };
     
     for(int i=0; i<p0.cnt-1; i++) 
@@ -1619,7 +1688,7 @@ bool broken_projection(dqi_poly& p0) {
     dqi_poly p = p0;
     p.tab = &v;
     p.offset = 0;
-    p.V.T = Id;
+    p.V.T = IT;
 
     /* we don't rotate h's back, just change p.V */
     for(int i=0; i<3; i++)
@@ -1628,7 +1697,12 @@ bool broken_projection(dqi_poly& p0) {
     if(fail) {
       if(p0.tinf) return true;
       dynamicval<bool> ib(in_broken, true);
-      ld part = ilerp(all[last_fail][0], all[last_fail+1][0], 0);
+
+      ld part = ilerp(all[last_fail][xcoord], all[last_fail+1][xcoord], 0);
+      if(both_broken && all[last_fail][ycoord] * all[last_fail+1][ycoord] < 0) {
+        ld part2 = ilerp(all[last_fail][ycoord], all[last_fail+1][ycoord], 0);
+        if(part2 > part) part = part2, swap(xcoord, ycoord);
+        }
       hyperpoint initial = normalize(lerp(all[last_fail], all[last_fail+1], 1 - (1-part) * .99));
       bool have_initial = true;
       v.push_back(glhr::pointtogl(initial));
@@ -1636,19 +1710,48 @@ bool broken_projection(dqi_poly& p0) {
       int at = last_fail;
       do {
         v.push_back(glhr::pointtogl(all[at]));
-        if(at == p0.cnt-1 && all[at] != all[0]) {
+        if(at == p0.cnt-1 && sqhypot_d(2, all[at] - all[0]) > 1e-6) {
           p.cnt = isize(v); p.draw(); v.clear(); at = 0;
           have_initial = false;
           }
         int next = at+1;
         if(next == p0.cnt) next = 0;
         if(break_in(all[at], all[next])) {
-          ld part = ilerp(all[at][0], all[next][0], 0);
+          ld part = ilerp(all[at][xcoord], all[next][xcoord], 0);
+          if(both_broken && all[at][ycoord] * all[next][ycoord] < 0) {
+            ld part2 = ilerp(all[at][ycoord], all[next][ycoord], 0);
+            if(part2 < part) part = part2, swap(xcoord, ycoord);
+            }
+
           hyperpoint final = normalize(lerp(all[at], all[next], part * .99));
           v.push_back(glhr::pointtogl(final));
           if(have_initial) {
             int max = 4 << vid.linequality;
-            if(final[0] * initial[0] > 0) {
+            if(both_broken) {
+              auto square_close_corner = [&] (hyperpoint h) {
+                hyperpoint end = -C0;
+                end[0] = 0.01 * signum(h[0]);
+                end[1] = 0.01 * signum(h[1]);
+                /* if(abs(h1[0]) > abs(h1[1]))
+                  end[0] = 0.01 * signum(h1[0]), end[1] = 0.001 * signum(h1[1]);
+                else
+                  end[1] = 0.01 * signum(h1[1]), end[0] = 0.001 * signum(h1[0]); */
+                return normalize(end);
+                };
+              hyperpoint endf = square_close_corner(final);
+              hyperpoint endi = square_close_corner(initial);
+              if(endf != endi) {
+                for(int i=1; i<=max; i++)
+                  v.push_back(glhr::pointtogl(lerp(final, endf, i * 1. / max)));
+                for(int i=0; i<=max; i++)
+                  v.push_back(glhr::pointtogl(lerp(endi, initial, i * 1. / max)));
+                }
+              else {
+                for(int i=1; i<=max; i++)
+                  v.push_back(glhr::pointtogl(lerp(final, initial, i * 1. / max)));
+                }
+              }
+            else if(final[xcoord] * initial[xcoord] > 0) {
               for(int i=1; i<=max; i++)
                 v.push_back(glhr::pointtogl(lerp(final, initial, i * 1. / max)));
               }
@@ -2157,7 +2260,7 @@ EX void quickqueue() {
 
 /* todo */
 ld xintval(const shiftpoint& h) {
-  if(sphereflipped()) return -h.h[2];
+  if(sphere_flipped) return -h.h[2];
   if(hyperbolic) return -h.h[2];
   return -intval(h.h, C0);
   }
@@ -2171,7 +2274,7 @@ int qp[PMAX], qp0[PMAX];
 
 color_t darken_color(color_t& color, bool outline) {
   int alpha = color & 255;
-  if(sphere && pmodel == mdDisk && pconf.alpha <= 1)
+  if(sphere && pmodel == mdDisk && pconf.alpha <= 0.99)
     return 0;
   else {
     if(outline && alpha < 255) 
@@ -2201,13 +2304,13 @@ EX void sort_drawqueue() {
 
   #if MINIMIZE_GL_CALLS
   map<color_t, vector<unique_ptr<drawqueueitem>>> subqueue;
-  for(auto& p: ptds) subqueue[(p->prio == PPR::CIRCLE || p->prio == PPR::OUTCIRCLE) ? 0 : p->outline_group()].push_back(move(p));
+  for(auto& p: ptds) subqueue[(p->prio == PPR::CIRCLE || p->prio == PPR::OUTCIRCLE) ? 0 : p->outline_group()].push_back(std::move(p));
   ptds.clear();
-  for(auto& p: subqueue) for(auto& r: p.second) ptds.push_back(move(r));
+  for(auto& p: subqueue) for(auto& r: p.second) ptds.push_back(std::move(r));
   subqueue.clear();
-  for(auto& p: ptds) subqueue[(p->prio == PPR::CIRCLE || p->prio == PPR::OUTCIRCLE) ? 0 : p->color].push_back(move(p));
+  for(auto& p: ptds) subqueue[(p->prio == PPR::CIRCLE || p->prio == PPR::OUTCIRCLE) ? 0 : p->color].push_back(std::move(p));
   ptds.clear();
-  for(auto& p: subqueue) for(auto& r: p.second) ptds.push_back(move(r));
+  for(auto& p: subqueue) for(auto& r: p.second) ptds.push_back(std::move(r));
   #endif
     
   for(auto& p: ptds) {
@@ -2228,7 +2331,7 @@ EX void sort_drawqueue() {
   vector<unique_ptr<drawqueueitem>> ptds2;  
   ptds2.resize(siz);
   
-  for(int i = 0; i<siz; i++) ptds2[qp[int(ptds[i]->prio)]++] = move(ptds[i]);
+  for(int i = 0; i<siz; i++) ptds2[qp[int(ptds[i]->prio)]++] = std::move(ptds[i]);
   swap(ptds, ptds2);
   }
 
@@ -2252,7 +2355,7 @@ EX void draw_backside() {
         ptd->draw();
     }
 
-  spherespecial = sphereflipped() ? 1 : -1;
+  spherespecial = sphere_flipped ? 1 : -1;
   reset_projection();
   
   if(pmodel == mdRotatedHyperboles) {
@@ -2429,7 +2532,7 @@ EX void drawqueue() {
   #endif
 
   #if MAXMDIM >= 4 && CAP_GL
-  if(WDIM == 2 && GDIM == 3 && hyperbolic && !vrhr::rendering()) make_air();
+  make_air();
   #endif
   
   #if CAP_VR
@@ -2479,6 +2582,25 @@ EX void drawqueue() {
         });
     }
 
+  if(draw_plain_floors && (default_flooralpha < 255 || svg::in)) for(PPR p: {PPR::FLOOR}) {
+    int pp = int(p);
+    if(qp0[pp] == qp[pp]) continue;
+    auto get_z = [&] (const unique_ptr<drawqueueitem>& p) -> ld {
+      auto d = dynamic_cast<dqi_poly*> (&*p);
+      if(!d) return 0;
+      hyperpoint h = Hypc;
+
+      for(int i=0; i<d->cnt; i++) h += glhr::gltopoint( (*d->tab)[d->offset + i] );
+      h /= d->cnt; normalize(h);
+      h = unshift(d->V) * h;
+      return h[2];
+      };
+    sort(&ptds[qp0[int(p)]], &ptds[qp[int(p)]],
+      [&] (const unique_ptr<drawqueueitem>& p1, const unique_ptr<drawqueueitem>& p2) {
+        return get_z(p1) > get_z(p2);
+        });
+    }
+
 #if CAP_SDL
   if(current_display->stereo_active() && !vid.usingGL) {
 
@@ -2619,6 +2741,10 @@ EX void curvepoint(const hyperpoint& H1) {
   curvedata.push_back(glhr::pointtogl(H1));
   }
 
+EX void curvepoint_first() {
+  curvedata.push_back(curvedata[curvestart]);
+  }
+
 EX dqi_poly& queuecurve(const shiftmatrix& V, color_t linecol, color_t fillcol, PPR prio) {
   auto &res = queuetable(V, curvedata, isize(curvedata)-curvestart, linecol, fillcol, prio);
   res.offset = curvestart;
@@ -2740,7 +2866,7 @@ EX void write_in_space(const shiftmatrix& V, int fsize, double size, const strin
     }
   
   if(frame) for(int i=0; i<360; i+=45) {
-    auto &res = queuetable(V * xspinpush(i*degree, frame*scale), curvedata, isize(curvedata)-curvestart, col & 0xFF, col & 0xFF, prio);
+    auto &res = queuetable(V * xspinpush(i*degree, frame*scale), curvedata, isize(curvedata)-curvestart, poly_outline, poly_outline, prio);
     res.offset = curvestart;
     res.offset_texture = fstart;
     res.tinf = &finf;
