@@ -46,10 +46,12 @@ EX namespace euc {
 
     switch(g) {
       case gCubeTiling:
+      case gMengerSponge:
         shifttable = { +D0, +D1, +D2 };
         break;
       
       case gRhombic3:
+      case gSierpinskiTet:
         shifttable = { D0+D1, D0+D2, D1+D2, D1-D2, D0-D2, D0-D1 };
         break;
       
@@ -58,10 +60,13 @@ EX namespace euc {
         break;
       
       case gEuclid:
+      case gSierpinski3:
+      case gSixFlake:
         shifttable = { D0, D1, D1-D0, -D0, -D1, D0-D1 };
         break;
 
       case gEuclidSquare:
+      case gSierpinski4:
         shifttable = { D0, D1, -D0, -D1 };
         break;
       
@@ -137,10 +142,10 @@ EX namespace euc {
     map<gp::loc, struct cdata> eucdata;
     
     void compute_tmatrix() {
+      cgi.require_basics();
       shifttable = get_shifttable();
       tmatrix.resize(S7);
-      for(int i=0; i<S7; i++) 
-        tmatrix[i] = eumove(shifttable[i]);
+      for(int i=0; i<S7; i++) tmatrix[i] = eumove(shifttable[i]);
       }
     
     void on_dim_change() override {
@@ -236,6 +241,7 @@ EX namespace euc {
       }
 
     transmatrix adj(cell *c, int i) override {
+      if(dont_inverse()) return adj(c->master, i);
       if(WDIM == 3) return adj(c->master, i);
       else return hrmap_standard::adj(c, i);
       }
@@ -255,8 +261,11 @@ EX namespace euc {
         bool draw = drawcell_subs(c, V * spin(master_to_c7_angle()));
         if(in_wallopt() && isWall3(c) && isize(dq::drawqueue) > 1000 && !hybrid::pmap) continue;
   
-        if(draw) for(int i=0; i<S7; i++)
-          dq::enqueue_by_matrix(h->move(i), optimized_shift(V * adj(h, i)));
+        if(draw) for(int i=0; i<S7; i++) {
+          auto V1 = V * adj(h, i);
+          if(geom3::apply_break_cylinder && cgi.emb->break_cylinder(V, V1)) continue;
+          dq::enqueue_by_matrix(h->move(i), optimized_shift(V1));
+          }
         }
       }
     
@@ -882,7 +891,7 @@ EX namespace euc {
     auto& T_edit = eu_edit.user_axes;
     auto& twisted_edit = eu_edit.twisted;
     cmode = sm::SIDE | sm::MAYDARK | sm::TORUSCONFIG;
-    gamescreen(1);  
+    gamescreen();
     dialog::init(XLAT("Euclidean quotient spaces"));
     
     for(int y=0; y<dim+1; y++)
@@ -1200,13 +1209,14 @@ EX transmatrix eumove(coord co) {
     }
   transmatrix Mat = Id;
   if(a4) {
-    Mat[0][LDIM] += co[0] * cgi.tessf;
-    Mat[1][LDIM] += co[1] * cgi.tessf;
+    Mat[0][2] += co[0] * cgi.tessf;
+    Mat[1][2] += co[1] * cgi.tessf;
     }
   else {
-    Mat[0][LDIM] += (co[0] + co[1] * .5) * cgi.tessf;
-    Mat[1][LDIM] += co[1] * q3 /2 * cgi.tessf;
+    Mat[0][2] += (co[0] + co[1] * .5) * cgi.tessf;
+    Mat[1][2] += co[1] * q3 /2 * cgi.tessf;
     }
+  if(embedded_plane) Mat = cgi.emb->base_to_actual(Mat);
   return Mat;
   }
 
@@ -1356,7 +1366,9 @@ EX void generate() {
  */
 EX bool in() { 
   if(fake::in()) return FPIU(in()); 
-  return euclid && standard_tiling(); 
+  if(geometry == gCubeTiling && (reg3::cubes_reg3 || !PURE)) return false;
+  if(cgflags & qEXPERIMENTAL) return false;
+  return meuclid && standard_tiling();
   }
 
 EX bool in(int dim) { return in() && WDIM == dim; }
