@@ -113,22 +113,23 @@ void expansion_analyzer::preliminary_grouping() {
       for(int v: rulegen::treestates[i].rules)
         if(v >= 0) children[i].push_back(v);
     }
-  else if(reg3::in_rule()) {
 #if MAXMDIM >= 4
+  else if(reg3::exact_rules()) {
     rootid = reg3::rule_get_root(0);
     auto& chi = reg3::rule_get_children();
-    N = isize(chi) / S7;    
+    auto& chpos = reg3::rule_get_childpos();
+    N = isize(chpos) - 1;
     children.resize(N);
     int k = 0;
-    for(int i=0; i<N; i++) for(int j=0; j<S7; j++) {
+    for(int i=0; i<N; i++) for(int j=0; j<chpos[i+1]-chpos[i]; j++) {
       int ck = chi[k];
       if(ck < -1) ck += (1<<16);
       if(ck >= 0)
         children[i].push_back(ck);
       k++;
       }
-#endif
     }
+#endif
   else {
     sample_id(currentmap->gamestart());
     // queue for, do not change to range-based for
@@ -144,7 +145,9 @@ void expansion_analyzer::preliminary_grouping() {
   }
 
 void expansion_analyzer::reduce_grouping() {
-  if(reg3::in_rule()) return;
+  #if MAXMDIM >= 4
+  if(reg3::exact_rules()) return;
+  #endif
   if(currentmap->strict_tree_rules()) return;
   int old_N = N;
   vector<int> grouping;
@@ -190,7 +193,7 @@ void expansion_analyzer::reduce_grouping() {
   for(int i=0; i<nogroups; i++) 
     for(int j: children[groupsample[i]])
       newchildren[i].push_back(grouping[j]);
-  children = move(newchildren);
+  children = std::move(newchildren);
   for(auto& p: codeid) p.second = grouping[p.second];
   N = nogroups;
   rootid = grouping[rootid];
@@ -238,7 +241,11 @@ bool expansion_analyzer::verify(int id) {
 
 int expansion_analyzer::valid(int v, int step) {
   if(step < 0) return 0;
-  int more = reg3::in_rule() ? 1 : 5;
+  #if MAXMDIM >= 4
+  int more = 5;
+  #else
+  int more = reg3::exact_rules() ? 1 : 5;
+  #endif
   #if CAP_GMP == 0
   if(get_descendants(step+v+v+more).approx_int() >= bignum::BASE) return 0;
   typedef ld val;
@@ -391,13 +398,15 @@ int type_in_quick(expansion_analyzer& ea, cell *c, const cellfunction& f) {
   }
 
 EX bool sizes_known() {
-  if(reg3::in_rule()) return true;
+  #if MAXMDIM >= 4
+  if(reg3::exact_rules()) return true;
+  #endif
   if(closed_manifold) return false;
   // Castle Anthrax is infinite
   if(bt::in()) return false;
   // not implemented
   if(arcm::in()) return false;
-  if(kite::in()) return false;
+  if(aperiodic) return false;
   if(currentmap->strict_tree_rules()) return true;
   if(arb::in()) return false;
   return true;  
@@ -421,10 +430,12 @@ string expansion_analyzer::approximate_descendants(int d, int max_length) {
   return XLAT("about ") + fts(pow(10, log_10 - more_digits)) + "E" + its(more_digits);
   }
 
+#if HDR
 enum eDistanceFrom { dfPlayer, dfStart, dfWorld };
+#endif
 EX string dfnames[3] = { "player", "start", "land" };
 
-eDistanceFrom distance_from = dfPlayer;
+EX eDistanceFrom distance_from = dfPlayer;
 
 #if HDR
 enum eNumberCoding { ncNone, ncDistance, ncType, ncDebug, ncError };
@@ -432,19 +443,31 @@ enum eNumberCoding { ncNone, ncDistance, ncType, ncDebug, ncError };
 EX string ncnames[5] = { "NO", "distance", "type", "debug", "error" };
 EX eNumberCoding number_coding = ncDistance;
 
-bool mod_allowed() {
-  return cheater || autocheat || arcm::in() || tour::on;
+EX bool mod_allowed() {
+  return cheater || autocheat || arcm::in() || arb::in() || tour::on;
+  }
+
+EX bool distances_legal(cell *c) {
+  if(mod_allowed()) return true;
+  switch(distance_from) {
+    case dfPlayer:
+      return true;
+    case dfStart:
+      return bt::in();
+    case dfWorld:
+      return c && among(c->land, laOcean, laIvoryTower, laEndorian, laDungeon, laTemple, laWhirlpool, laCanvas);
+    }
+  return false;
   }
 
 EX int curr_dist(cell *c) {
+  if(!distances_legal(c)) return 0;
   switch(distance_from) {
     case dfPlayer:
       return c->cpdist < INFD ? c->cpdist : celldistance(cwt.at, c);
     case dfStart:
       return celldist(c);
     case dfWorld:
-      if(!mod_allowed() && !among(c->land, laOcean, laIvoryTower, laEndorian, laDungeon, laTemple, laWhirlpool, laCanvas))
-        return 0;
       if((isCyclic(c->land) || among(c->land, laCanvas, laCaribbean, laStorms, laRlyeh))) {
         if(eubinary || c->master->alt) return celldistAlt(c);
         return UNKNOWN;
@@ -543,7 +566,7 @@ EX bool use_color_codes = true;
 EX bool use_analyzer = true;
 EX bool show_distance_lists = true;
 
-int first_distance = 0, scrolltime = 0;
+int last_distance = 16;
 bool scrolling_distances = false;
 
 EX map<int, color_t> expcolors;
@@ -580,7 +603,7 @@ void celldrawer::do_viewdist() {
       }
     case ncType: {
       int t = -1;
-      if(reg3::in_rule()) switch(distance_from) {
+      if(reg3::exact_rules()) switch(distance_from) {
         case dfPlayer: 
           t = -1;
           break;
@@ -621,9 +644,6 @@ void celldrawer::do_viewdist() {
   
   if(!dist_label_colored) dc = dist_label_color;
 
-  // string label = its(fieldpattern::getriverdistleft(c)) + its(fieldpattern::getriverdistright(c));
-  /* queuepolyat(V, shFloor[ct6], darkena(gradient(0, distcolors[cd&7], 0, .25, 1), fd, 0xC0),
-    PPR::TEXT); */
   if(label != "")
     queuestr(V, (isize(label) > 1 ? .6 : 1), label, 0xFF000000 + dc, 1);
   }
@@ -631,7 +651,7 @@ void celldrawer::do_viewdist() {
 EX void viewdist_configure_dialog() {
   dialog::init("");
   cmode |= sm::SIDE | sm::MAYDARK | sm::EXPANSION;
-  gamescreen(0);
+  gamescreen();
   
   dialog::addSelItem(XLAT("which distance"), XLAT(dfnames[distance_from]), 'c');
   dialog::add_action([] () { distance_from = mod_allowed() ? eDistanceFrom((distance_from + 1) % 3) : eDistanceFrom(2 - distance_from); });
@@ -641,25 +661,11 @@ EX void viewdist_configure_dialog() {
 
   dialog::addBoolItem_action(XLAT("color codes"), use_color_codes, 'u');
 
-  dialog::addSelItem(XLAT("display distances from"), its(first_distance), 'd');
-  dialog::add_action([] () { 
-    scrolling_distances = false;
-    dialog::editNumber(first_distance, 0, 3000, 1, 0, XLAT("display distances from"), "");
-    dialog::bound_low(0);
-    });
-
   dialog::addBoolItem(XLAT("strict tree maps"), currentmap->strict_tree_rules(), 's');
   dialog::add_action_push(rulegen::show);
 
-  int id = 0;
-  using namespace linepatterns;
-  for(auto& lp: {&patTriTree, &patTriRings, &patTriOther}) {
-    dialog::addColorItem(XLAT(lp->lpname), lp->color, '1'+(id++));
-    dialog::add_action([&lp] () {
-      dialog::openColorDialog(lp->color, NULL);
-      dialog::dialogflags |= sm::MAYDARK | sm::SIDE | sm::EXPANSION;
-      });
-    }
+  dialog::addItem(XLAT("line patterns"), 'L');
+  dialog::add_action_push(linepatterns::showMenu);
   
   if(!mod_allowed()) {
     dialog::addItem(XLAT("enable the cheat mode for additional options"), 'C');
@@ -715,21 +721,20 @@ string produce_coef_formula(vector<int> coef) {
   return fmt;
   }
 
+EX bool auto_extend = true;
+
 void expansion_analyzer::view_distances_dialog() {
   static int lastticks;
   if(scrolling_distances && !closed_manifold) {
-    scrolltime += SDL_GetTicks() - lastticks;
-    first_distance += scrolltime / scrollspeed;
-    scrolltime %= scrollspeed;
+    dialog::list_skip += (SDL_GetTicks() - lastticks) * dialog::dfspace / scrollspeed;
     }
   lastticks = SDL_GetTicks();
-  if(first_distance < 0) first_distance = 0;
   
   dynamicval<color_t> dv(distcolors[0], forecolor);
   dialog::init("");
-  cmode |= sm::DIALOG_STRICT_X | sm::EXPANSION;
+  cmode |= sm::DIALOG_STRICT_X | sm::EXPANSION | sm::AUTO_VALUES | sm::NARROW_LINES;
   
-  int maxlen = closed_manifold ? 128 : 16 + first_distance;
+  int maxlen = last_distance;
   vector<bignum> qty(maxlen);
   auto& expansion = get_expansion();
   
@@ -737,7 +742,7 @@ void expansion_analyzer::view_distances_dialog() {
   
   if(really_use_analyzer) {
     int t;
-    if(reg3::in_rule() || currentmap->strict_tree_rules()) {
+    if(reg3::exact_rules() || currentmap->strict_tree_rules()) {
       if(!N) preliminary_grouping();      
       t = rootid;
       }
@@ -768,13 +773,14 @@ void expansion_analyzer::view_distances_dialog() {
       }
     #endif
     }
-  
-  dialog::addBreak(100 - 100 * scrolltime / scrollspeed);
 
-  for(int i=first_distance; i<maxlen; i++) if(!qty[i].digits.empty())
-    dialog::addInfo(its(i) + ": " + qty[i].get_str(100), distcolors[i]);
-  
-  dialog::addBreak(100 * scrolltime / scrollspeed);
+  dialog::start_list(1600, 1600, 'a');
+  for(int i=0; i<maxlen; i++) if(!qty[i].digits.empty()) {
+    dialog::addSelItem(qty[i].get_str(100), " " + its(i), dialog::list_fake_key);
+    auto& last = dialog::lastItem();
+    last.color = last.colorv = distcolors[i];
+    }
+  dialog::end_list();
 
   if(sizes_known() || bt::in()) {
     if(euclid && !arb::in()) {
@@ -798,12 +804,24 @@ void expansion_analyzer::view_distances_dialog() {
   
   dialog::addItem(XLAT("scroll"), 'S');
   dialog::addItem(XLAT("configure"), 'C');
+  dialog::addSelItem(XLAT("display distances up to"), its(last_distance), 'D');
+  dialog::add_action([] () {
+    scrolling_distances = false;
+    dialog::editNumber(last_distance, 0, 3000, 1, 0, XLAT("display distances up to"), "");
+    dialog::bound_low(0);
+    dialog::extra_options = [] {
+      add_edit(auto_extend);
+      };
+    });
+
   dialog::display();
+  if(auto_extend && dialog::list_skip + dialog::list_actual_size == dialog::list_full_size) last_distance++;
   }
 
 EX void enable_viewdists() {
-  first_distance = 0;
-  scrolltime = 0;
+  last_distance = closed_manifold ? 128 : 16;
+  dialog::list_skip = 0;
+  scrolling_distances = false;
   viewdists = true;
   if(!mod_allowed()) {
     number_coding = ncDistance;
@@ -814,10 +832,11 @@ EX void enable_viewdists() {
 
 bool expansion_handleKey(int sym, int uni) {
   if((cmode & sm::NORMAL) && viewdists) {
+    dialog::handleNavigation(sym, uni);
     if(uni == 'S' && (cmode & sm::EXPANSION)) scrolling_distances = !scrolling_distances;
     else if(uni == 'C') pushScreen(viewdist_configure_dialog);
     else if(uni == 'A' && (cmode & sm::EXPANSION)) use_analyzer = !use_analyzer;
-    else if(sym == SDLK_ESCAPE) first_distance = 0, viewdists = false;
+    else if(sym == SDLK_ESCAPE) dialog::list_skip = 0, viewdists = false;
     else return false;
     return true;
     }
@@ -1060,7 +1079,7 @@ EX int hyperbolic_celldistance(cell *c1, cell *c2) {
       else {
         if(cl1 == cr2 || cr1 == cl2) found_distance = d;
         }
-      }    
+      }
     
     if(d >= found_distance) {
       if(sl_used == sibling_limit && IRREGULAR) { 
@@ -1082,4 +1101,4 @@ EX int hyperbolic_celldistance(cell *c1, cell *c2) {
     }
   }
 
-}
\ No newline at end of file
+}
