@@ -32,6 +32,7 @@ EX namespace fake {
     if(in()) return true;
     if(WDIM == 2 && standard_tiling() && (PURE || BITRUNCATED)) return true;
     if(arcm::in() && PURE) return true;
+    if(hat::in()) return true;
     if(WDIM == 2) return false;
     if(among(geometry, gBitrunc3)) return false;
     #if MAXMDIM >= 4
@@ -89,7 +90,7 @@ EX namespace fake {
 
     hyperpoint get_corner(cell *c, int cid, ld cf=3) override { 
 
-      if(arcm::in()) {
+      if(arcm::in() || hat::in()) {
         return underlying_map->get_corner(c, cid, cf);
         }
 
@@ -99,6 +100,7 @@ EX namespace fake {
       }
 
     transmatrix adj(cell *c, int d) override {
+      if(hat::in()) return underlying_map->adj(c, d);
       if(variation == eVariation::coxeter) {
         array<int, 3> which;
         in_underlying([&which, c, d] {
@@ -237,19 +239,17 @@ EX namespace fake {
         auto h1 = V * befake(FPIU(get_corner_position(c, (i+1) % c->type)));
         ld b0 = atan2(unshift(h0));
         ld b1 = atan2(unshift(h1));
-        while(b1 < b0) b1 += 2 * M_PI;
+        while(b1 < b0) b1 += TAU;
         if(a0 == -1) {
           draw_recursive(c->move(i), optimized_shift(V * adj(c, i)), b0, b1, c, depth+1);
           }
         else {
           if(b1 - b0 > M_PI) continue;
           
-          if(b0 < a0 - M_PI) b0 += 2 * M_PI;
-          if(b0 > a0 + M_PI) b0 -= 2 * M_PI;
+          cyclefix(b0, a0);
           if(b0 < a0) b0 = a0;
           
-          if(b1 > a1 + M_PI) b1 -= 2 * M_PI;
-          if(b1 < a1 - M_PI) b1 += 2 * M_PI;
+          cyclefix(b1, a1);
           if(b1 > a1) b1 = a1;
           
           if(b0 > b1) continue;
@@ -537,8 +537,10 @@ EX ld compute_euclidean() {
   #if CAP_ARCM
   if(arcm::in()) return arcm::current.N * 2 / arcm::current.euclidean_angle_sum;
   #endif
+  if(underlying == gAperiodicHat) return 6;
   if(WDIM == 2) return 4 / (S7-2.) + 2;
 
+
   if(underlying == gRhombic3) return 3;
   if(underlying == gBitrunc3) return 2.55208;
   int middle = get_middle();
@@ -553,6 +555,7 @@ EX ld around_orig() {
   if(arcm::in())
     return arcm::current.N;
   #endif
+  if(hat::in()) return 6;
   if(WDIM == 2)
     return S3;
   if(underlying == gRhombic3)
@@ -657,8 +660,9 @@ void set_gfake(ld _around) {
   compute_scale();
   check_cgi();
   cgi.require_basics();
-  
+
   if(currentmap) new hrmap_fake(currentmap);
+  if(hat::in()) hat::reshape();
   }
 
 EX void change_around() {
@@ -682,7 +686,7 @@ EX void change_around() {
 
     /* to compute scale */
     if(WDIM == 2)
-      cgi.prepare_basics();
+      cgi.require_basics();
     }
 
   t = scale / t;
