@@ -26,9 +26,16 @@ EX vector<basic_textureinfo> floor_texture_vertices;
 EX vector<glvertex> floor_texture_map;
 EX struct renderbuffer *floor_textures;
 
+EX basic_textureinfo* get_floor_texture_vertices(int index) {
+  if(noGUI || !vid.usingGL) return nullptr;
+  return &floor_texture_vertices[index];
+  }
+
 /* 0: generate no floorshapes; 1: generate only plain floorshapes; 2: generate all */
 EX int floorshapes_level = 2;
 
+EX ld global_boundary_ratio = 1;
+
 void geometry_information::init_floorshapes() {
   if(floorshapes_level == 0) return;
   all_escher_floorshapes.clear();
@@ -104,17 +111,17 @@ mesher msh(eGeometry g, int sym, ld main, ld v0, ld v1, ld bspi, ld scale) {
   dynamicval<eGeometry> dg(geometry, g);
 
   hyperpoint rot = xpush(v0) * xspinpush0(M_PI - M_PI/sym, main);
-  hyperpoint bnlfar = xpush(v0) * spin(M_PI) * rspintox(rot) * rspintox(rot) * rspintox(rot) * xpush0(hdist0(rot));
-  hyperpoint bnrfar = xpush(v0) * spin(M_PI) * spintox(rot) * spintox(rot) * spintox(rot) * xpush0(hdist0(rot));
+  hyperpoint bnlfar = xpush(v0) * spin180() * rspintox(rot) * rspintox(rot) * rspintox(rot) * xpush0(hdist0(rot));
+  hyperpoint bnrfar = xpush(v0) * spin180() * spintox(rot) * spintox(rot) * spintox(rot) * xpush0(hdist0(rot));
 
-  m.lcorner = xspinpush0 (bspi-M_PI/sym, main);
-  m.rcorner = xspinpush0 (bspi+M_PI/sym, main);
+  m.lcorner = xspinpush0 (bspi - M_PI/sym, main);
+  m.rcorner = xspinpush0 (bspi + M_PI/sym, main);
   m.mfar[0] = xspinpush0 (bspi, v0);
   m.mfar[1] = xspinpush0 (bspi, v1);
   m.vfar[0] = spin(bspi) * bnlfar;
   m.vfar[2] = spin(bspi) * bnrfar;
-  m.vfar[1] = spin(-2*M_PI/sym) * m.vfar[2];
-  m.vfar[3] = spin(+2*M_PI/sym) * m.vfar[0];
+  m.vfar[1] = spin(-TAU/sym) * m.vfar[2];
+  m.vfar[3] = spin(+TAU/sym) * m.vfar[0];
   
   return m;
   }
@@ -129,17 +136,12 @@ matrixitem genitem(const transmatrix& m1, const transmatrix& m2, int nsym) {
   mi.first = m1;
   mi.second.resize(nsym);
   for(int i=0; i<nsym; i++)
-    mi.second[i] = spin(2*M_PI*i/nsym) * m2;
+    mi.second[i] = spin(TAU*i/nsym) * m2;
   return mi;
   }
 
 bool do_kleinize() { return S3 >= OINF || (cgflags & qIDEAL); }
 
-EX hyperpoint kleinize(hyperpoint h) { 
-  if(GDIM == 2) return point3(h[0]/h[2], h[1]/h[2], 1); 
-  else return point31(h[0]/h[3], h[1]/h[3], h[2]/h[3]);
-  }
-
 EX hyperpoint may_kleinize(hyperpoint h) { 
   if(do_kleinize()) return kleinize(h);
   else return h;
@@ -147,7 +149,7 @@ EX hyperpoint may_kleinize(hyperpoint h) {
 
 void addmatrix(matrixlist& matrices, hyperpoint o0, hyperpoint o1, hyperpoint o2, hyperpoint n0, hyperpoint n1, hyperpoint n2, int d, int osym, int nsym) {
   if(do_kleinize()) o0 = kleinize(o0), o1 = kleinize(o1), o2 = kleinize(o2), n0 = kleinize(n0), n1 = kleinize(n1), n2 = kleinize(n2);
-  matrices.v.push_back(genitem(inverse(spin(2*M_PI*d/osym)*build_matrix(o0, o1, o2,C02)), spin(2*M_PI*d/nsym)*build_matrix(n0, n1, n2,C02), nsym));
+  matrices.v.push_back(genitem(inverse(spin(TAU*d/osym)*build_matrix(o0, o1, o2,C02)), spin(TAU*d/nsym)*build_matrix(n0, n1, n2,C02), nsym));
   }
 
 matrixlist hex_matrices, hept_matrices;
@@ -223,20 +225,20 @@ void geometry_information::bshape2(hpcshape& sh, PPR prio, int shapeid, matrixli
   hyperpoint lstmid = hpxyz(0,0,0);
   for(auto pp: lst) lstmid += pp;
   transmatrix T = spin(-m.o.bspi);
-  while((spin(2*M_PI / rots) * T* lstmid)[0] < (T*lstmid)[0])
-    T = spin(2*M_PI / rots) * T;
-  while((spin(-2*M_PI / rots) * T* lstmid)[0] < (T*lstmid)[0])
-    T = spin(-2*M_PI / rots) * T;
+  while((spin(TAU / rots) * T* lstmid)[0] < (T*lstmid)[0])
+    T = spin(TAU / rots) * T;
+  while((spin(-TAU / rots) * T* lstmid)[0] < (T*lstmid)[0])
+    T = spin(-TAU / rots) * T;
   T = spin(m.o.bspi) * T;
   for(auto &pp: lst) pp = T * pp;
   
-  if(osym % rots && rots % osym) printf("warning: rotation oddity (shapeid %d, osym=%d rots=%d)\n", shapeid, osym, rots);
+  if(osym % rots && rots % osym && (debugflags & DF_GEOM)) printf("warning: rotation oddity (shapeid %d, osym=%d rots=%d)\n", shapeid, osym, rots);
 
   if(rots > osym && rots % osym == 0) {
     int rep = rots / osym;
     int s = lst.size();
     for(int i=0; i<s*(rep-1); i++)
-      lst.push_back(spin(2*M_PI/rots) * lst[i]);
+      lst.push_back(spin(TAU/rots) * lst[i]);
     rots /= rep;
     }
   
@@ -258,7 +260,7 @@ void geometry_information::bshape2(hpcshape& sh, PPR prio, int shapeid, matrixli
           nh = m.second[r] * z, mapped++;
           }
         }
-      if(mapped == 0) printf("warning: not mapped (shapeid %d)\n", shapeid);
+      if(mapped == 0 && (debugflags & DF_GEOM)) printf("warning: not mapped (shapeid %d)\n", shapeid);
       if(invalid) {
         apeirogonal = true;        
         for(auto h: head) tail.push_back(h);
@@ -277,6 +279,10 @@ template<class T> void sizeto(T& t, int n) {
   if(isize(t) <= n) t.resize(n+1);
   }
 
+template<class T, class U> void sizeto(T& t, int n, const U& val) {
+  if(isize(t) <= n) t.resize(n+1, val);
+  }
+
 void geometry_information::bshape_regular(floorshape &fsh, int id, int sides, ld shift, ld size, cell *c) {
   
   sizeto(fsh.b, id);
@@ -286,20 +292,23 @@ void geometry_information::bshape_regular(floorshape &fsh, int id, int sides, ld
   if(bt::in()) {
 
     const int STEP = vid.texture_step;
-    
+
     for(int t=0; t<2; t++) {
     
       if(t == 0) 
         bshape(fsh.b[id], fsh.prio);
       if(t == 1)
         bshape(fsh.shadow[id], fsh.prio);
+
+      int STEP1 = STEP;
+      if((embedded_plane || geom3::flipped) && t == 0) STEP1 = 1;
       
       for(int i=0; i<sides; i++) {
         hyperpoint h0 = bt::get_corner_horo_coordinates(c, i) * size;
         hyperpoint h1 = bt::get_corner_horo_coordinates(c, i+1) * size;
         if(t) h0 *= SHADMUL, h1 *= SHADMUL;
-        hyperpoint hd = (h1 - h0) / STEP;
-        for(int j=0; j<STEP; j++) {
+        hyperpoint hd = (h1 - h0) / STEP1;
+        for(int j=0; j<STEP1; j++) {
           hpcpush(bt::get_horopoint(h0 + hd * j));
           if(geometry == gBinary4 && among(i, 2, 4)) break;
           if(geometry == gBinaryTiling && among(i, 0, 4)) break;
@@ -331,11 +340,11 @@ void geometry_information::bshape_regular(floorshape &fsh, int id, int sides, ld
   
   bshape(fsh.b[id], fsh.prio);
   for(int t=0; t<=sides; t++)
-    hpcpush(xspinpush0(t*2 * M_PI / sides + shift * M_PI / S42, size));
+    hpcpush(xspinpush0(t * TAU / sides + shift * S_step, size));
 
   bshape(fsh.shadow[id], fsh.prio);
   for(int t=0; t<=sides; t++)
-    hpcpush(xspinpush0(t*2 * M_PI / sides + shift * M_PI / S42, size * SHADMUL));
+    hpcpush(xspinpush0(t * TAU / sides + shift * S_step, size * SHADMUL));
 
   for(int k=0; k<SIDEPARS; k++) {
     fsh.side[k].resize(2);
@@ -343,6 +352,17 @@ void geometry_information::bshape_regular(floorshape &fsh, int id, int sides, ld
     hpcpush(xspinpush0(+M_PI/sides, size));
     hpcpush(xspinpush0(-M_PI/sides, size));
     chasmifyPoly(dlow_table[k], dhi_table[k], k);
+
+    if(cgi.emb->is_euc_in_noniso()) {
+      fsh.gpside[k].resize(c->type);
+      for(int i=0; i<c->type; i++) {
+        sizeto(fsh.gpside[k][i], id);
+        bshape(fsh.gpside[k][i][id], PPR::LAKEWALL);
+        hpcpush(xspinpush0(M_PI - i * TAU / sides + shift * S_step, size));
+        hpcpush(xspinpush0(M_PI - (i + 1) * TAU / sides + shift * S_step, size));
+        chasmifyPoly(dlow_table[k], dhi_table[k], k);
+        }
+      }
     }
   }
 
@@ -366,6 +386,24 @@ void geometry_information::finish_apeirogon(hyperpoint center) {
     }
   }
 
+hyperpoint get_circumscribed_corner(cell *c, int t, hyperpoint h) {
+
+  hyperpoint h0 = currentmap->adjmod(c, t) * h;
+  hyperpoint h1 = currentmap->adjmod(c, t-1) * h;
+  transmatrix T;
+  array<hyperpoint, 3> hs = {h, h0, h1};
+  set_column(T, 3, C03);
+  hyperpoint res = C03;
+  for(int i=0; i<3; i++) {
+    hyperpoint ahs = hs[i];
+    if(hyperbolic) ahs[3] *= -1;
+    set_column(T, i, ahs);
+    res[i] = dot_d(4, hs[i], ahs);
+    }
+  T = transpose(T);
+  return inverse(T) * res;
+  }
+
 // !siid equals pseudohept(c)
 
 void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, int sidir) {
@@ -421,26 +459,28 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
     if(&fsh == &shTriheptaFloor) {
       if(!siid) {
         for(int i=0; i<cor; i++)
-          cornerlist.push_back(midcorner(c, i, .49));
+          cornerlist.push_back(midcorner(c, i, .5 - .01 * global_boundary_ratio));
         }
       else {
         for(int i=0; i<cor; i++) {
           int ri = i;
           if((i&1) == ((sidir+siid)&1)) ri--;
           ri = c->c.fix(ri);
-          cornerlist.push_back(mid(get_corner_position(c, ri, 3.1), get_corner_position(c, c->c.fix(ri+1), 3.1)));
+          ld val = 3 + 0.1 * global_boundary_ratio;
+          cornerlist.push_back(mid(get_corner_position(c, ri, val), get_corner_position(c, c->c.fix(ri+1), val)));
           }
         }
       }
     
     else if(&fsh == &shBigTriangle) {
+      ld val = 1 - 0.06 * global_boundary_ratio;
       if(!siid) {
         for(int i=0; i<cor; i++) cornerlist.push_back(hpxy(0,0));
         }
       else if(geosupport_chessboard()) {
         for(int i=0; i<cor; i++) {
           hyperpoint nc = nearcorner(c, i);
-          cornerlist.push_back(mid_at(hpxy(0,0), nc, .94));
+          cornerlist.push_back(mid_at(hpxy(0,0), nc, val));
           }
         }
       else {
@@ -449,16 +489,17 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
           if((i&1) != ((sidir+siid)&1)) ri--;
           ri = c->c.fix(ri);
           hyperpoint nc = nearcorner(c, ri);
-          cornerlist.push_back(mid_at(hpxy(0,0), nc, .94));
+          cornerlist.push_back(mid_at(hpxy(0,0), nc, val));
           }
         }
       }
   
     else if(&fsh == &shBigHepta) {
+      ld val = 1 - 0.06 * global_boundary_ratio;
       if(!siid) {
         for(int i=0; i<cor; i++) {
           hyperpoint nc = nearcorner(c, i);
-          cornerlist.push_back(mid_at(hpxy(0,0), nc, .94));
+          cornerlist.push_back(mid_at(hpxy(0,0), nc, val));
           }
         }
       else {
@@ -466,7 +507,7 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         }
       }
     
-    else if(arb::in()) {
+    else if(arb::in() || aperiodic) {
       vector<hyperpoint> actual;
       for(int j=0; j<cor; j++) 
         actual.push_back(get_corner_position(c, j));
@@ -480,8 +521,8 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         }
       
       auto &ac = arb::current_or_slided();
-      ld dist = min_dist * (1 - 3 / sca) * ac.boundary_ratio;
-      
+
+      ld dist = min_dist * (1 - 3 / sca) * (arb::in() ? ac.boundary_ratio : 1);
       ld area = 0;
       for(int j=0; j<cor; j++) {
         hyperpoint current = kleinize(actual[j]);
@@ -490,9 +531,14 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         }
       if(area < 0) dist = -dist;
 
-      int id = arb::id_of(c->master);
-      auto& sh = ac.shapes[id];
-      apeirogonal = sh.apeirogonal;
+      apeirogonal = false;
+      int id = 0;
+      arb::shape *sh = nullptr;
+      if(arb::in()) {
+        int id = arb::id_of(c->master);
+        sh = &ac.shapes[id];
+        apeirogonal = sh->apeirogonal;
+        }
       
       for(int j=0; j<cor; j++) {
         hyperpoint last = actual[j?j-1:cor-1];
@@ -500,9 +546,9 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         hyperpoint next = actual[j<cor-1?j+1:0];
 
         if(apeirogonal) {
-          if(j == 0) last = arb::get_adj(arb::current_or_slided(), id, cor-1, id, cor-2) * actual[cor-3];
-          if(j == cor-2) next = arb::get_adj(arb::current_or_slided(), id, cor-2, id, cor-1) * actual[1];
-          if(j == cor-1) { cornerlist.push_back(sh.vertices.back()); continue; }
+          if(j == 0) last = arb::get_adj(arb::current_or_slided(), id, cor-1, id, cor-2, false) * actual[cor-3];
+          if(j == cor-2) next = arb::get_adj(arb::current_or_slided(), id, cor-2, id, cor-1, false) * actual[1];
+          if(j == cor-1) { cornerlist.push_back(sh->vertices.back()); continue; }
           }
 
         auto T = gpushxto0(current);
@@ -640,7 +686,7 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
           int id = arb::id_of(c->master);
           auto &ac = arb::current_or_slided();
           auto& sh = ac.shapes[id];
-          hpcpush(arb::get_adj(arb::current_or_slided(), id, cor-2, id, cor-1) * starting_point);
+          hpcpush(arb::get_adj(arb::current_or_slided(), id, cor-2, id, cor-1, false) * starting_point);
           finish_apeirogon(sh.vertices.back());
           }
         }
@@ -648,7 +694,7 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
     }
   
   #if MAXMDIM >= 4
-  if(WDIM == 2 && GDIM == 3) {
+  if(embedded_plane) {
     finishshape();
     for(auto pfsh: all_plain_floorshapes) {
       auto& fsh = *pfsh;
@@ -660,42 +706,57 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         sizeto(fsh.levels[k], id);
         bshape(fsh.levels[k][id], fsh.prio);    
         last->flags |= POLY_TRIANGLES;
-        last->tinf = &floor_texture_vertices[fsh.id];
+        last->tinf = get_floor_texture_vertices(fsh.id);
         last->texture_offset = 0;
 
-        #if CAP_BT
-        if(bt::in()) 
-          for(int t=0; t<c->type; t++)
-            texture_order([&] (ld x, ld y) {
-              hyperpoint left = bt::get_corner_horo_coordinates(c, t);
-              hyperpoint right = bt::get_corner_horo_coordinates(c, t+1);
-              hpcpush(orthogonal_move(bt::get_horopoint(left * x + right * y), dfloor_table[k]));
-              });
-        else 
-        #endif
         if(1) {
           int s = fsh.b[id].s;
           int e = fsh.b[id].e-1;
-          hyperpoint ctr = Hypc;
-          for(int t=0; t<e-s; t++)
-            ctr += orthogonal_move(may_kleinize(hpc[s+t]), dfloor_table[k]);
-          ctr = normalize(ctr);
-          if(vid.pseudohedral) for(int t=0; t<e-s; t++) {
-            hyperpoint v1 = orthogonal_move(may_kleinize(hpc[s+t]), dfloor_table[k]) - ctr;
-            hyperpoint v2 = orthogonal_move(may_kleinize(hpc[s+t+1]), dfloor_table[k]) - ctr;
-            texture_order([&] (ld x, ld y) {
-              hpcpush(normalize(ctr + v1 * x + v2 * y));
-              });
+
+          if(vid.pseudohedral == phInscribed) {
+            hyperpoint ctr = Hypc;
+            for(int t=0; t<e-s; t++)
+              ctr += kleinize(cgi.emb->orthogonal_move(hpc[s+t], dfloor_table[k]));
+            ctr = normalize(ctr);
+
+            for(int t=0; t<e-s; t++) {
+              hyperpoint v1 = kleinize(cgi.emb->orthogonal_move(hpc[s+t], dfloor_table[k])) - ctr;
+              hyperpoint v2 = kleinize(cgi.emb->orthogonal_move(hpc[s+t+1], dfloor_table[k])) - ctr;
+              texture_order([&] (ld x, ld y) {
+                hpcpush(normalize(ctr + v1 * x + v2 * y));
+                });
+              }
             }
-          if(!vid.pseudohedral) for(int t=0; t<e-s; t++) {
-            hyperpoint v1 = may_kleinize(hpc[s+t]) - C0;
-            hyperpoint v2 = may_kleinize(hpc[s+t+1]) - C0;
+          if(vid.pseudohedral == phCircumscribed) {
+
+            vector<hyperpoint> hs(c->type);
+            hyperpoint z = Hypc; z[2] = dfloor_table[k];
+            hyperpoint ctr = cgi.emb->logical_to_actual(z);
+            for(int t=0; t<c->type; t++) hs[t] = get_circumscribed_corner(c, t, ctr);
+            // for(int t=0; t<c->type; t++) hs[t] = xspinpush0(t * TAU / c->type, 0.2); //  kleinize(get_circumscribed_corner(c, t, ctr));
+
+            for(int t=0; t<c->type; t++) {
+              hyperpoint v1 = hs[t] - ctr;
+              hyperpoint v2 = atmod(hs, t+1) - ctr;
+              texture_order([&] (ld x, ld y) {
+                hpcpush(normalize(ctr + v1 * x + v2 * y));
+                });
+              }
+            /* also affect the plain floor */
+            bshape(fsh.b[id], PPR::FLOOR);
+            for(hyperpoint& h: hs) hpcpush(h);
+            hpcpush(hs[0]);
+            }
+          if(vid.pseudohedral == phOFF) for(int t=0; t<e-s; t++) {
+
+            hyperpoint v1 = cgi.emb->actual_to_logical(hpc[s+t]);
+            hyperpoint v2 = cgi.emb->actual_to_logical(hpc[s+t+1]);
+
             texture_order([&] (ld x, ld y) { 
-              hpcpush(
-                orthogonal_move(
-                  normalize(C0 + v1 * x + v2 * y)
-                  , dfloor_table[k])
-                ); 
+              hyperpoint a = v1 * x + v2 * y;
+              a[2] = dfloor_table[k];
+              auto c = cgi.emb->logical_to_actual(a);
+              cgi.hpcpush(c);
               });
             }
           }
@@ -708,27 +769,20 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
         sizeto(fsh.cone[co], id);
         bshape(fsh.cone[co][id], fsh.prio);    
         last->flags |= POLY_TRIANGLES;
-        last->tinf = &floor_texture_vertices[fsh.id];
+        last->tinf = get_floor_texture_vertices(fsh.id);
         last->texture_offset = 0;
         ld h = (FLOOR - WALL) / (co+1);
         ld top = co ? (FLOOR + WALL) / 2 : WALL;
-        #if CAP_BT
-        if(bt::in())
-          for(int t=0; t<c->type; t++)
-            texture_order([&] (ld x, ld y) {
-              hyperpoint left = bt::get_corner_horo_coordinates(c, t);
-              hyperpoint right = bt::get_corner_horo_coordinates(c, t+1);
-              hpcpush(orthogonal_move(bt::get_horopoint(left * x + right * y), top + h * (x+y)));
-              });
-        else
-        #endif
         if(1) {
           int s = fsh.b[id].s;
           int e = fsh.b[id].e-1;        
           for(int t=0; t<e-s; t++) {
-            hyperpoint v1 = may_kleinize(hpc[s+t]) - C0;
-            hyperpoint v2 = may_kleinize(hpc[s+t+1]) - C0;
-            texture_order([&] (ld x, ld y) { hpcpush(orthogonal_move(normalize(C0 + v1 * x + v2 * y), top + h * (x+y))); });
+            hyperpoint v1 = cgi.emb->actual_to_logical(hpc[s+t]);
+            hyperpoint v2 = cgi.emb->actual_to_logical(hpc[s+t+1]);
+            texture_order([&] (ld x, ld y) {
+              hyperpoint a = v1 * x + v2 * y; a[2] = top + h * (x+y);
+              hpcpush(cgi.emb->logical_to_actual(a));
+              });
             }
           }
 
@@ -776,6 +830,8 @@ void geometry_information::generate_floorshapes_for(int id, cell *c, int siid, i
   #endif
   }
 
+EX hookset<bool(cell*)> hooks_floorshapes;
+
 void geometry_information::generate_floorshapes() {
 
   DEBBI(DF_POLY, ("generate_floorshapes"));
@@ -784,19 +840,23 @@ void geometry_information::generate_floorshapes() {
   cell model;
   model.master = &modelh;
   modelh.c7 = &model;
-  model.type = modelh.type = S7;
+  model.type = modelh.type = FULL_EDGE;
   
   auto mmerge1 = [&] (int i, int j) { model.c.setspin(i, j, false); modelh.c.setspin(i, j, false); };  
   auto mmerge = [&] (int i, int j) { mmerge1(i, j); mmerge1(j, i); };  
 
-  for(int i=0; i<S7; i++) {
+  for(int i=0; i<FULL_EDGE; i++) {
     model.move(i) = &model;
     modelh.move(i) = &modelh;
     model.c.setspin(i, i, false);
     modelh.c.setspin(i, i, false);
     }
 
-  if(WDIM == 3) ;
+  model.type = modelh.type = S7;
+
+  if(callhandlers(false, hooks_floorshapes, &model)) ;
+
+  else if(WDIM == 3) ;
   
   #if CAP_IRR
   else if(IRREGULAR) {
@@ -822,6 +882,14 @@ void geometry_information::generate_floorshapes() {
 
   else if(inforder::mixed()) { /* will be generated on the fly */ }
   
+  else if(hat::in()) {
+    dynamicval<bool> ncor(approx_nearcorner, true);
+    for(int i=0; i<2; i++) {
+      modelh.c7 = i == 1 ? &model : nullptr;
+      generate_floorshapes_for(i, &model, 0, 0);
+      }
+    }
+
   #if CAP_BT
   else if(kite::in()) {
     dynamicval<bool> ncor(approx_nearcorner, true);
@@ -836,21 +904,7 @@ void geometry_information::generate_floorshapes() {
   
   #if CAP_ARCM
   else if(arcm::in()) {
-    arcm::parent_index_of(&modelh) = 0;
-    auto &ac = arcm::current;
-    for(int i=0; i<2*ac.N + 2; i++) {
-      if(ac.regular && i>=2 && i < 2*ac.N) continue;
-      arcm::id_of(&modelh) = i;
-      model.type = isize(ac.triangles[i]);
-      if(DUAL) model.type /= 2, arcm::parent_index_of(&modelh) = !(i&1);
-      
-      if(BITRUNCATED)
-        generate_floorshapes_for(i, &model, !arcm::pseudohept(&model), arcm::pseudohept(&model) ? 0 : 1^(i&1));
-      else if(geosupport_football() == 2)
-        generate_floorshapes_for(i, &model, !arcm::pseudohept(&model), i >= 4 ? 1 : 0);
-      else
-        generate_floorshapes_for(i, &model, 0, 0);
-      }
+    /* will be generated on the fly */
     }
   #endif
   
@@ -885,7 +939,7 @@ void geometry_information::generate_floorshapes() {
         ms.c.connect(j, &models[co.sid], co.eid, co.mirror);
         }
       }
-    for(int i=0; i<n; i++) generate_floorshapes_for(i, &models[i], 0, 0);
+    for(int i=0; i<n; i++) generate_floorshapes_for(i, &models[i], c.shapes[i].football_type < 2, c.shapes[i].football_type == 0);
     }
   
   else if(geometry == gBinary4) {
@@ -1033,6 +1087,22 @@ EX void set_floor(const transmatrix& spin, hpcshape& sh) {
   qfi.usershape = -1;
   }
 
+/** currently only for arcm */
+EX void ensure_floorshape_generated(int id, cell *c) {
+  hpcshape nul; nul.s = -1;
+  sizeto(cgi.shFloor.b, id, nul);
+  if(cgi.shFloor.b[id].s == -1) {
+    if(BITRUNCATED)
+      cgi.generate_floorshapes_for(id, c, !arcm::pseudohept(c), arcm::pseudohept(c) ? 0 : 1^(id&1));
+    else if(geosupport_football() == 2)
+      cgi.generate_floorshapes_for(id, c, !arcm::pseudohept(c), id >= 4 ? 1 : 0);
+    else
+      cgi.generate_floorshapes_for(id, c, 0, 0);
+    cgi.finishshape();
+    cgi.extra_vertices();
+    }
+  }
+
 EX int shvid(cell *c) {
   return currentmap->shvid(c);
   }
@@ -1068,7 +1138,6 @@ int hrmap_standard::shvid(cell *c) {
       cgi.crossf = cgi.tessf;
       
       cgi.require_shapes();
-      println(hlog, "generating floorshapes for ", t);
       cgi.generate_floorshapes_for(t, &model, 0, 0);
       cgi.finishshape();
       cgi.extra_vertices();
@@ -1105,12 +1174,14 @@ EX struct dqi_poly *draw_shapevec(cell *c, const shiftmatrix& V, const vector<hp
   #endif
   #if CAP_ARCM
   else if(arcm::in()) {
-    return &queuepolyat(V, shv[shvid(c)], col, prio);
+    int id = shvid(c);
+    ensure_floorshape_generated(id, c);
+    return &queuepolyat(V, shv[id], col, prio);
     }
   #endif
   else if(GOLDBERG && ishex1(c)) 
     return &queuepolyat(V * pispin, shv[0], col, prio);
-  else if(!(S7&1) && PURE && !kite::in() && !a4) {
+  else if(!(S7&1) && PURE && !aperiodic && !a4) {
     auto si = patterns::getpatterninfo(c, patterns::PAT_COLORING, 0);
     if(si.id == 8) si.dir++;
     transmatrix D = applyPatterndir(c, si);
@@ -1213,7 +1284,7 @@ void draw_shape_for_texture(floorshape* sh) {
     queuepoly(shiftless(eupush(gx+a, gy+b)), sh->b[0], 0xFFFFFFFF);
 
   if(sh == &cgi.shCrossFloor) {
-    queuepoly(shiftless(eupush(gx, gy) * spin(M_PI/4)), cgi.shCross, 0x808080FF);
+    queuepoly(shiftless(eupush(gx, gy) * spin(45._deg)), cgi.shCross, 0x808080FF);
     }
 
   if(1) {
@@ -1232,15 +1303,17 @@ void draw_shape_for_texture(floorshape* sh) {
     ld d = hdist(h1, h2);
     hyperpoint h3 = h1 + (h2-h1) /d * min(d, .1);
     for(int a=0; a<4; a++) {
-      curvepoint(eupush(gx,gy) * eupush(spin(90*degree*a) * h1) * C0);
-      curvepoint(eupush(gx,gy) * eupush(spin(90*degree*a) * h3) * C0);
+      curvepoint(eupush(gx,gy) * eupush(spin(90._deg*a) * h1) * C0);
+      curvepoint(eupush(gx,gy) * eupush(spin(90._deg*a) * h3) * C0);
       queuecurve(shiftless(Id), 0x10101010, 0, PPR::LINE);
       }
     }
   
-  auto& ftv = floor_texture_vertices[sh->id];
-  ftv.tvertices.clear();
-  ftv.texture_id = floor_textures->renderedTexture;
+  auto ftv = get_floor_texture_vertices(sh->id);
+  if(ftv) {
+    ftv->tvertices.clear();
+    ftv->texture_id = floor_textures->renderedTexture;
+    }
   
   hyperpoint center = eupush(gx, gy) * C0;
   hyperpoint v1 = hpxyz3(sd, sd, 0, 0);
@@ -1268,9 +1341,9 @@ void draw_shape_for_texture(floorshape* sh) {
     };
   
   // SL2 needs 6 times more
-  texture_order([&] (ld x, ld y) {
+  if(ftv) texture_order([&] (ld x, ld y) {
     auto v = tvec_at(x, y);
-    ftv.tvertices.push_back(glhr::makevertex(v[0], v[1], 0));
+    ftv->tvertices.push_back(glhr::makevertex(v[0], v[1], 0));
     });
   
   floor_texture_square_size = 2 * (tvec_at(1, 0)[0] - tvec_at(0, 0)[0]);
@@ -1287,11 +1360,12 @@ EX void ensure_vertex_number(basic_textureinfo& bti, int qty) {
 
 /** ensure_vertex_number for a hpcshape */
 EX void ensure_vertex_number(hpcshape& sh) {
+  if(!sh.tinf) return;
   ensure_vertex_number(*sh.tinf, sh.e - sh.s);
   }
 
 EX void bind_floor_texture(hpcshape& li, int id) {
-  li.tinf = &floor_texture_vertices[id];
+  li.tinf = get_floor_texture_vertices(id);
   ensure_vertex_number(li);
   }
 
@@ -1391,7 +1465,9 @@ EX void make_floor_textures() {
   dynamicval<eModel> gm(pmodel, mdDisk);
   dynamicval<eVariation> va(variation, eVariation::pure);
   dynamicval<geometryinfo1> gie(ginf[geometry].g, giEuclid2);
+  dynamicval<flagtype> gief(ginf[geometry].flags, qOPTQ);
   dynamicval<geometryinfo1> gih(ginf[gNormal].g, giHyperb2);
+  dynamicval<flagtype> gihf(ginf[gNormal].flags, 0);
   dynamicval<bool> a3(vid.always3, false);
   dynamicval<bool> hq(inHighQual, true);
   dynamicval<int> hd(darken, 0);
