@@ -40,7 +40,8 @@ string euchelp =
 #if CAP_FIELD
 void showQuotientConfig() {
   using namespace fieldpattern;
-  gamescreen(2);
+  cmode = sm::SIDE | sm::MAYDARK;
+  gamescreen();
   dialog::init(XLAT("field quotient"));
   fgeomextra& gxcur = fgeomextras[current_extra];
   for(int i=0; i<isize(fgeomextras); i++) {
@@ -102,7 +103,8 @@ void showQuotientConfig() {
     auto& cfp = currfp;
     auto triplets = cfp.find_triplets();
     pushScreen([triplets] {
-      gamescreen(2);
+      cmode = sm::SIDE | sm::MAYDARK;
+      gamescreen();
       dialog::init(XLAT("alternate manifolds"));
       int id = 0;
       for(auto t: triplets) {
@@ -153,7 +155,7 @@ string validclasses[4] = {" (X)", " (½)", "", " (!)"};
   
 EX void ge_land_selection() {
   cmode = sm::SIDE | sm::MAYDARK;
-  gamescreen(0);  
+  gamescreen();
 
   if(cheater) for(int i=0; i<landtypes; i++) landvisited[i] = true;
 
@@ -185,17 +187,17 @@ EX void ge_land_selection() {
     });
   stable_sort(landlist.begin(), landlist.end(), [] (eLand l1, eLand l2) { return land_validity(l1).quality_level > land_validity(l2).quality_level; });
   
-  for(int i=0; i<euperpage; i++) {
-    if(euperpage * eupage + i >= isize(landlist)) { dialog::addBreak(100); break; }
-    eLand l = landlist[euperpage * eupage + i];
-    char ch = '1'+i;
+  dialog::start_list(900, 900, '1');
+  
+  for(auto& l: landlist) {
+  
     string s = XLAT1(linf[l].name);
 
     if(landvisited[l]) {
-      dialog::addBoolItem(s, l == specialland, ch);
+      dialog::addBoolItem(s, l == specialland, dialog::list_fake_key++);
       }
     else {
-      dialog::addSelItem(s, XLAT("(locked)"), ch);
+      dialog::addSelItem(s, XLAT("(locked)"), dialog::list_fake_key++);
       }
     
     dialog::lastItem().color = linf[l].color;
@@ -215,13 +217,13 @@ EX void ge_land_selection() {
         }));
       });
     }
-  dialog::addItem(XLAT("next page"), '-');
-  dialog::addInfo(XLAT("press letters to search"));
+  dialog::end_list();
 
   dialog::addBreak(25);
   validity_info();    
   dialog::addBreak(25);
   
+  dialog::addInfo(XLAT("press letters to search"));
   dual::add_choice();  
   dialog::addBack();
   dialog::display();
@@ -229,11 +231,7 @@ EX void ge_land_selection() {
   keyhandler = [] (int sym, int uni) {
     dialog::handleNavigation(sym, uni);
     
-    if(uni == '-' || uni == PSEUDOKEY_WHEELUP || uni == PSEUDOKEY_WHEELDOWN) {
-      eupage++;
-      if(eupage * euperpage >= isize(landlist)) eupage = 0;
-      }
-    else if(dialog::editInfix(uni)) eupage = 0;
+    if(dialog::editInfix(uni)) dialog::list_skip = 0;
     else if(doexiton(sym, uni)) popScreen();
     };
   }
@@ -247,6 +245,7 @@ EX void activate_ge_land_selection() {
 #if HDR
 struct geometry_filter {
   string name;
+  int hotkey;
   /** test if the current geometry matches the filter */
   function<bool()> test; 
   };
@@ -256,35 +255,36 @@ EX geometry_filter *current_filter;
 
 bool forced_quotient() { return quotient && !(cgflags & qOPTQ); }
 
-EX geometry_filter gf_hyperbolic = {"hyperbolic", [] { return (arcm::in() || arb::in() || hyperbolic) && !forced_quotient(); }};
-EX geometry_filter gf_spherical = {"spherical", [] { return (arcm::in() || arb::in() || sphere) && !forced_quotient(); }};
-EX geometry_filter gf_euclidean = {"Euclidean", [] { return (arcm::in() || arb::in() || euclid) && !forced_quotient(); }};
-EX geometry_filter gf_other = {"non-isotropic", [] { return prod || nonisotropic; }};
-EX geometry_filter gf_regular_2d = {"regular 2D tesselations", [] { 
+EX geometry_filter gf_hyperbolic = {"hyperbolic", 'h', [] { return (arcm::in() || arb::in() || hyperbolic) && !forced_quotient(); }};
+EX geometry_filter gf_spherical = {"spherical", 's', [] { return (arcm::in() || arb::in() || sphere) && !forced_quotient(); }};
+EX geometry_filter gf_euclidean = {"Euclidean", 'e', [] { return (arcm::in() || arb::in() || euclid) && !forced_quotient(); }};
+EX geometry_filter gf_other = {"non-isotropic", 'n', [] { return mproduct || nonisotropic; }};
+EX geometry_filter gf_regular_2d = {"regular 2D tesselations", 'r', [] { 
   return standard_tiling() && WDIM == 2 && !forced_quotient();
   }};
-EX geometry_filter gf_regular_3d = {"regular 3D honeycombs", [] { 
+EX geometry_filter gf_regular_3d = {"regular 3D honeycombs", '3', [] { 
   if(euclid) return geometry == gCubeTiling;
-  return !bt::in() && !kite::in() && WDIM == 3 && !forced_quotient() && !nonisotropic && !prod;
+  return !bt::in() && !kite::in() && WDIM == 3 && !forced_quotient() && !nonisotropic && !mproduct;
   }};
-EX geometry_filter gf_quotient = {"interesting quotient spaces", [] { 
+EX geometry_filter gf_quotient = {"interesting quotient spaces", 'q', [] { 
   return forced_quotient() && !elliptic;
   }};
+EX geometry_filter gf_tes_file = {"load from file", 'f', [] { return arb::in(); }};
+EX geometry_filter gf_no_filters = {"no filters", '-', [] { return true; }};
   
-EX vector<geometry_filter*> available_filters = { &gf_hyperbolic, &gf_spherical, &gf_euclidean, &gf_other, &gf_regular_2d, &gf_regular_3d, &gf_quotient };
+EX vector<geometry_filter*> available_filters = { &gf_hyperbolic, &gf_spherical, &gf_euclidean, &gf_other, &gf_regular_2d, &gf_regular_3d, &gf_quotient, &gf_tes_file, &gf_no_filters };
 
 void ge_select_filter() {
   cmode = sm::SIDE | sm::MAYDARK;
-  gamescreen(2);  
+  gamescreen();
 
   dialog::init(XLAT("geometries"));
   
-  char x = 'a';
   for(auto f: available_filters) {
     if(current_filter)
-      dialog::addBoolItem(XLAT(f->name), f == current_filter, x++);
+      dialog::addBoolItem(XLAT(f->name), f == current_filter, f->hotkey);
     else
-      dialog::addItem(XLAT(f->name), x++);
+      dialog::addItem(XLAT(f->name), f->hotkey);
     dialog::add_action([f] { current_filter = f; popScreen(); });
     }
 
@@ -322,7 +322,7 @@ void set_or_configure_geometry(eGeometry g) {
       if(g == gRotSpace) {
         bool ok = true;
         if(arcm::in()) ok = PURE;
-        else if(bt::in() || kite::in()) ok = false;
+        else if(bt::in() || aperiodic) ok = false;
         else ok = PURE || BITRUNCATED;
         if(!ok) {
           addMessage(XLAT("Only works with (semi-)regular tilings"));
@@ -346,9 +346,11 @@ void set_or_configure_geometry(eGeometry g) {
 
 /** is g2 the same tiling as the current geometry (geometry)? */
 bool same_tiling(eGeometry g2) {
+  /* no quotients for fractals */
+  if(cgflags & qFRACTAL) return g2 == geometry;
   if(g2 == gCrystal)
     return S3 == 4;
-  if(g2 == gFieldQuotient && hyperbolic && standard_tiling())
+  if(g2 == gFieldQuotient && (hyperbolic || (geometry == gCubeTiling && reg3::cubes_reg3)) && standard_tiling())
     return true;
   if(g2 == gFieldQuotient && geometry != gFieldQuotient) {
     int ce = 0;
@@ -364,8 +366,9 @@ bool same_tiling(eGeometry g2) {
   }
 
 void ge_select_tiling() {
+  if(current_filter == &gf_tes_file) { popScreen(); set_or_configure_geometry(gArbitrary); }
   cmode = sm::SIDE | sm::MAYDARK;
-  gamescreen(0);  
+  gamescreen();
 
   if(!current_filter) { popScreen(); return; }
   dialog::init();
@@ -376,7 +379,8 @@ void ge_select_tiling() {
 
   dialog::addBreak(100);
   
-  char letter = 'a';
+  dialog::start_list(500, 1500, 'a');
+  
   for(int i=0; i<isize(ginf); i++) {
     eGeometry g = eGeometry(i);
     if(among(g, gProduct, gRotSpace)) hybrid::configure(g);
@@ -404,11 +408,12 @@ void ge_select_tiling() {
     dialog::addBoolItem(
       is_product ? XLAT("current geometry x E") : 
       is_rotspace ? XLAT("space of rotations in current geometry") : 
-      XLAT(ginf[g].menu_displayed_name), on, letter++);
+      XLAT(ginf[g].menu_displayed_name), on, dialog::list_fake_key++);
     dialog::lastItem().value += validclasses[land_validity(specialland).quality_level];
     dialog::add_action([g] { set_or_configure_geometry(g); });
     }
   
+  dialog::end_list();
   dialog::addBreak(100);
   dual::add_choice();  
   dialog::addBack();
@@ -447,7 +452,8 @@ EX string dim_name() {
 EX void showQuotientConfig3() {
 
   using namespace fieldpattern;
-  gamescreen(2);
+  cmode = sm::SIDE | sm::MAYDARK;
+  gamescreen();
   dialog::init(XLAT("field quotient"));
   
   auto& ds = discoveries[cginf.tiling_name];
@@ -503,34 +509,27 @@ EX void showQuotientConfig3() {
   }
 #endif
 
-EX string geometry_name() {
-  switch(ginf[geometry].cclass) {
+EX string geometry_name(eGeometryClass gc) {
+  switch(gc) {
     case gcHyperbolic:
-      return XLAT("hyperbolic") + dim_name();
+      return XLAT("hyperbolic");
 
     case gcEuclid: 
       if(cgflags & qAFFINE)
-        return XLAT("affine") + dim_name();
-      return XLAT("flat") + dim_name();
+        return XLAT("affine");
+      return XLAT("flat");
     
     case gcSphere:
-      return XLAT("spherical") + dim_name();
+      return XLAT("spherical");
 
-    case gcSolNIH:
-#if CAP_SOLV
-      switch(sn::geom()) {
-        case gSol:
-          return XLAT("Sol");
-        case gNIH:
-          return XLAT("hyperbolic (3:2)");
-        case gSolN:
-          return XLAT("Sol (3:2)");
-        default:
-          return "unknown";
-        }
-#else
+    case gcSol:
       return XLAT("Sol");
-#endif
+
+    case gcNIH:
+      return XLAT("hyperbolic (3:2)");
+
+    case gcSolN:
+      return XLAT("Sol (3:2)");
 
     case gcNil:
       return XLAT("Nil");
@@ -544,9 +543,22 @@ EX string geometry_name() {
   return "?";
   }
 
+EX string geometry_name() {
+  if(cgi.emb && cgi.emb->is_same_in_same())
+    return geometry_name(geom3::mgclass());
+  else if(embedded_plane && gproduct)
+    return geometry_name(geom3::mgclass()) + " (x E)";
+  else if(embedded_plane)
+    return geometry_name(geom3::mgclass()) + " @ " + geometry_name(geom3::ggclass());
+  else if(among(ginf[geometry].cclass, gcHyperbolic, gcEuclid, gcSphere))
+    return geometry_name(ginf[geometry].cclass) + dim_name();
+  else
+    return geometry_name(ginf[geometry].cclass);
+  };
+
 EX void select_quotient_screen() {
   cmode = sm::SIDE | sm::MAYDARK;
-  gamescreen(0);
+  gamescreen();
 
   dialog::init(XLAT("quotient spaces in ") + ginf[geometry].tiling_name);
   char key = 'a';
@@ -592,7 +604,7 @@ EX void select_quotient_screen() {
   }
 
 EX void select_quotient() {
-  if(euclid && !kite::in() && !arcm::in()) {
+  if(meuclid && !aperiodic && !arcm::in() && !reg3::cubes_reg3 && !(cgflags & qFRACTAL)) {
     euc::prepare_torus3();
     pushScreen(euc::show_torus3);
     }
@@ -606,7 +618,7 @@ EX void select_quotient() {
     pushScreen(asonov::show_config);
     }
   #endif
-  else if(prod)
+  else if(mproduct)
     pushScreen(product::show_config);
   else if(rotspace)
     hybrid::configure_period();
@@ -630,7 +642,7 @@ EX void select_quotient() {
 EX string full_geometry_name() {
   string qstring = ginf[geometry].quotient_name;
   bool variable =
-    !(prod || hybri || bt::in() || (WDIM == 3 && !reg3::in()) || kite::in() || arb::in());
+    !(mproduct || mhybrid || bt::in() || (WDIM == 3 && !reg3::in()) || aperiodic || arb::in());
   
   string fgname = XLAT(ginf[geometry].tiling_name);
   if(qstring != "none") fgname += " " + XLAT(qstring);
@@ -647,7 +659,7 @@ void action_change_variation() {
   else if(arcm::in()) arcm::next_variation();
   #endif
   #if MAXMDIM >= 4
-  else if(reg3::in()) reg3::configure_variation();
+  else if(reg3::in() || geometry == gCubeTiling) reg3::configure_variation();
   #endif
   else if(euc::in(2,4) || !CAP_GP) dialog::do_if_confirmed([] {
     set_variation(PURE ? eVariation::bitruncated : eVariation::pure);
@@ -757,9 +769,9 @@ EX geometry_data compute_geometry_data() {
   gd.area = PURE ? 1 : 3;
 
   gd.euler = 0;
-  if(euclid) gd.euler = 0;
-  else if(sphere && nonorientable) gd.euler = 1;
-  else if(sphere) gd.euler = 2;
+  if(meuclid) gd.euler = 0;
+  else if(msphere && nonorientable) gd.euler = 1;
+  else if(msphere) gd.euler = 2;
   else if(!closed_manifold) gd.euler = -2;
   else if(WDIM == 3) gd.euler = 0;
   else switch(geometry) {
@@ -802,7 +814,7 @@ EX geometry_data compute_geometry_data() {
   gd.nom *= gd.euler;
   gd.denom *= 2;
         
-  if(hybri) gd.nom *= hybrid::csteps, gd.denom *= cgi.single_step;
+  if(mhybrid) gd.nom *= hybrid::csteps, gd.denom *= cgi.single_step;
 
   int g = gcd(gd.nom, gd.denom);
   if(g) {
@@ -897,7 +909,7 @@ EX geometry_data compute_geometry_data() {
   gd.size_str =
     disksize ? its(isize(currentmap->allcells())) :
     #if CAP_BT
-    bt::in() ? fts(8 * M_PI * sqrt(2) * log(2) / pow(vid.binary_width, WDIM-1), 4) + " exp(∞)" :
+    bt::in() ? fts(1440._deg * sqrt(2) * log(2) / pow(vid.binary_width, WDIM-1), 4) + " exp(∞)" :
     #endif
     #if CAP_ARCM
     arcm::in() && (WDIM == 2) ? arcm::current.world_size() :
@@ -932,7 +944,7 @@ EX void showEuclideanMenu() {
   // for(int i=2; i<lt; i++) landvisited[i] = true;
 
   cmode = sm::SIDE | sm::MAYDARK;
-  gamescreen(0);
+  gamescreen();
 
   dialog::init(XLAT("experiment with geometry"));
 
@@ -946,6 +958,7 @@ EX void showEuclideanMenu() {
     });
 
   string qstring = ginf[geometry].quotient_name;
+  if(meuclid && quotient) qstring = ONOFF(true);
 
   if(qstring == "none")
     dialog::addBoolItem(XLAT("quotient space"), false, 'q');
@@ -984,6 +997,7 @@ EX void showEuclideanMenu() {
     dialog::add_action_push(arb::set_sliders);
     }
   if(arb::is_apeirogonal()) add_edit(arb::apeirogon_simplified_display);
+  if(arb::in()) add_edit(debug_tiles);
   
   #if MAXMDIM >= 4
   if(cgflags & qULTRA) {
@@ -995,13 +1009,14 @@ EX void showEuclideanMenu() {
     }
   #endif
   
-  if(prod) {
+  if(mproduct) {
     dialog::addSelItem(XLAT("Z-level height factor"), fts(vid.plevel_factor), 'Z');
     dialog::add_action([] {
       dialog::editNumber(vid.plevel_factor, 0, 2, 0.1, 0.7, XLAT("Z-level height factor"), "");
+      dialog::reaction = ray::reset_raycaster;
       });
     }
-  else if(hybri) {
+  else if(mhybrid) {
     dialog::addSelItem(XLAT("number of levels"), its(hybrid::csteps / cgi.single_step), 'L');
     dialog::add_action(hybrid::configure_period);
     }
@@ -1009,10 +1024,16 @@ EX void showEuclideanMenu() {
     menuitem_binary_width('v');
     add_edit_wall_quality('W');
     }
+  else if(hat::in()) {
+    add_edit(hat::hat_param);
+    add_edit(hat::hat_param_imag);
+    }
   else if(nil) {
     menuitem_nilwidth('v');
     }
-  else if((WDIM == 3 || kite::in() || arb::in()) && !reg3::in()) dialog::addBreak(100);
+  else if((WDIM == 3 || aperiodic || arb::in()) && !reg3::in() && geometry != gCubeTiling) dialog::addBreak(100);
+  else if(cgclass & qFRACTAL)
+    dialog::addBreak(100);
   else 
     menuitem_change_variation('v');
 
@@ -1031,7 +1052,7 @@ EX void showEuclideanMenu() {
     }
   
   #if MAXMDIM >= 4
-  if(hybri) {
+  if(mhybrid) {
     auto r = rots::underlying_scale;
     dialog::addSelItem(XLAT("view the underlying geometry"), r > 0 ? fts(r)+"x" : ONOFF(false), '6');
     dialog::add_action([] {
@@ -1096,8 +1117,9 @@ EX void showEuclideanMenu() {
     dialog::add_action_push(show3D);
     }
   menuitem_projection('1');
-  if(nonisotropic && !sl2)
-    dialog::addBoolItem_action(XLAT("geodesic movement in Sol/Nil"), nisot::geodesic_movement, 'G');
+  if(lie_movement_available()) {
+    add_edit(nisot::geodesic_movement);
+    }
   #if CAP_CRYSTAL && MAXMDIM >= 4
   crystal::add_crystal_transform('x');  
   #endif
@@ -1117,7 +1139,7 @@ EX void showEuclideanMenu() {
   dialog::addTitle(XLAT("info about: %1", full_geometry_name()), 0xFFFFFF, 150);
   
   auto gd = compute_geometry_data();
-  if(WDIM == 2 && !arb::in() && !kite::in()) dialog::addSelItem(XLAT("faces per vertex"), gd.spf, 0);
+  if(WDIM == 2 && !arb::in() && !aperiodic) dialog::addSelItem(XLAT("faces per vertex"), gd.spf, 0);
   
   if(arb::in() && arb::current.comment != "") {
     dialog::addBreak(100);
@@ -1126,7 +1148,7 @@ EX void showEuclideanMenu() {
 
   dialog::addSelItem(XLAT("size of the world"), gd.size_str, '3');
   
-  if(WDIM == 2 || reg3::in_rule()) dialog::add_action([] {
+  if(WDIM == 2 || reg3::exact_rules()) dialog::add_action([] {
     if(!viewdists) { enable_viewdists(); pushScreen(viewdist_configure_dialog); }
     else if(viewdists) viewdists = false;
     });
