@@ -274,6 +274,8 @@ EX namespace gp {
     if(do_adjm) {
       get_adj(wcw.at, wcw.spin) = inverse(wc.adjm) * wc1.adjm;
       get_adj(wcw1.at, wcw1.spin) = inverse(wc1.adjm) * wc.adjm;
+      if(geom3::flipped) gp_swapped.emplace(wcw.at, wcw.spin);
+      if(geom3::flipped) gp_swapped.emplace(wcw1.at, wcw1.spin);
       }
     }
 
@@ -283,6 +285,7 @@ EX namespace gp {
     }
   
   EX map<pair<cell*, int>, transmatrix> gp_adj;
+  EX set<pair<cell*, int>> gp_swapped;
   
   EX transmatrix& get_adj(cell *c, int i) { return gp_adj[make_pair(c,i)]; }
 
@@ -646,22 +649,24 @@ EX namespace gp {
       }
     if(sp>SG3) sp -= SG6;
 
-    return normalize(spin(2*M_PI*sp/S7) * cornmul(T, corner));
+    return normalize(spin(TAU*sp/S7) * cornmul(T, corner));
     }
   
   transmatrix dir_matrix(int i) {
     auto ddspin = [] (int d) -> transmatrix { 
-      return spin(M_PI - d * 2 * M_PI / S7 - cgi.hexshift);
+      return spin(M_PI - d * TAU / S7 - cgi.hexshift);
       };
     return spin(-cgi.gpdata->alpha) * build_matrix(
-      C0, 
-      ddspin(i) * xpush0(cgi.tessf),
-      ddspin(i+1) * xpush0(cgi.tessf),
+      geom3::flipped ? C02 : tile_center(),
+      geom3::flipped ? ddspin(i) * xpush0(cgi.tessf) : ddspin(i) * lxpush0(cgi.tessf),
+      geom3::flipped ? ddspin(i+1) * xpush0(cgi.tessf) : ddspin(i+1) * lxpush0(cgi.tessf),
       C03
       );
     }
   
-  void prepare_matrices() {
+  EX void prepare_matrices(bool inv) {
+    if(!(GOLDBERG_INV || inv)) return;
+    if(embedded_plane) geom3::light_flip(true);
     cgi.gpdata->corners = inverse(build_matrix(
       loctoh_ort(loc(0,0)),
       loctoh_ort(param),
@@ -669,6 +674,9 @@ EX namespace gp {
       C03
       ));
     cgi.gpdata->Tf.resize(S7);
+
+    /* should work directly without flipping but it does not... flipping for now */
+
     for(int i=0; i<S7; i++) {
       transmatrix T = dir_matrix(i);
       for(int x=-GOLDBERG_LIMIT_HALF; x<GOLDBERG_LIMIT_HALF; x++)
@@ -678,9 +686,21 @@ EX namespace gp {
         
         hyperpoint h = atz(T, cgi.gpdata->corners, at, 6);
         hyperpoint hl = atz(T, cgi.gpdata->corners, at + eudir(d), 6);
-        cgi.gpdata->Tf[i][x&GOLDBERG_MASK][y&GOLDBERG_MASK][d] = rgpushxto0(h) * rspintox(gpushxto0(h) * hl) * spin(M_PI);
+        auto& res = cgi.gpdata->Tf[i][x&GOLDBERG_MASK][y&GOLDBERG_MASK][d];
+        res = rgpushxto0(h) * rspintox(gpushxto0(h) * hl) * spin180();
         }
       }       
+
+    if(geom3::flipped) {
+    geom3::light_flip(false);
+    for(int i=0; i<S7; i++) {
+      for(int x=-GOLDBERG_LIMIT_HALF; x<GOLDBERG_LIMIT_HALF; x++)
+      for(int y=-GOLDBERG_LIMIT_HALF; y<GOLDBERG_LIMIT_HALF; y++)
+      for(int d=0; d<(S3==3?6:4); d++) {
+        auto& T = cgi.gpdata->Tf[i][x&GOLDBERG_MASK][y&GOLDBERG_MASK][d];
+        T = cgi.emb->base_to_actual(T);
+        }
+      } }
     }
 
   EX hyperpoint get_corner_position(const local_info& li, int cid, ld cf IS(3)) {
@@ -729,7 +749,6 @@ EX namespace gp {
         cgi.base_distlimit = 2 * param.first + 2 * param.second + 1;
       if(cgi.base_distlimit > SEE_ALL)
         cgi.base_distlimit = SEE_ALL;
-      prepare_matrices();
       DEBB(DF_GEOM | DF_POLY, ("scale = ", scale));
       }
     }
@@ -806,7 +825,7 @@ EX namespace gp {
 
   void show() {
     cmode = sm::SIDE | sm::MAYDARK;
-    gamescreen(0);  
+    gamescreen();
     dialog::init(XLAT("variations"));
     
     int min_quality_chess = 0;
@@ -1444,5 +1463,9 @@ EX namespace gp {
   #endif
 
 
+auto hooksw = addHook(hooks_swapdim, 100, [] {
+  for(auto& p: gp_adj) swapmatrix(p.second);
+  });
+
     
   }}
