@@ -40,12 +40,12 @@ namespace spiral {
     CY = band[0]->h;
     SX = out->w;
     SY = out->h;
+    
+    ld prec = 2*M_PI*M_PI;
 
-    ld k = -2*M_PI*M_PI / log(2.6180339);
+    ld k = -prec / log(2.6180339);
 
-//   cxld mnoznik = cxld(0, M_PI) / cxld(k, M_PI);
-
-    cxld factor = cxld(0, -CY/2/M_PI/M_PI) * cxld(k, M_PI);
+    cxld factor = cxld(0, -CY/prec) * cxld(k, M_PI);
     
     Yshift = CY * k / M_PI;
     
@@ -190,14 +190,16 @@ EX namespace history {
   int lastprogress;
   
   EX void progress_screen() {
-    gamescreen(0);
+    gamescreen();
     mouseovers = "";
     }
 
+  EX int progress_each = 250;
+
   EX void progress(string str) {
 #if CAP_SDL
     int tick = SDL_GetTicks();
-    if(tick > lastprogress + 250) {
+    if(tick > lastprogress + progress_each) {
       lastprogress = tick;
       msgs.clear();
       addMessage(str);
@@ -370,7 +372,7 @@ EX namespace history {
       v[ph+1]->at * C0;
   
     View = xpush(-(phase-ph) * hdist(now, next)) * View;
-    if(WDIM == 2 || prod) {
+    if(WDIM == 2 || mproduct) {
       View = models::rotmatrix() * View;
       }
     else {
@@ -568,7 +570,7 @@ EX namespace history {
   
   EX void history_menu() {
     cmode = sm::SIDE | sm::MAYDARK;
-    gamescreen(0);
+    gamescreen();
     
     dialog::init(XLAT("history mode"));
 
