@@ -45,6 +45,7 @@ EX int subclass(int i) {
 #define GLYPH_LOCAL2     256
 #define GLYPH_TARGET     512
 #define GLYPH_INSQUARE   1024
+#define GLYPH_INLANDSCAPE 2048
 
 #if HDR
 enum eGlyphsortorder {
@@ -147,7 +148,7 @@ int glyphflags(int gid) {
       if(i == localshardof(cwt.at->land)) f |= GLYPH_LOCAL2;
       }
     if(i == treasureType(cwt.at->land) || daily::on) 
-      f |= GLYPH_LOCAL | GLYPH_LOCAL2 | GLYPH_IMPORTANT | GLYPH_INSQUARE;
+      f |= GLYPH_LOCAL | GLYPH_LOCAL2 | GLYPH_IMPORTANT | GLYPH_INSQUARE | GLYPH_INPORTRAIT | GLYPH_INLANDSCAPE;
     if(i == itHolyGrail) {
       if(items[i] >= 3 && !inv::on) f |= GLYPH_MARKOVER;
       }
@@ -156,19 +157,21 @@ int glyphflags(int gid) {
       else if(items[i] < 10) f |= GLYPH_MARKTODO;
       }
     else {
-      f |= GLYPH_IMPORTANT | GLYPH_INSQUARE;
+      f |= GLYPH_IMPORTANT | GLYPH_INSQUARE | GLYPH_INPORTRAIT | GLYPH_INLANDSCAPE;
       if(itemclass(i) == IC_ORB && items[i] < 10) f |= GLYPH_RUNOUT;
       }
     if(i == orbToTarget) f |= GLYPH_TARGET;
-    f |= GLYPH_INPORTRAIT;
+    if(!less_in_portrait) f |= GLYPH_INPORTRAIT;
+    if(!less_in_landscape) f |= GLYPH_INLANDSCAPE;
     }
   else {
     eMonster m = eMonster(gid-ittypes);
-    if(m == moLesser) f |= GLYPH_IMPORTANT | GLYPH_DEMON | GLYPH_INPORTRAIT | GLYPH_INSQUARE;
+    if(m == moLesser) f |= GLYPH_IMPORTANT | GLYPH_DEMON | GLYPH_INPORTRAIT | GLYPH_INSQUARE | GLYPH_INLANDSCAPE;
     int isnat = isNative(cwt.at->land, m);
-    if(isnat) f |= GLYPH_LOCAL | GLYPH_IMPORTANT | GLYPH_INPORTRAIT | GLYPH_INSQUARE;
+    if(isnat) f |= GLYPH_LOCAL | GLYPH_IMPORTANT | GLYPH_INPORTRAIT | GLYPH_INSQUARE | GLYPH_INLANDSCAPE;
     if(isnat == 2) f |= GLYPH_LOCAL2;
     if(m == monsterToSummon) f |= GLYPH_TARGET;
+    if(!less_in_landscape) f |= GLYPH_INLANDSCAPE;
     }
   return f;
   }
@@ -194,6 +197,7 @@ bool displayglyph(int cx, int cy, int buttonsize, char glyph, color_t color, int
   
   if(graphglyph(isMonster)) {
     initquickqueue();
+    poly_outline = OUTLINE_DEFAULT;
     if(isMonster) {
       eMonster m = eMonster(id - ittypes);
       double bsize = buttonsize * 2/3;
@@ -211,11 +215,19 @@ bool displayglyph(int cx, int cy, int buttonsize, char glyph, color_t color, int
     else {
       eItem it = eItem(id);
       double bsize = buttonsize / 2;
-      if(glyph =='*') bsize *= 2;
+      if(glyph == '*') bsize *= 2;
+      if(glyph == '%') bsize *= 2;
+      if(glyph == '!') bsize *= 2;
+      if(glyph == '?') bsize *= 2;
+      if(glyph == ';') bsize *= 2;
       if(glyph == '$') bsize = (bsize*5+2)/3;
       if(glyph == 'o') bsize = (bsize*3+1)/2;
+      if(glyph == 'c') bsize = (bsize*3+1)/2;
       if(glyph == 't') bsize = bsize*5/2;
       if(glyph == '(') bsize = bsize*2.5;
+      if(glyph == '/') bsize = bsize*2;
+      if(glyph == ']') bsize = bsize*1.5;
+      if(glyph == 'x') bsize = bsize*1.5;
       if(it == itWarning) bsize *= 2;
       if(it == itBombEgg || it == itTrollEgg || it == itDodeca) bsize = bsize*3/2;
       int icol = color;
@@ -228,8 +240,17 @@ bool displayglyph(int cx, int cy, int buttonsize, char glyph, color_t color, int
         ((glyph == 't' && qty%5) || it == itOrbYendor) ? ticks/2 : 
         it == itTerra ? glyphphase[id] * 3 * M_PI + 900 * M_PI:
         glyphphase[id] * 2;
+
       drawItemType(it, NULL, shiftless(V), icol, t, false);
+      
+      int c1 = max(color_diff(color, backcolor), color_diff(color, bordcolor));
+      if(c1 < 0x80) {
+        color = orb_auxiliary_color(it);
+        }
       }
+    int c1 = max(color_diff(color, backcolor), color_diff(color, bordcolor));
+    if(c1 < 0x80)
+      color = gradient(color, forecolor, 0, .4, 1);
     sortquickqueue();
     quickqueue();
     }
@@ -405,6 +426,8 @@ EX void draw_crosshair() {
     }
   return;
   }
+
+EX bool less_in_portrait, less_in_landscape;
   
 EX void drawStats() {
   if(vid.stereo_mode == sLR) return;
@@ -417,6 +440,7 @@ EX void drawStats() {
   
   first_cell_to_draw = true;
   bool h = hide_player();
+  instat = false;
 
   bool cornermode = (vid.xres > vid.yres * 85/100 && vid.yres > vid.xres * 85/100);
   
@@ -504,20 +528,22 @@ EX void drawStats() {
   
   else {
   
-    instat = false;
     bool portrait = vid.xres < vid.yres;
-    int colspace = portrait ? (vid.yres - vid.xres - vid.fsize*3) : (vid.xres - vid.yres - 16) / 2;
+    int colspace = portrait ? (current_display->ycenter - current_display->scrsize - 3 * vid.fsize) : (vid.xres - vid.yres - 16) / 2;
     int rowspace = portrait ? vid.xres - 16 : vid.yres - vid.fsize * (vid.msgleft ? 9 : 4);
     int colid[4], rowid[4];
     int maxbyclass[4];
     for(int z=0; z<4; z++) maxbyclass[z] = 0;
+
+    flagtype flag = portrait ? GLYPH_INPORTRAIT : GLYPH_INLANDSCAPE;
+
     for(int i=0; i<glyphs; i++) if(ikappear(i))
-      if(!portrait || (glyphflags(i) | GLYPH_INPORTRAIT))
+      if(glyphflags(i) & flag)
         maxbyclass[glyphclass(i)]++;
     int buttonsize;
     int columns, rows;
     bool imponly = false;
-    int minsize = vid.fsize * (portrait ? 4 : 2);  
+    int minsize = vid.fsize * (portrait ? 3 : 2);
     rows = 0;
     while((buttonsize = minsize - vid.killreduction)) {
       columns = colspace / buttonsize;
@@ -549,6 +575,7 @@ EX void drawStats() {
       if(!ikappear(i)) continue;
       int z = glyphclass(i);
       int imp = glyphflags(i);
+      if(!(imp & flag)) continue;
       if(imponly) { z &=~1; if(!(imp & GLYPH_IMPORTANT)) continue; }
   
       int cx, cy;
