@@ -13,8 +13,8 @@
 #define _HYPER_H_
 
 // version numbers
-#define VER "12.0q"
-#define VERNUM_HEX 0xA911
+#define VER "12.1q"
+#define VERNUM_HEX 0xA931
 
 #include "sysconfig.h"
 
@@ -110,6 +110,11 @@ struct hr_shortest_path_exception { };
 #define GEN_F 1
 #define GEN_N 2
 #define GEN_O 3
+#define GEN_BASE_MASK   7
+
+#define GENF_ELISION    (1 << 3)
+#define GENF_PLURALONLY (1 << 4)
+#define GENF_PROPER     (1 << 5)
 
 // Add a message to the GUI.
 // If multiple messages appear with the same spamtype != 0, the older ones disappear quickly
@@ -124,7 +129,7 @@ void addMessage(string s, char spamtype = 0);
 #define hyperbolic_37 (S7 == 7 && S3 == 3 && !bt::in() && !arcm::in())
 #define hyperbolic_not37 ((S7 > 7 || S3 > 3 || bt::in() || arcm::in()) && hyperbolic)
 #define weirdhyperbolic ((S7 > 7 || S3 > 3 || !STDVAR || bt::in() || arcm::in() || arb::in()) && hyperbolic)
-#define stdhyperbolic (S7 == 7 && S3 == 3 && STDVAR && !bt::in() && !arcm::in())
+#define stdhyperbolic (S7 == 7 && S3 == 3 && STDVAR && !bt::in() && !arcm::in() && !arb::in())
 
 #define cgflags cginf.flags 
 
@@ -139,15 +144,13 @@ void addMessage(string s, char spamtype = 0);
 #define cgclass (cginf.cclass)
 #define euclid (cgclass == gcEuclid)
 #define sphere (cgclass == gcSphere)
-#define sol (cgflags & qSOL)
-#define nih (cgflags & qNIH)
+#define sol (among(cgclass, gcSol, gcSolN))
+#define nih (among(cgclass, gcNIH, gcSolN))
 #define nil (cgclass == gcNil)
 #define sl2 (cgclass == gcSL2)
-#define prod (cgclass == gcProduct)
-#define hybri (cgflags & qHYBRID)
 #define rotspace (geometry == gRotSpace)
 #define hyperbolic (cgclass == gcHyperbolic)
-#define nonisotropic (among(cgclass, gcSolNIH, gcNil, gcSL2))
+#define nonisotropic (among(cgclass, gcSol, gcSolN, gcNIH, gcSL2, gcNil))
 #define translatable (euclid || nonisotropic)
 #define nonorientable (cgflags & qNONORIENTABLE)
 #define elliptic (cgflags & qELLIPTIC)
@@ -155,6 +158,19 @@ void addMessage(string s, char spamtype = 0);
 #define smallbounded ((cgflags & qSMALL) || disksize)
 #define closed_manifold (cgflags & qCLOSED)
 #define closed_or_bounded (closed_manifold || disksize)
+/** the actual map has hybrid geometry, not just the graphics */
+#define mhybrid (cgflags & qHYBRID)
+/** graphics based on a product geometry -- either embedded or actual hybrid product */
+#define gproduct (cgclass == gcProduct)
+
+/** 2D geometry embedded in 3D */
+#define embedded_plane (WDIM == 2 && GDIM == 3)
+/** the actual map is product, not just the graphics */
+#define mproduct (gproduct && !embedded_plane)
+/** the actual map is product, not just the graphics */
+#define meuclid (geom3::mgclass() == gcEuclid)
+#define msphere (geom3::mgclass() == gcSphere)
+#define mhyperbolic (geom3::mgclass() == gcHyperbolic)
 
 // Dry Forest burning, heat transfer, etc. are performed on the whole universe
 #define doall (closed_or_bounded)
@@ -184,6 +200,7 @@ void addMessage(string s, char spamtype = 0);
 
 #define INVERSE among(variation, eVariation::unrectified, eVariation::warped, eVariation::untruncated )
 
+#define aperiodic (cgflags & qAPERIODIC)
 #define UNRECTIFIED (variation == eVariation::unrectified)
 #define WARPED (variation == eVariation::warped)
 #define UNTRUNCATED (variation == eVariation::untruncated)
@@ -241,7 +258,7 @@ enum eModel : int;
 struct projection_configuration {
   eModel model;            /**< which projection, see classes.cpp */
   ld xposition, yposition; /**< move the center to another position */
-  ld scale, alpha, camera_angle, fisheye_param, twopoint_param, stretch, ballangle, ballproj, euclid_to_sphere;
+  ld scale, alpha, camera_angle, fisheye_param, twopoint_param, axial_angle, stretch, ballangle, ballproj, euclid_to_sphere;
   ld clip_min, clip_max;
   ld model_orientation, halfplane_scale, model_orientation_yz;  
   ld collignon_parameter; 
@@ -292,11 +309,16 @@ struct projection_configuration {
     vr_scale_factor = 1;
     back_and_front = 0;
     dualfocus_autoscale = false;
+    axial_angle = 90;
     }
   };
 
 enum eThreatLevel { tlNoThreat, tlSpam, tlNormal, tlHighThreat };
 
+constexpr ld use_the_default_value = -20.0625;
+
+enum ePseudohedral { phOFF, phInscribed, phCircumscribed };
+
 struct videopar {
   projection_configuration projection_config, rug_config;
   ld yshift;
@@ -338,7 +360,7 @@ struct videopar {
   int fsize, abs_fsize, fontscale;
   int flashtime;
   
-  int wallmode, monmode, axes, highlightmode;
+  int wallmode, monmode, axes, highlightmode, orbmode;
   bool axes3;
   bool revcontrol;
   
@@ -400,10 +422,12 @@ struct videopar {
   ld depth;      // world level below the plane
   ld camera;     // camera level above the plane
   ld wall_height, creature_scale, height_width;
-  ld lake_top, lake_bottom;
+  ld lake_top, lake_bottom, lake_shallow, wall_height2, wall_height3;
+  ld lowsky_height, sky_height, star_height, infdeep_height, star_size, sun_size;
+  bool height_limits;
   ld rock_wall_ratio;
   ld human_wall_ratio;
-  bool pseudohedral; // in 3D modes
+  ePseudohedral pseudohedral;
   ld depth_bonus;   // to fiix the placement of 3D models in pseudogonal -- not working currently
 
   int tc_alpha, tc_depth, tc_camera;
@@ -673,8 +697,6 @@ static const int MAXPLAYER = 7;
 // just in case if I change my mind about when Orbs lose their power
 #define ORBBASE 0
 
-#define mmscale(V, x) (mmspatial ? (ivoryz ? mzscale(V,x) : mscale(V, x)) : (V))
-
 #define SHADOW_WALL 0x60
 #define SHADOW_SL   0x18
 #define SHADOW_MON  0x30
@@ -864,6 +886,18 @@ const typename Map::mapped_type *at_or_null(const Map& map, const Key& key) {
   return (it == map.end()) ? nullptr : &it->second;
   }
 
+int gmod(int i, int j);
+
+// vector::at(i) modulo its size (const version)
+template<class T> const T& atmod(const vector<T>& container, int index) {
+  return container[gmod(index, isize(container))];
+  }
+
+// vector::at(i) modulo its size (non-const version)
+template<class T> T& atmod(vector<T>& container, int index) {
+  return container[gmod(index, isize(container))];
+  }
+
 namespace daily {
   extern bool on;
   extern int daily_id;
