@@ -6,7 +6,7 @@ EX namespace intra {
 
 EX bool in;
 
-#if CAP_RAY && MAXMDIM >= 4
+#if CAP_PORTALS
 #if HDR
 /** information per every space connected with intra-portals */
 struct intra_data {
@@ -18,6 +18,9 @@ struct intra_data {
 
 EX vector<intra_data> data;
 
+/** tells gamedata store that we are just storing one world */
+EX bool switching;
+
 /** index of the space we are currently in */
 EX int current;
 
@@ -46,7 +49,7 @@ struct portal_data {
 #endif
 
 hyperpoint portal_data::to_poco(hyperpoint h) const {
-  if(prod && kind == 1) {
+  if(mproduct && kind == 1) {
     auto dec = product_decompose(h);
     h = dec.second;
     if(bt::in()) {
@@ -64,7 +67,7 @@ hyperpoint portal_data::to_poco(hyperpoint h) const {
     if(d<0) h[2] = -h[2], h[0] = -h[0];
     return h;
     }
-  else if(prod && kind == 0) {
+  else if(mproduct && kind == 0) {
     h = T * h;
     ld z = product_decompose(h).first;
     h /= exp(z);
@@ -102,7 +105,7 @@ hyperpoint portal_data::to_poco(hyperpoint h) const {
   }
 
 hyperpoint portal_data::from_poco(hyperpoint h) const {
-  if(prod && kind == 1) {
+  if(mproduct && kind == 1) {
     ld xd = h[2];
     if(d<0) xd = -xd, h[0] = -h[0];
     #if CAP_BT
@@ -115,7 +118,7 @@ hyperpoint portal_data::from_poco(hyperpoint h) const {
     auto z = product_decompose(h).first;
     return iT * h * exp(d+xd-z);
     }
-  else if(prod && kind == 0) {
+  else if(mproduct && kind == 0) {
     auto h0 = h;
     h[0] = sin_auto(h0[2]);
     h[1] = sin_auto(h0[0]) * cos_auto(h0[2]);
@@ -153,14 +156,14 @@ EX portal_data make_portal(cellwalker cw, int spin) {
   id.scale = 1;
   id.T = Id;
   auto gg = geometry;
-  if(prod && cw.spin >= cw.at->type - 2) {
+  if(mproduct && cw.spin >= cw.at->type - 2) {
     id.kind = 1;
     id.d = product_decompose(fac[0]).first;
     id.v0 = C0 * exp(id.d);
     #if CAP_BT
     if(bt::in()) {
       for(auto h: fac)
-        println(hlog, PIU(deparabolic13(normalize_flat(h))));
+        println(hlog, PIU(deparabolic13(cgi.emb->normalize_flat(h))));
       if(cw.spin == cw.at->type - 2)
         fac.pop_back();
       else
@@ -173,19 +176,19 @@ EX portal_data make_portal(cellwalker cw, int spin) {
     else {
       hyperpoint ctr = Hypc;
       for(auto p: fac) ctr += product_decompose(p).second;
-      ctr = normalize_flat(ctr);
+      ctr = cgi.emb->normalize_flat(ctr);
       id.T = gpushxto0(ctr);
       }
     }
-  else if(prod) {
+  else if(mproduct) {
     id.kind = 0;
     id.v0 = Hypc;
     id.scale = cgi.plevel;
     for(auto p: fac) id.v0 += p;
-    id.v0 = normalize_flat(id.v0);
-    hyperpoint h = normalize_flat(fac[0]);
-    id.T = cspin(0, 1, -90*degree) * spintox(gpushxto0(id.v0) * h) * gpushxto0(id.v0);
-    if((id.T * C0)[0] > 0) id.T = cspin(0, 1, 180*degree) * id.T;
+    id.v0 = cgi.emb->normalize_flat(id.v0);
+    hyperpoint h = cgi.emb->normalize_flat(fac[0]);
+    id.T = cspin90(1, 0) * spintox(gpushxto0(id.v0) * h) * gpushxto0(id.v0);
+    if((id.T * C0)[0] > 0) id.T = spin180() * id.T;
     for(int i=0; i<3; i++) id.T[3][i] = id.T[i][3] = i==3;
     if(debug_portal & 128)
     for(int a=0; a<4; a++) {
@@ -230,13 +233,13 @@ EX portal_data make_portal(cellwalker cw, int spin) {
     id.T = gpushxto0(id.v0);
     for(auto p: fac1) {
       if(abs((id.T * p)[2]) > 1e-3 && abs((id.T * p)[0]) < 1e-3)
-        id.T = cspin(2, 0, 90*degree) * id.T;
+        id.T = cspin90(2, 0) * id.T;
       if(abs((id.T * p)[2]) > 1e-3 && abs((id.T * p)[1]) < 1e-3)
-        id.T = cspin(2, 1, 90*degree) * id.T;
+        id.T = cspin90(2, 1) * id.T;
       }
-    if((id.T * C03)[2] > 0) id.T = cspin(2, 0, 180*degree) * id.T;
-    if(abs((id.T * removed)[0]) > 1e-2) id.T = cspin(0, 1, 90*degree) * id.T;
-    if((id.T * removed)[1] < -1e-2) id.T = cspin(0, 1, 180*degree) * id.T;
+    if((id.T * C03)[2] > 0) id.T = cspin180(2, 0) * id.T;
+    if(abs((id.T * removed)[0]) > 1e-2) id.T = cspin90(0, 1) * id.T;
+    if((id.T * removed)[1] < -1e-2) id.T = cspin180(0, 1) * id.T;
     vector<hyperpoint> v;
     geometry = gg;
     for(auto f: fac) v.push_back(id.to_poco(f));
@@ -251,7 +254,7 @@ EX portal_data make_portal(cellwalker cw, int spin) {
   else {
     id.kind = 0;
     id.v0 = project_on_triangle(fac[0], fac[1], fac[2]);
-    id.T = cpush(2, -hdist0(id.v0)) * cspin(2, 0, 90*degree) * spintox(id.v0);
+    id.T = cpush(2, -hdist0(id.v0)) * cspin90(2, 0) * spintox(id.v0);
     hyperpoint ctr = Hypc;
     for(auto p: fac) ctr += id.T*p;
     ctr = normalize(ctr);
@@ -285,9 +288,9 @@ EX portal_data make_portal(cellwalker cw, int spin) {
       println(hlog, "deep       ", x, " = ", hdist(res(x*.1,0,0), res(x*.1,0,0.001)));
       }
     hyperpoint a = hyperpoint(.4, .2, .1, 1);
-    println(hlog, "a = ", a);
-    println(hlog, "b = ", id.from_poco(a));
-    println(hlog, "c = ", id.to_poco(id.from_poco(a)));
+    println(hlog, "a = ", kz(a));
+    println(hlog, "b = ", kz(id.from_poco(a)));
+    println(hlog, "c = ", kz(id.to_poco(id.from_poco(a))));
     }
 
   if(debug_portal & 1) {
@@ -327,6 +330,7 @@ EX connection_data* find_connection(int a, int b) {
 
 EX void switch_to(int id) {
   if(current == id) return;
+  dynamicval<bool> is(switching, true);
   data[current].gd.storegame();
   current = id;
   ginf[gProduct] = data[current].gi;
@@ -435,6 +439,7 @@ EX void regenerate_full_sample_list() {
 
 /** make currentmap into one of the spaces in intra */
 EX void become() {
+  dynamicval<bool> is(switching, true);
   if(intra::in) {
     /* let them add more spaces in this case */
     data[current].gd.storegame();
@@ -462,6 +467,7 @@ EX void become() {
 EX void start(int id IS(0)) {
   in = true;
   current = id;
+  dynamicval<bool> is(switching, true);
   data[current].gd.restoregame();
   ginf[gProduct] = data[current].gi;
 
@@ -512,7 +518,7 @@ EX void analyze_orthonormal(array<hyperpoint, 4> ds, ld sca) {
   vector<ld> orths;
   for(int i: {1,2,3}) {
     ds[i] = T * ds[i];
-    if(prod) ds[i][2]--;
+    if(mproduct) ds[i][2]--;
     }
   for(int i=0; i<3; i++)
   for(int j=0; j<3; j++)
@@ -540,13 +546,8 @@ EX void shift_view_portal(hyperpoint H) {
 
 EX const connection_data* through_portal() {
   transmatrix iView = view_inverse(View);
-  ld dist = hdist0(iView * C0);
-  int nei = -1;
-  for(int i=0; i<centerover->type; i++) {
-    ld dist1 = hdist0(currentmap->ray_iadj(centerover, i) * iView * C0);
-    if(dist1 < dist) nei = i, dist = dist1;
-    }
-
+  int nei = through_wall(centerover, iView * C0);
+  if(nei == -1) return nullptr;
   auto cw1 = cellwalker(centerover, nei);
   return at_or_null(connections, cw1);
   }
@@ -641,24 +642,105 @@ void erase_unconnected(cellwalker cw) {
       unconnected.erase(unconnected.begin() + i);
   }
 
+EX string portal_help =
+  "In portal maps, you can create portals on walls. Such portals can be entered to end up in another world. "
+  "The portals are 'immersive', that is, you can see through the portals. In the current implementation, "
+  "the immersive view works only in the raycasting mode, for rendering walls -- so any other objects will not "
+  "be rendered. (You can turn raycasting off to see the world without any portals.) As a consequence, "
+  "there is no HyperRogue game in a portal map -- you can only move the camera around the scene. Pick "
+  "'set recommended settings' in the manage portals menu to optimize settings for this.\n\n"
+  "To create portals between different geometries, you need to create a map in the other geometry and save it "
+  "(using the map editor), then in your portal map such a world can be loaded to be added to the scene "
+  "(in the manage portals menu). Such portals work correctly when their intrinsic shapes are equal (including the "
+  "curvature), their extrinsic curvatures match, and are actually implemented; see the video 'Portals to "
+  "Non-Euclidean Geometries' for some portals that work.\n\n"
+  ;
+
+EX void become_menu() {
+  cmode = sm::SIDE | sm::MAYDARK;
+  gamescreen();
+  dialog::init(XLAT("Become a portal map"));
+  dialog::addHelp(XLAT(portal_help));
+  dialog::addItem(XLAT("yes, that's what I want"), 'y');
+  dialog::add_action([] {
+    intra::become();
+    intra::start();
+    game_keys_scroll = true;
+    mapeditor::drawplayer = false;
+    popScreen();
+    pushScreen(show_portals);
+    });
+
+  dialog::addBack();
+  dialog::display();
+  }
+
 int edit_spin;
 
+EX void world_list() {
+  cmode = sm::SIDE | sm::MAYDARK;
+  gamescreen();
+  dialog::init(XLAT("world list"));
+  dialog::start_list(900, 900, '1');
+  int c = current;
+  for(int i=0; i<isize(data); i++) {
+    switch_to(i);
+    dialog::addBoolItem(full_geometry_name(), i == c, dialog::list_fake_key++);
+    dialog::add_action([i] {
+      int ic = i;
+      switch_to(ic);
+      });
+    }
+  switch_to(c);
+  dialog::end_list();
+  dialog::addBreak(100);
+  #if CAP_EDIT
+  dialog::addItem("add a saved world to the scene", 'a');
+  dialog::add_action([] {
+    dialog::openFileDialog(levelfile, XLAT("level to load:"), ".lev", [] () {
+      intra::become();
+      if(mapstream::loadMap(levelfile.c_str())) {
+        addMessage(XLAT("Map loaded from %1", levelfile));
+        intra::become();
+        intra::start();
+        mapeditor::map_version++;
+        ray::reset_raycaster();
+        return true;
+        }
+      else {
+        addMessage(XLAT("Failed to load map from %1", levelfile));
+        intra::start();
+        return false;
+        }
+      });
+    });
+  #endif
+  dialog::addBack();
+  dialog::display();
+  }
+
 EX void show_portals() {
-  gamescreen(1);
+  cmode = sm::SIDE | sm::MAYDARK | sm::PANNING;
+  if(mapeditor::building_mode) cmode |= sm::EDIT_INSIDE_WALLS;
+  gamescreen();
 
   dialog::init(XLAT("manage portals"));
 
-  cellwalker cw(centerover, point_direction);
-  bool valid = point_direction >= 0 && point_direction < centerover->type;
+  cellwalker cw(mouseover2, point_direction);
+  bool valid = point_direction >= 0 && point_direction < mouseover2->type;
 
-  dialog::addItem(XLAT("move to the next space"), 'm');
-  dialog::add_action([] {
-    int ic = (current + 1) % isize(data);
-    switch_to(ic);
-    });
+  if(valid) {
+    initquickqueue();
+    for (const shiftmatrix& V : hr::span_at(current_display->all_drawn_copies, mouseover2)) {
+      dynamicval<color_t> p(poly_outline, 0xFFFFFF);
+      int ofs = currentmap->wall_offset(mouseover2);
+      queuepolyat(V, cgi.shWireframe3D[ofs + point_direction], 0, PPR::SUPERLINE);
+      }
+    quickqueue();
+    }
 
-  dialog::addSelItem(XLAT("mode"), ray::fixed_map ? "perf" : "edit", 'e');
-  dialog::add_action([] { ray::fixed_map = !ray::fixed_map; });
+  dialog::addItem(XLAT("view another world"), 'm');
+  dialog::add_action_push(world_list);
 
   if(debug_portal) {
     dialog::addItem(XLAT("debug"), 'd');
@@ -683,6 +765,7 @@ EX void show_portals() {
       unconnected.push_back(tcw);
       connections.erase(cw);
       connections.erase(tcw);
+      mapeditor::map_version++;
       });
     }
   else if(in_list) {
@@ -694,13 +777,16 @@ EX void show_portals() {
   else {
     dialog::addItem(XLAT("add to list"), 'a');
     dialog::add_action([cw] { unconnected.push_back(cw); });
+    dialog::start_list(500, 500, '1');
     for(auto p: unconnected) {
-      dialog::addItem(XLAT("connect " + lalign(0, p)), '1');
+      dialog::addItem(XLAT("connect " + lalign(0, p)), dialog::list_fake_key++);
       dialog::add_action([p, cw] {
         connect_portal(cw, p, edit_spin);
+        mapeditor::map_version++;
         erase_unconnected(p);
         });
       }
+    dialog::end_list();
     dialog::addSelItem(XLAT("portal orientation"), its(edit_spin), 'o');
     dialog::add_action([] { edit_spin = edit_spin + 1; });
     if(debug_portal) {
@@ -713,8 +799,36 @@ EX void show_portals() {
       }
     }
 
+  if(!game_keys_scroll || mapeditor::drawplayer || !ray::fixed_map || vid.cells_drawn_limit > 100) {
+    dialog::addItem(XLAT("set recommended settings"), 'A');
+    dialog::add_action([] {
+      game_keys_scroll = true;
+      mapeditor::drawplayer = false;
+      ray::fixed_map = true;
+      vid.cells_drawn_limit = 100;
+      });
+    }
+  else {
+    dialog::addItem(XLAT("configure raycasting"), 'A');
+    dialog::add_action_push(ray::configure);
+    }
+
+  if(mproduct && point_direction < mouseover2->type - 2) {
+    ld r = get_ratio_edge(mouseover2, point_direction);
+    dialog::addSelItem(XLAT("height-to-width ratio"), fts(r), 'r');
+    dialog::add_action([] {
+      be_ratio_edge(mouseover2, point_direction);
+      mapeditor::map_version++;
+      ray::reset_raycaster();
+      });
+    }
+  else dialog::addBreak(100);
+
   walking::add_options();
 
+  dialog::addHelp();
+  dialog::add_action([] { gotoHelp(portal_help); });
+
   dialog::display();
   }
 
@@ -748,14 +862,28 @@ EX void be_ratio(ld v IS(1)) {
   cgi.require_basics();
   }
 
-EX void be_ratio_edge(int i, ld v IS(1)) {
-  start_game();
-  ld len = hdist(currentmap->get_corner(cwt.at, i), currentmap->get_corner(cwt.at, (i+1)%cwt.at->type));
-  PIU( vid.plevel_factor = v * len / cgi.scalefactor );
+EX ld get_edge_length(cell *c, int i) {
+  auto c1 = hybrid::get_where(c).first;
+  return PIU( hdist(currentmap->get_corner(c1, i), currentmap->get_corner(c1, (i+c1->type-1)%c1->type)) );
+  }
+
+EX ld get_ratio_edge(cell *c, int i) {
+  ld len = get_edge_length(c, i);
+  return vid.plevel_factor * cgi.scalefactor / len;
+  }
+
+EX void be_ratio_edge(cell *c, int i, ld v IS(1)) {
+  ld len = get_edge_length(c, i);
+  vid.plevel_factor = v * len / cgi.scalefactor;
   check_cgi();
   cgi.require_basics();
   }
 
+EX void be_ratio_edge(int i, ld v IS(1)) {
+  start_game();
+  be_ratio_edge(cwt.at, i, v);
+  }
+
 /** Remove the space with the given id. Turns off intra */
 EX void kill(int id) {
   if(in) become();
@@ -781,6 +909,7 @@ EX void kill(int id) {
 
 EX void erase_all_maps() {
   println(hlog, "erase_all_maps called");
+  dynamicval<bool> is(switching, true);
   data[current].gd.storegame();
   in = false;
   for(int i=0; i<isize(data); i++) {
@@ -845,7 +974,7 @@ EX ld eye_level = 0.2174492;
 EX ld eye_angle = 0;
 EX ld eye_angle_scale = 1;
 
-int ticks_end, ticks_last;
+EX int ticks_end, ticks_last;
 
 EX set<color_t> colors_of_floors;
 
@@ -858,6 +987,7 @@ EX bool isFloor(cell *c) {
 
 EX void handle() {
   if(playermoved || !on) return;
+  make_actual_view();
 
   if(floor_dir == -1 || on_floor_of != centerover) {
     vector<int> choices;
@@ -920,13 +1050,13 @@ EX void handle() {
       z[2] = bt::minkowski_to_bt(fac.h0)[2];
       return bt::bt_to_minkowski(z);
       }
-    else if(prod && bt::in()) {
+    else if(mproduct && bt::in()) {
       auto dec = product_decompose(at);
       hyperpoint dep = PIU( deparabolic13(dec.second) );
       hyperpoint h = product_decompose(fac.h0).second;
       h = PIU( deparabolic13(h) );
       dep[0] = h[0];
-      return zshift(PIU(parabolic13(dep)), dec.first);
+      return orthogonal_move(PIU(parabolic13(dep)), dec.first);
       }
     #else
     if(false) {}
@@ -1014,16 +1144,18 @@ EX void handle() {
   ticks_last = ticks;
   }
 
+EX void switch_walking() {
+  on = !on;
+  if(on && auto_eyelevel) eye_level = -1;
+  floor_dir = -1;
+  on_floor_of = nullptr;
+  ticks_last = ticks;
+  ticks_end = ticks + 1000;
+  }
+
 EX void add_options() {
   dialog::addBoolItem("walking mode", on, 'w');
-  dialog::add_action([] {
-    on = !on;
-    if(on && auto_eyelevel) eye_level = -1;
-    floor_dir = -1;
-    on_floor_of = nullptr;
-    ticks_last = ticks;
-    ticks_end = ticks + 1000;
-    });
+  dialog::add_action(switch_walking);
   add_edit(eye_level);
   add_edit(eye_angle);
   if(point_direction >= 0 && point_direction < centerover->type) {
@@ -1042,17 +1174,17 @@ EX void add_options() {
 auto a = addHook(hooks_configfile, 100, [] {
   param_b(auto_eyelevel, "auto_eyelevel")
       -> editable("keep eye level when walking enabled", 'L');
-  param_f(eye_level, "eye_level")
+  param_f(eye_level, "walk_eye_level")
       -> editable(0, 5, .1, "walking eye level",
       "Distance from the floor to the eye in the walking mode, in absolute units. In VR this is adjusted automatically.",
       'e')
       ->set_extra([] { add_edit(auto_eyelevel); });
-  param_f(eye_angle, "eye_angle")
+  param_f(eye_angle, "walk_eye_angle")
       -> editable(-90, 90, 15, "walking eye angle",
       "0 = looking forward, 90 = looking upward. In VR this is adjusted automatically.",
       'k')
       ->set_extra([] { add_edit(eye_angle_scale); });
-  param_f(eye_angle_scale, "eye_angle_scale")
+  param_f(eye_angle_scale, "walk_eye_angle_scale")
       -> editable(-2, 0, 2, "eye angle scale",
       "1 = the angle can be changed with keyboard or mouse movements, 0 = the angle is fixed",
       'k');
