@@ -10,9 +10,13 @@ namespace hr {
 
 EX namespace mapeditor {
 
+  /* changes when the map is changed */
+  EX int map_version;
+
   EX bool drawing_tool;
   EX bool intexture;
   EX bool snapping;
+  EX bool building_mode = true;
 
   #if HDR
   enum eShapegroup { sgPlayer, sgMonster, sgItem, sgFloor, sgWall };
@@ -85,7 +89,7 @@ EX namespace mapeditor {
       int ll = ceil(360 * len);
       shiftmatrix W = V * rgpushxto0(s);
       for(int i=0; i<=ll; i++)
-        curvepoint(xspinpush0(360*degree*i/ll, radius));
+        curvepoint(xspinpush0(TAU*i/ll, radius));
       queuecurve(W, col, fill, PPR::LINE);
       }
 
@@ -395,7 +399,7 @@ EX namespace mapstream {
   EX vector<cell*> cellbyid;
   EX vector<char> relspin;
   
-  void load_drawing_tool(fhstream& hs) {
+  void load_drawing_tool(hstream& hs) {
     using namespace mapeditor;
     if(hs.vernum < 0xA82A) return;
     int i = hs.get<int>();
@@ -514,7 +518,7 @@ EX namespace mapstream {
       f.write(asonov::period_z);
       }
     #endif
-    if(prod) {
+    if(mproduct) {
       f.write(hybrid::csteps);
       f.write(product::cspin);
       f.write(product::cmirror);
@@ -523,7 +527,7 @@ EX namespace mapstream {
     if(rotspace) {
       f.write(hybrid::csteps);
       }
-    if(hybri) {
+    if(mhybrid) {
       hybrid::in_underlying_geometry([&] { save_geometry(f); });
       }
     if(fake::in()) {
@@ -657,7 +661,7 @@ EX namespace mapstream {
     if(geometry == gRotSpace && vernum >= 0xA833) {
       f.read(hybrid::csteps);
       }
-    if(hybri && vernum >= 0xA80C) {
+    if(mhybrid && vernum >= 0xA80C) {
       auto g = geometry;
       load_geometry(f);
       set_geometry(g);
@@ -680,17 +684,20 @@ EX namespace mapstream {
     geometry_settings(was_default);
     }
   
-  EX hookset<void(fhstream&)> hooks_savemap, hooks_loadmap_old;
-  EX hookset<void(fhstream&, int)> hooks_loadmap;
+  EX hookset<void(hstream&)> hooks_savemap, hooks_loadmap_old;
+  EX hookset<void(hstream&, int)> hooks_loadmap;
   
   EX cell *save_start() {
-    return (closed_manifold || euclid || prod || arcm::in() || sol || INVERSE) ? currentmap->gamestart() : cwt.at->master->c7;
+    return (closed_manifold || euclid || mproduct || arcm::in() || sol || INVERSE) ? currentmap->gamestart() : cwt.at->master->c7;
     }
 
 #if CAP_EDIT  
-  void save_only_map(fhstream& f) {
+  void save_only_map(hstream& f) {
     f.write(patterns::whichPattern);
     save_geometry(f);
+    #if CAP_RACING
+    if(racing::on) racing::restore_goals();
+    #endif
     
     // game settings
     f.write(safety);
@@ -717,7 +724,7 @@ EX namespace mapstream {
     }
     
     addToQueue(save_start());
-    #if MAXMDIM >= 4 && CAP_RAY
+    #if CAP_PORTALS
     if(intra::in) intra::prepare_need_to_save();
     #endif
     for(int i=0; i<isize(cellbyid); i++) {
@@ -752,23 +759,25 @@ EX namespace mapstream {
         f.write_char(dat.dir);
         f.write_char(dat.mirrored);
         }
-      // f.write_char(c->barleft);
-      // f.write_char(c->barright);
       f.write_char(c->item);
       if(c->item == itBabyTortoise)
         f.write(tortoise::babymap[c]);
       f.write_char(c->mpdist);
-      // f.write_char(c->bardir);
+      if(inmirrororwall(c)) {
+        f.write_char(c->barleft);
+        f.write_char(c->barright);
+        f.write_char(c->bardir);
+        }
       f.write(c->wparam); f.write(c->landparam);
       f.write_char(c->stuntime); f.write_char(c->hitpoints);
       bool blocked = false;
-      #if MAXMDIM >= 4 && CAP_RAY
+      #if CAP_PORTALS
       if(intra::in && isWall3(c) && !intra::need_to_save.count(c)) blocked = true;
       #endif
       if(!blocked)
       for(int j=0; j<c->type; j++) {
         cell *c2 = c->move(j);
-        if(c2 && c2->land != laNone) addToQueue(c2);
+        if(c2 && c2->land != laNone && c2->land != laMemory) addToQueue(c2);
         }
       }
     printf("cells saved = %d\n", isize(cellbyid));
@@ -796,7 +805,7 @@ EX namespace mapstream {
       for(int i=0; i<multi::players; i++)
         f.write(cellids[multi::player[i].at]);
 
-    #if MAXMDIM >= 4 && CAP_RAY
+    #if CAP_PORTALS
     if(intra::in) {
       for(int i=0; i<isize(intra::portals_to_save); i++) {
         auto& p = intra::portals_to_save[i];
@@ -817,6 +826,20 @@ EX namespace mapstream {
       }
     #endif
       
+    if(f.vernum >= 0xA912) {
+      #if CAP_RACING
+      f.write(racing::on);
+      if(racing::on) {
+        f.write<int>(isize(racing::track));
+        for(auto& t: racing::track) f.write<int>(cellids[t]);
+        racing::save_ghosts(f);
+        }
+      #else
+      bool on = false;
+      f.write(on);
+      #endif
+      }
+
     callhooks(hooks_savemap, f);
     f.write<int>(0);
 
@@ -824,7 +847,7 @@ EX namespace mapstream {
     cellbyid.clear();
     }
   
-  void load_usershapes(fhstream& f) {
+  void load_usershapes(hstream& f) {
     if(f.vernum >= 7400) while(true) {
       int i = f.get<int>();
       if(i == -1) break;
@@ -854,7 +877,7 @@ EX namespace mapstream {
       }    
     }
   
-  void load_only_map(fhstream& f) {
+  void load_only_map(hstream& f) {
     stop_game();
     if(f.vernum >= 10420 && f.vernum < 10503) {
       int i;
@@ -903,7 +926,7 @@ EX namespace mapstream {
       for(int k=0; k<i; k++) f.read(kills[k]);    
       }
 
-    int sub = hybri ? 2 : 0;
+    int sub = mhybrid ? 2 : 0;
     while(true) {
       cell *c;
       int rspin;
@@ -924,7 +947,7 @@ EX namespace mapstream {
         
         // spinval becomes xspinval
         rspin = gmod(c2->c.spin(dir) - f.read_char(), c->type - sub);
-        if(GDIM == 3 && rspin && !hybri) {
+        if(GDIM == 3 && rspin && !mhybrid) {
           println(hlog, "rspin in 3D");
           throw hstream_exception();
           }
@@ -953,6 +976,11 @@ EX namespace mapstream {
         f.read(tortoise::babymap[c]);
       c->mpdist = f.read_char();
       c->bardir = NOBARRIERS;
+      if(inmirrororwall(c) && f.vernum >= 0xA912) {
+        c->barleft = (eLand) f.read_char();
+        c->barright = (eLand) f.read_char();
+        c->bardir = fixspin(rspin, f.read_char(), c->type, f.vernum);
+        }
       // fixspin(rspin, f.read_char(), c->type);
       if(f.vernum < 7400) {
         short z;
@@ -1007,6 +1035,7 @@ EX namespace mapstream {
       walking::colors_of_floors.clear();
       for(auto c: v) walking::colors_of_floors.insert(c);
       }
+    else walking::colors_of_floors.clear();
 
     load_drawing_tool(f);
 
@@ -1040,7 +1069,7 @@ EX namespace mapstream {
           }
       }
 
-    #if MAXMDIM >= 4 && CAP_RAY
+    #if CAP_PORTALS
     if(intra::in) {
       while(true) {
         char k = f.get<char>();
@@ -1055,6 +1084,25 @@ EX namespace mapstream {
         }
       }
     #endif
+    
+    if(f.vernum >= 0xA912) {
+      #if CAP_RACING
+      f.read(racing::on);
+      if(racing::on) {
+        if(!shmup::on) {
+          shmup::on = true;
+          shmup::init();
+          }
+        racing::track.resize(f.get<int>());
+        for(auto& t: racing::track) t = cellbyid[f.get<int>()];
+        racing::load_ghosts(f);
+        racing::configure_track(false);
+        }
+      #else
+      bool on;
+      f.read(on);
+      #endif
+      }
 
     if(f.vernum >= 0xA848) {
       int i;
@@ -1073,7 +1121,7 @@ EX namespace mapstream {
     game_active = true;
     }
   
-  void save_usershapes(fhstream& f) {
+  void save_usershapes(hstream& f) {
     int32_t n;
     #if CAP_POLY    
     for(int i=0; i<mapeditor::USERSHAPEGROUPS; i++) for(auto usp: usershapes[i]) {
@@ -1098,16 +1146,21 @@ EX namespace mapstream {
   EX bool saveMap(const char *fname) {
     fhstream f(fname, "wb");
     if(!f.f) return false;
-    f.write(f.vernum);
+    saveMap(f);
+    return true;
+    }
+
+  EX void saveMap(hstream& f) {
+    f.write(f.get_vernum());
     f.write(dual::state);
-    #if MAXMDIM >= 4 && CAP_RAY
+    #if CAP_PORTALS
     int q = intra::in ? isize(intra::data) : 0;
     f.write(q);
     #else
     int q = 0;
     #endif
     if(q) {
-      #if MAXMDIM >= 4 && CAP_RAY
+      #if CAP_PORTALS
       intra::prepare_to_save();
       int qp = isize(intra::portals_to_save);
       f.write(qp);
@@ -1125,12 +1178,15 @@ EX namespace mapstream {
       dual::split_or_do([&] { save_only_map(f); });
       }
     save_usershapes(f);
-    return true;
     }
   
   EX bool loadMap(const string& fname) {
     fhstream f(fname, "rb");
     if(!f.f) return false;
+    return loadMap(f);
+    }
+    
+  EX bool loadMap(hstream& f) {
     f.read(f.vernum);
     if(f.vernum > 10505 && f.vernum < 11000) 
       f.vernum = 11005;
@@ -1148,7 +1204,7 @@ EX namespace mapstream {
     if(q) {
       int qp;
       f.read(qp);
-      #if MAXMDIM >= 4 && CAP_RAY
+      #if CAP_PORTALS
       intra::portals_to_save.resize(qp);
       for(auto& ps: intra::portals_to_save) {
         f.read(ps.spin);
@@ -1186,6 +1242,7 @@ EX namespace mapeditor {
 
 #if CAP_EDIT
   int paintwhat = 0;
+  int paintwhat_alt_wall = 0;
   int painttype = 0;
   int paintstatueid = 0;
   int radius = 0;
@@ -1322,7 +1379,11 @@ EX namespace mapeditor {
 
   EX void showMapEditor() {
     cmode = sm::MAP | sm::PANNING;
-    gamescreen(0);
+    if(building_mode) {
+      if(anyshiftclick) cmode |= sm::EDIT_INSIDE_WALLS;
+      else cmode |= sm::EDIT_BEFORE_WALLS;
+      }
+    gamescreen();
   
     int fs = editor_fsize();
     
@@ -1335,7 +1396,14 @@ EX namespace mapeditor {
     
     getcstat = '-';
 
-    displayfr(8, 8 + fs, 2, vid.fsize, paintwhat_str, forecolor, 0);
+    if(anyshiftclick) {
+      displayfr(8, 8 + fs, 2, vid.fsize,
+        (painttype == 6 && (GDIM == 3)) ? "wall" :
+        painttype == 3 ? XLATN(winf[paintwhat_alt_wall].name) : "clear",
+        forecolor, 0);
+      }
+    else
+      displayfr(8, 8 + fs, 2, vid.fsize, paintwhat_str, forecolor, 0);
     displayfr(8, 8+fs*2, 2, vid.fsize, XLAT("use at your own risk!"), 0x800000, 0);
 
     displayButton(8, 8+fs*4, XLAT("0-9 = radius (%1)", its(radius)), ('0' + (radius+1)%10), 0);
@@ -1350,6 +1418,10 @@ EX namespace mapeditor {
       displayButton(8, 8+fs*12, XLAT("f = flip %1", ONOFF(copysource.mirrored)), 'u', 0);
     displayButton(8, 8+fs*13, XLAT("r = regular"), 'r', 0);
     displayButton(8, 8+fs*14, XLAT("p = paint"), 'p', 0);
+    if(painttype == 3)
+      displayButton(8, 8+fs*15, XLAT("z = set Shift+click"), 'z', 0);
+    if(WDIM == 3)
+      displayButton(8, 8+fs*16, XLAT("B = build on walls ") + ONOFF(building_mode), 'B', 0);
 
     displayFunctionKeys();
     displayButton(8, vid.yres-8-fs*4, XLAT("F8 = settings"), SDLK_F8, 0);
@@ -1391,6 +1463,7 @@ EX namespace mapeditor {
     saveUndo(c);
     switch(painttype) {
       case 0: {
+        if(anyshiftclick) { c->monst = moNone; mirror::destroyKilled(); break; }
         eMonster last = c->monst;
         c->monst = eMonster(paintwhat);
         c->hitpoints = 3;
@@ -1421,6 +1494,7 @@ EX namespace mapeditor {
         break;
         }
       case 1: {
+        if(anyshiftclick) { c->item = itNone; break; }
         eItem last = c->item;
         c->item = eItem(paintwhat);
         if(c->item == itBabyTortoise)
@@ -1428,6 +1502,7 @@ EX namespace mapeditor {
         break;
         }
       case 2: {
+        if(anyshiftclick) { c->land = laNone; c->wall = waNone; map_version++; break; }
         eLand last = c->land;
         c->land = eLand(paintwhat);
         if(isIcyLand(c) && isIcyLand(last))
@@ -1444,7 +1519,8 @@ EX namespace mapeditor {
         }
       case 3: {
         eWall last = c->wall;
-        c->wall = eWall(paintwhat);
+        c->wall = eWall(anyshiftclick ? paintwhat_alt_wall : paintwhat);
+        map_version++;
         
         if(last != c->wall) {
           if(hasTimeout(c))
@@ -1467,6 +1543,7 @@ EX namespace mapeditor {
         break;
         }
       case 5:
+        map_version++;
         c->land = laNone;
         c->wall = waNone;
         c->item = itNone;
@@ -1475,11 +1552,13 @@ EX namespace mapeditor {
         // c->tmp = -1;
         break;
       case 6:
+        map_version++;
         c->land = laCanvas;
-        c->wall = GDIM == 3 ? waWaxWall : waNone;
+        c->wall = ((GDIM == 3) ^ anyshiftclick) ? waWaxWall : waNone;
         c->landparam = paintwhat >> 8;
         break;
       case 4: {
+        map_version++;
         cell *copywhat = where.second.at;
         c->wall = copywhat->wall;
         c->item = copywhat->item;
@@ -1549,7 +1628,11 @@ EX namespace mapeditor {
   void allInPattern(cellwalker where) {
 
     manual_celllister cl;
-    if(!patterns::whichPattern) {
+    bool call_editAt = !patterns::whichPattern;
+#if CAP_TEXTURE
+    call_editAt |= (texture::config.tstate == texture::tsActive);
+#endif
+    if (call_editAt) {
       editAt(where, cl);
       return;
       }
@@ -1621,10 +1704,12 @@ EX namespace mapeditor {
     }
   
   void showList() {
+    string caption;
     dialog::v.clear();
     if(painttype == 4) painttype = 0;
     switch(painttype) {
       case 0: 
+        caption = "monsters";
         for(int i=0; i<motypes; i++) {
           eMonster m = eMonster(i);
           if(
@@ -1639,12 +1724,15 @@ EX namespace mapeditor {
           }
         break;
       case 1:
+        caption = "items";
         for(int i=0; i<ittypes; i++) dialog::vpush(i, iinf[i].name);
         break;
       case 2:
+        caption = "lands";
         for(int i=0; i<landtypes; i++) dialog::vpush(i, linf[i].name);
         break;
       case 3:
+        caption = "walls";
         for(int i=0; i<walltypes; i++) if(i != waChasmD) dialog::vpush(i, winf[i].name);
         break;
       }
@@ -1652,26 +1740,18 @@ EX namespace mapeditor {
     
     if(dialog::infix != "") mouseovers = dialog::infix;
     
-    int q = dialog::v.size();
-    int percolumn = vid.yres / (vid.fsize+5) - 4;
-    int columns = 1 + (q-1) / percolumn;
+    cmode = 0;
+    gamescreen();
+    dialog::init(caption);
+    if(dialog::infix != "") mouseovers = dialog::infix;
+    dialog::addBreak(50);
+    dialog::start_list(900, 900, '1');
     
-    for(int i=0; i<q; i++) {
-      int x = 16 + (vid.xres * (i/percolumn)) / columns;
-      int y = (vid.fsize+5) * (i % percolumn) + vid.fsize*2;
-      
-      int actkey = 1000 + i;
-      string vv = dialog::v[i].first;
-      if(i < 9) { vv += ": "; vv += ('1' + i); }
-      
-      displayButton(x, y, vv, actkey, 0);
-      }
-    keyhandler = [] (int sym, int uni) {
-      if(uni >= '1' && uni <= '9') uni = 1000 + uni - '1';
-      if(sym == SDLK_RETURN || sym == SDLK_KP_ENTER || sym == '-' || sym == SDLK_KP_MINUS) uni = 1000;
-      for(int z=0; z<isize(dialog::v); z++) if(1000 + z == uni) {
-        paintwhat = dialog::v[z].second;
-        paintwhat_str = dialog::v[z].first;
+    for(auto& vi: dialog::v) {
+      dialog::addItem(vi.first, dialog::list_fake_key++);
+      dialog::add_action([&vi] {
+        paintwhat = vi.second;
+        paintwhat_str = vi.first;
 
         mousepressed = false;
         popScreen();
@@ -1680,9 +1760,18 @@ EX namespace mapeditor {
           dialog::editNumber(paintstatueid, 0, 127, 1, 1, XLAT1("editable statue"), 
             XLAT("These statues are designed to have their graphics edited in the Vector Graphics Editor. Each number has its own, separate graphics.")
             );
-        return;
-        }
-      if(dialog::editInfix(uni)) ;
+        });
+      }
+
+    dialog::end_list();
+    dialog::addBreak(50);
+    dialog::addInfo(XLAT("press letters to search"));
+    dialog::addBack();
+    dialog::display();
+
+    keyhandler = [] (int sym, int uni) {
+      dialog::handleNavigation(sym, uni);
+      if(dialog::editInfix(uni)) dialog::list_skip = 0;
       else if(doexiton(sym, uni)) popScreen();
       };    
     }
@@ -1694,17 +1783,21 @@ EX namespace mapeditor {
     if(uni == '-' && !holdmouse) undoLock();
     if(uni == '-' && mouseover) {
       allInPattern(mouseover_cw(false));
-      holdmouse = true;
+      if(!(GDIM == 3 && building_mode))
+        holdmouse = true;
       }
     
     if(mouseover) for(int i=0; i<mouseover->type; i++) createMov(mouseover, i);
     if(uni == 'u') applyUndo();
     else if(uni == 'v' || sym == SDLK_F10 || sym == SDLK_ESCAPE) popScreen();
+    else if(uni == 'A') popScreen();
     else if(uni >= '0' && uni <= '9') radius = uni - '0';
     else if(uni == 'm') pushScreen(showList), painttype = 0, dialog::infix = "";
     else if(uni == 'i') pushScreen(showList), painttype = 1, dialog::infix = "";
     else if(uni == 'l') pushScreen(showList), painttype = 2, dialog::infix = "";
     else if(uni == 'w') pushScreen(showList), painttype = 3, dialog::infix = "";
+    else if(uni == 'z' && painttype == 3) paintwhat_alt_wall = paintwhat;
+    else if(uni == 'B') building_mode = !building_mode;
     else if(uni == 'r') pushScreen(patterns::showPattern);
     else if(uni == 't' && mouseover) {
       playermoved = true;
@@ -1890,18 +1983,18 @@ EX namespace mapeditor {
         }
       transmatrix T;
       if(front_config == eFront::equidistants) T = Id;
-      else if(front_config == eFront::const_x) T = cspin(2, 0, M_PI/2);
-      else if(front_config == eFront::const_y) T = cspin(2, 1, M_PI/2);
+      else if(front_config == eFront::const_x) T = cspin90(2, 0);
+      else if(front_config == eFront::const_y) T = cspin90(2, 1);
       else return;
       for(int i=0; i<4; i+=2) {
         for(int u=2; u<=20; u++) {
           PRING(d) {
-            curvepoint(T * xspinpush(M_PI*d/cgi.S42, u/20.) * zpush0(front_edit));
+            curvepoint(T * xspinpush(d*cgi.S_step, u/20.) * zpush0(front_edit));
             }
           queuecurve(d2, cols[i + (u%5 != 0)], 0, i < 2 ? PPR::LINE : PPR::SUPERLINE);
           }
         for(int d=0; d<cgi.S84; d++) {
-          for(int u=0; u<=20; u++) curvepoint(T * xspinpush(M_PI*d/cgi.S42, u/20.) * zpush(front_edit) * C0);
+          for(int u=0; u<=20; u++) curvepoint(T * xspinpush(d*cgi.S_step, u/20.) * zpush(front_edit) * C0);
           queuecurve(d2, cols[i + (d % (cgi.S84/drawcell->type) != 0)], 0, i < 2 ? PPR::LINE : PPR::SUPERLINE);
           }
         }
@@ -1910,11 +2003,11 @@ EX namespace mapeditor {
 
     for(int d=0; d<cgi.S84; d++) {
       unsigned col = (d % (cgi.S84/drawcell->type) == 0) ? gridcolor : lightgrid;
-      queueline(d2 * C0, d2 * xspinpush0(M_PI*d/cgi.S42, 1), col, 4 + vid.linequality);
+      queueline(d2 * C0, d2 * xspinpush0(d*cgi.S_step, 1), col, 4 + vid.linequality);
       }
     for(int u=2; u<=20; u++) {
       PRING(d) {
-        curvepoint(xspinpush0(M_PI*d/cgi.S42, u/20.));
+        curvepoint(xspinpush0(d*cgi.S_step, u/20.));
         }
       queuecurve(d2, (u%5==0) ? gridcolor : lightgrid, 0, PPR::LINE);
       }
@@ -1968,8 +2061,7 @@ EX namespace mapeditor {
         // total: rh2 - rh1
         // ld z = degree;
         ld x = b2 - b1 + M_PI;
-        while(x > M_PI) x -= 2 * M_PI;
-        while(x < -M_PI) x += 2 * M_PI;
+        cyclefix(x, 0);
         area += x;
         }
       }
@@ -2002,7 +2094,7 @@ EX namespace mapeditor {
   EX void showDrawEditor() {
 #if CAP_POLY
     cmode = sm::DRAW | sm::PANNING;
-    gamescreen(0);
+    gamescreen();
     drawGrid();
     if(callhandlers(false, hooks_prestats)) return;
 
@@ -2331,7 +2423,7 @@ EX namespace mapeditor {
           }
         if(uni == 'c') dsCur->list.push_back(best);
         else if(uni == 'd') {
-          vector<hyperpoint> oldlist = move(dsCur->list);
+          vector<hyperpoint> oldlist = std::move(dsCur->list);
           dsCur->list.clear();
           int i;
           for(i=0; i<isize(oldlist); i+=3)
@@ -2351,7 +2443,7 @@ EX namespace mapeditor {
       }
     
     if(uni == 'a' && haveshape) {
-      mh = spin(2*M_PI*-ew.rotid/dsCur->rots) * mh;
+      mh = spin(TAU*-ew.rotid/dsCur->rots) * mh;
       if(ew.symid) mh = Mirror * mh;
     
       if(ew.pointid < 0 || ew.pointid >= isize(dsCur->list)) 
@@ -2373,7 +2465,7 @@ EX namespace mapeditor {
 
       if(i < 0 || i >= isize(dsCur->list)) return;
 
-      mh = spin(2*M_PI*-ew.rotid/dsCur->rots) * mh;
+      mh = spin(TAU*-ew.rotid/dsCur->rots) * mh;
       if(ew.symid) mh = Mirror * mh;
 
       if(uni == 'm' || uni == 'M') 
@@ -2690,6 +2782,7 @@ EX namespace mapeditor {
       }
 
     if(sym == SDLK_F10) popScreen();
+    else if(uni == 'A') popScreen();
 
     (void)clickused;
     
@@ -2749,7 +2842,7 @@ EX namespace mapeditor {
           shiftmatrix T = rgpushxto0(lstart);
           texture::where = lstartcell;
           for(int i=0; i<circp; i++)
-            texture::drawPixel(T * xspinpush0(2 * M_PI * i / circp, rad), tcolor);
+            texture::drawPixel(T * xspinpush0(TAU * i / circp, rad), tcolor);
           lstartcell = NULL;
           }
 #endif          
@@ -2833,6 +2926,32 @@ EX namespace mapeditor {
             println(hlog);
             }
           }
+
+        for(int i=0; i<USERSHAPEGROUPS; i++) for(auto usp: usershapes[i]) {
+          auto us = usp.second;
+          if(!us) continue;
+
+          for(int l=0; l<USERLAYERS; l++) if(isize(us->d[l].list)) {
+            usershapelayer& ds(us->d[l]);
+            println(hlog, spaced("//", i, usp.first, l, "[", ds.color, double(ds.zlevel), "]"));
+            print(hlog, "{");
+
+            for(int r=0; r<us->d[l].rots; r++) {
+              for(int i=0; i<isize(us->d[l].list); i++) {
+                hyperpoint h = us->d[l].list[i];
+                h = spin(TAU * r / us->d[l].rots) * h;
+                for(int d=0; d<GDIM; d++) print(hlog, fts(h[d]), ", ");
+                }
+              if(us->d[l].sym) for(int i=isize(us->d[l].list)-1; i>=0; i--) {
+                hyperpoint h = us->d[l].list[i];
+                h[1] = -h[1];
+                h = spin(TAU * r / us->d[l].rots) * h;
+                for(int d=0; d<GDIM; d++) print(hlog, fts(h[d]), ", ");
+                }
+              }
+            println(hlog, "}, ");
+            }
+          }
         }
   
       if(uni == 'p') {
@@ -2906,7 +3025,7 @@ EX namespace mapeditor {
     if(radius > .1) circp *= 2; 
     
     for(int j=0; j<circp; j++)
-      pts.push_back(xspinpush0(M_PI*j*2/circp, radius));
+      pts.push_back(xspinpush0(TAU * j / circp, radius));
     for(int j=0; j<circp; j++) curvepoint(pts[j]);
     curvepoint(pts[0]);
     queuecurve(Ctr, dtcolor, 0, PPR::LINE);
@@ -2941,7 +3060,7 @@ EX namespace mapeditor {
 
       for(int j=0; j<=texture::texturesym; j++)
       for(int i=0; i<c->type; i += sih.symmetries) {
-        shiftmatrix M2 = V * applyPatterndir(c, sih) * spin(2*M_PI*i/c->type);
+        shiftmatrix M2 = V * applyPatterndir(c, sih) * spin(TAU * i / c->type);
         if(j) M2 = M2 * Mirror;
         switch(holdmouse ? mousekey : 'd') {
           case 'c':
@@ -2977,7 +3096,8 @@ EX namespace mapeditor {
         hpcshape& sh(cgi.ushr[&ds]);
     
         if(sh.s != sh.e) {
-          auto& last = queuepolyat(mmscale(V, GDIM == 3 ? 0 : geom3::lev_to_factor(ds.zlevel)), sh, ds.color ? ds.color : color, prio);
+          shiftmatrix V1 = GDIM == 3 ? V : orthogonal_move(V, ds.zlevel);
+          auto& last = queuepolyat(V1, sh, ds.color ? ds.color : color, prio);
           if(GDIM == 3) {
             last.tinf = &user_triangles_texture;
             last.offset_texture = ds.texture_offset;
@@ -3025,14 +3145,14 @@ EX namespace mapeditor {
     
           if(mouseout()) break;
     
-          shiftpoint P2 = V * spin(2*M_PI*a/ds.rots) * (b?Mirror*mh:mh);
+          shiftpoint P2 = V * spin(TAU * a / ds.rots) * (b?Mirror*mh:mh);
         
           queuestr(P2, 10, "x", 0xFF00FF);
           }
         
         if(isize(ds.list) == 0) return us;
         
-        shiftpoint Plast = V * spin(-2*M_PI/ds.rots) * (ds.sym?Mirror*ds.list[0]:ds.list[isize(ds.list)-1]);
+        shiftpoint Plast = V * spin(-TAU / ds.rots) * (ds.sym?Mirror*ds.list[0]:ds.list[isize(ds.list)-1]);
         int state = 0;
         int gstate = 0;
         double dist2 = 0;
@@ -3041,14 +3161,14 @@ EX namespace mapeditor {
         for(int a=0; a<ds.rots; a++) 
         for(int b=0; b<(ds.sym?2:1); b++) {
         
-          hyperpoint mh2 = spin(2*M_PI*-ew.rotid/ds.rots) * mh;
+          hyperpoint mh2 = spin(TAU * -ew.rotid/ds.rots) * mh;
           if(ew.symid) mh2 = Mirror * mh2;
-          shiftpoint pseudomouse = V * spin(2*M_PI*a/ds.rots) * mirrorif(mh2, b);      
+          shiftpoint pseudomouse = V * spin(TAU * a / ds.rots) * mirrorif(mh2, b);
         
           for(int t=0; t<isize(ds.list); t++) {
             int ti = b ? isize(ds.list)-1-t : t;
   
-            shiftpoint P2 = V * spin(2*M_PI*a/ds.rots) * mirrorif(ds.list[ti], b);
+            shiftpoint P2 = V * spin(TAU * a / ds.rots) * mirrorif(ds.list[ti], b);
             
             if(!mouseout()) {
               double d = hdist(moh, P2);
@@ -3118,7 +3238,7 @@ EX namespace mapeditor {
 
   EX void map_settings() {
     cmode = sm::SIDE | sm::MAYDARK;
-    gamescreen(1);
+    gamescreen();
   
     dialog::init(XLAT("Map settings"));
   
@@ -3138,7 +3258,22 @@ EX namespace mapeditor {
     dialog::addBoolItem(XLAT("god mode"), autocheat, 'G');
     dialog::add_action([] () { autocheat = true; });
     dialog::addInfo(XLAT("(unlock all, allow cheats, normal character display, cannot be turned off!)"));
+
+    add_edit(game_keys_scroll);
+    dialog::addInfo(XLAT("hint: shift+A to enter the map editor"));
     
+    #if CAP_PORTALS
+    if(WDIM == 3 && !intra::in) {
+      dialog::addBoolItem(XLAT("become a portal map"), intra::in, 'm');
+      dialog::add_action_push(intra::become_menu);
+      }
+
+    if(WDIM == 3 && intra::in) {
+      dialog::addItem(XLAT("manage portals"), 'm');
+      dialog::add_action_push(intra::show_portals);
+      }
+    #endif
+
     dialog::addItem(XLAT("change the pattern/color of new Canvas cells"), 'c');
     dialog::add_action_push(patterns::showPrePatternNoninstant);
 
