@@ -108,8 +108,8 @@ EX namespace polygonal {
   EX }
 
 #if HDR
-inline bool mdAzimuthalEqui() { return among(pmodel, mdEquidistant, mdEquiarea, mdEquivolume); }
-inline bool mdBandAny() { return mdinf[pmodel].flags & mf::pseudoband; }
+inline bool mdAzimuthalEqui() { return (mdinf[pmodel].flags & mf::azimuthal) && (mdinf[pmodel].flags & (mf::equidistant | mf::equiarea | mf::equivolume) && !(mdinf[pmodel].flags & mf::twopoint)); }
+inline bool mdBandAny() { return mdinf[pmodel].flags & mf::pseudocylindrical; }
 inline bool mdPseudocylindrical() { return mdBandAny() && !(mdinf[pmodel].flags & mf::cylindrical); }
 #endif
 
@@ -135,7 +135,7 @@ EX namespace models {
   #endif
 
   EX transmatrix rotmatrix() {
-    if(GDIM == 2 || prod) return spin(rotation * degree);
+    if(GDIM == 2 || gproduct) return spin(rotation * degree);
     return spin(rotation_xy2 * degree) * cspin(0, 2, -rotation_xz * degree) * spin(rotation * degree);
     }
   
@@ -176,7 +176,7 @@ EX namespace models {
       euclidean_spin = pispin * iso_inverse(cview().T * currentmap->master_relative(centerover, true));
       euclidean_spin = gpushxto0(euclidean_spin * C0) * euclidean_spin;
       hyperpoint h = inverse(euclidean_spin) * (C0 + (euc::eumove(gp::loc{1,0})*C0 - C0) * vpconf.spiral_x + (euc::eumove(gp::loc{0,1})*C0 - C0) * vpconf.spiral_y);
-      spiral_multiplier = cld(0, 2 * M_PI) / cld(h[0], h[1]);
+      spiral_multiplier = cld(0, TAU) / cld(h[0], h[1]);
       }
     
     if(centerover && !history::on)
@@ -185,23 +185,31 @@ EX namespace models {
       }
     }
   
+  /** mdRelPerspective and mdRelOrthogonal in hyperbolic space only make sense if it is actually a de Sitter visualization */
+  EX bool desitter_projections;
+
+  EX vector<bool_reaction_t> avail_checkers;
+
   EX bool model_available(eModel pm) {
+    if(pm < isize(avail_checkers) && avail_checkers[pm]) return avail_checkers[pm]();
     if(mdinf[pm].flags & mf::technical) return false;
-    if(prod) {
+    if(gproduct) {
       if(pm == mdPerspective) return true;
       if(among(pm, mdBall, mdHemisphere)) return false;
       return PIU(model_available(pm));
       }
-    if(sl2) return pm == mdGeodesic;
+    if(hyperbolic && desitter_projections && among(pm, mdRelPerspective, mdRelOrthogonal)) return true;
+    if(sl2) return among(pm, mdGeodesic, mdEquidistant, mdRelPerspective, mdRelOrthogonal, mdHorocyclic, mdPerspective);
+    if(among(pm, mdRelOrthogonal, mdRelPerspective)) return false;
     if(nonisotropic) return among(pm, mdDisk, mdPerspective, mdHorocyclic, mdGeodesic, mdEquidistant, mdFisheye, mdLiePerspective, mdLieOrthogonal);
-    if(sphere && (pm == mdHalfplane || pm == mdBall))
-      return false;
+    if(sphere && pm == mdBall) return false;
+    if(sphere && (mdinf[pm].flags & mf::horocyclic)) return false;
     if(GDIM == 2 && is_perspective(pm)) return false;
     if(pm == mdGeodesic && !nonisotropic) return false;
     if(pm == mdLiePerspective && sphere) return false;
     if(pm == mdLieOrthogonal && sphere) return false;
     if(GDIM == 2 && pm == mdEquivolume) return false;
-    if(pm == mdThreePoint && !(GDIM == 3 && !nonisotropic && !prod)) return false;
+    if(pm == mdThreePoint && !(GDIM == 3 && !nonisotropic && !gproduct)) return false;
     if(GDIM == 3 && among(pm, mdBall, mdHyperboloid, mdFormula, mdPolygonal, mdRotatedHyperboles, mdSpiral, mdHemisphere)) return false;
     if(pm == mdCentralInversion && !euclid) return false;
     if(pm == mdPoorMan) return hyperbolic;
@@ -210,17 +218,14 @@ EX namespace models {
     }    
   
   EX bool has_orientation(eModel m) {
-    if(among(m, mdHorocyclic, mdLieOrthogonal, mdLiePerspective))
-      return hyperbolic || in_h2xe();
-    if((m == mdPerspective || m == mdGeodesic) && panini_alpha) return true;
-    return
-      among(m, mdHalfplane, mdPolynomial, mdPolygonal, mdTwoPoint, mdJoukowsky, mdJoukowskyInverted, mdSpiral, mdSimulatedPerspective, mdTwoHybrid, mdHorocyclic, mdAxial, mdAntiAxial, mdQuadrant,
-        mdWerner, mdAitoff, mdHammer, mdLoximuthal, mdWinkelTripel, mdThreePoint) || mdBandAny();
+    if(is_perspective(m) && panini_alpha) return true;
+    if(nonisotropic) return false;    
+    return (mdinf[m].flags & mf::orientation);
     }
 
   /** @brief returns the broken coordinate, or zero */
   EX int get_broken_coord(eModel m) {
-    if(m == mdWerner) return 1;
+    if(mdinf[m].flags & mf::werner) return 1;
     if(sphere) return (mdinf[m].flags & mf::broken) ? 2 : 0;
     return 0;
     }
@@ -230,7 +235,7 @@ EX namespace models {
     }
   
   EX bool is_perspective(eModel m) {
-    return among(m, mdPerspective, mdGeodesic, mdLiePerspective);
+    return mdinf[m].flags & mf::perspective;
     }
 
   EX bool is_3d(const projection_configuration& p) {
@@ -239,12 +244,12 @@ EX namespace models {
     }
   
   EX bool has_transition(eModel m) {
-    return among(m, mdJoukowsky, mdJoukowskyInverted, mdBand, mdAxial) && GDIM == 2;
+    return (mdinf[m].flags & mf::transition) && GDIM == 2;
     }
   
   EX bool product_model(eModel m) {
-    if(!prod) return false;
-    if(among(m, mdPerspective, mdHyperboloid, mdEquidistant, mdThreePoint)) return false;
+    if(!gproduct) return false;
+    if(mdinf[m].flags & mf::product_special) return false;
     return true;
     }
   
@@ -252,13 +257,13 @@ EX namespace models {
   
   EX string get_model_name(eModel m) {
     if(m == mdDisk && GDIM == 3 && (hyperbolic || nonisotropic)) return XLAT("ball model/Gans");
-    if(m == mdPerspective && prod) return XLAT("native perspective");
-    if(prod) return PIU(get_model_name(m));
+    if(m == mdPerspective && gproduct) return XLAT("native perspective");
+    if(gproduct) return PIU(get_model_name(m));
     if(nonisotropic) {
       if(m == mdHorocyclic && !sol) return XLAT("simple model: projection");
       if(m == mdPerspective) return XLAT("simple model: perspective");
       if(m == mdGeodesic) return XLAT("native perspective");
-      if(among(m, mdEquidistant, mdFisheye, mdHorocyclic, mdLiePerspective, mdLieOrthogonal)) return XLAT(mdinf[m].name_hyperbolic);
+      if(among(m, mdEquidistant, mdFisheye, mdHorocyclic, mdLiePerspective, mdLieOrthogonal, mdRelPerspective, mdRelOrthogonal)) return XLAT(mdinf[m].name_hyperbolic);
       }
     if(m == mdDisk && GDIM == 3) return XLAT("perspective in 4D");
     if(m == mdHalfplane && GDIM == 3 && hyperbolic) return XLAT("half-space");
@@ -266,7 +271,7 @@ EX namespace models {
       return XLAT(mdinf[m].name_spherical);
     if(euclid) 
       return XLAT(mdinf[m].name_euclidean);
-    if(hyperbolic) 
+    if(hyperbolic)
       return XLAT(mdinf[m].name_hyperbolic);
     return "?";
     }
@@ -319,11 +324,11 @@ EX namespace models {
       initquickqueue();
       queuereset(mdPixel, PPR::LINE);              
       for(int a=-1; a<=1; a++) {
-        curvepoint(point2(-M_PI/2 * current_display->radius, a*current_display->radius));
-        curvepoint(point2(+M_PI/2 * current_display->radius, a*current_display->radius));
+        curvepoint(point2(-90._deg * current_display->radius, a*current_display->radius));
+        curvepoint(point2(+90._deg * current_display->radius, a*current_display->radius));
         queuecurve(shiftless(Id), forecolor, 0, PPR::LINE);
-        curvepoint(point2(a*current_display->radius, -M_PI/2*current_display->radius));
-        curvepoint(point2(a*current_display->radius, +M_PI/2*current_display->radius));
+        curvepoint(point2(a*current_display->radius, -90._deg * current_display->radius));
+        curvepoint(point2(a*current_display->radius, +90._deg * current_display->radius));
         queuecurve(shiftless(Id), forecolor, 0, PPR::LINE);
         }
       queuereset(vpconf.model, PPR::LINE);
@@ -361,17 +366,18 @@ EX namespace models {
 
   EX void model_list() {
     cmode = sm::SIDE | sm::MAYDARK | sm::CENTER;
-    gamescreen(0);
+    gamescreen();
     dialog::init(XLAT("models & projections"));
     #if CAP_RUG
     USING_NATIVE_GEOMETRY_IN_RUG;
     #endif
 
+    dialog::start_list(2000, 2000, 'a');
     for(int i=0; i<isize(mdinf); i++) {
       eModel m = eModel(i);
       if(m == mdFormula && ISMOBILE) continue;
       if(model_available(m)) {
-        dialog::addBoolItem(get_model_name(m), vpconf.model == m, (i < 26 ? 'a'+i : 'A'+i-26));
+        dialog::addBoolItem(get_model_name(m), vpconf.model == m, dialog::list_fake_key++);
         dialog::add_action([m] () {
           if(m == mdFormula) {
             edit_formula();
@@ -389,6 +395,9 @@ EX namespace models {
         }
       }
     
+    dialog::end_list();
+    dialog::addBreak(100);
+    dialog::addBack();
     dialog::display();
     }
       
@@ -417,7 +426,7 @@ EX namespace models {
 
   EX void model_menu() {
     cmode = sm::SIDE | sm::MAYDARK | sm::CENTER;
-    gamescreen(0);
+    gamescreen();
     #if CAP_RUG
     USING_NATIVE_GEOMETRY_IN_RUG;
     #endif
@@ -616,12 +625,15 @@ EX namespace models {
     if(vpmodel == mdJoukowskyInverted)
       add_edit(vpconf.dualfocus_autoscale);
     
-    if(vpmodel == mdHemisphere && euclid) 
+    if(vpmodel == mdHemisphere && euclid)
       add_edit(vpconf.euclid_to_sphere);
       
-    if(among(vpmodel, mdTwoPoint, mdSimulatedPerspective, mdTwoHybrid, mdThreePoint)) 
+    if(mdinf[vpmodel].flags & mf::twopoint)
       add_edit(vpconf.twopoint_param);
     
+    if(mdinf[vpmodel].flags & mf::axial)
+      add_edit(vpconf.axial_angle);
+
     if(vpmodel == mdFisheye) 
       add_edit(vpconf.fisheye_param);
 
@@ -704,8 +716,8 @@ EX namespace models {
     }
     
   EX void quick_model() {
-    cmode = sm::CENTER;
-    gamescreen(1);
+    cmode = sm::CENTER | sm::SIDE | sm::MAYDARK;
+    gamescreen();
     dialog::init("models & projections");
     
     if(GDIM == 2 && !euclid) {
@@ -996,8 +1008,12 @@ EX namespace models {
 
       param_f(p.twopoint_param, pp+"twopoint", sp+"twopoint parameter", 1)
       -> editable(1e-3, 10, .1, "two-point parameter", "In two-point-based models, this parameter gives the distance from each of the two points to the center.", 'b')
-      -> set_sets(dialog::scaleLog)
-;
+      -> set_sets(dialog::scaleLog);
+
+      param_f(p.axial_angle, pp+"axial", sp+"axial angle", 90)
+      -> editable(1e-3, 10, .1, "angle between the axes", "In two-axe-based models, this parameter gives the angle between the two axes.", 'x')
+      -> set_sets(dialog::scaleLog);
+
       param_f(p.fisheye_param, pp+"fisheye", sp+"fisheye parameter", 1)
       -> editable(1e-3, 10, .1, "fisheye parameter", "Size of the fish eye.", 'b')
       -> set_sets(dialog::scaleLog);
@@ -1027,7 +1043,7 @@ EX namespace models {
           "Here you can change this parameter.", 'b');
       param_f(p.miller_parameter, sp+"miller");
       param_f(p.loximuthal_parameter, sp+"loximuthal")
-      -> editable(-M_PI/2, M_PI/2, .1, "loximuthal parameter", 
+      -> editable(-90._deg, 90._deg, .1, "loximuthal parameter",
           "Loximuthal is similar to azimuthal equidistant, but based on loxodromes (lines of constant geographic direction) rather than geodesics. "
           "The loximuthal projection maps (the shortest) loxodromes to straight lines of the same length, going through the starting point. "
           "This setting changes the latitude of the starting point.\n\n"
