@@ -218,7 +218,7 @@ void wanderingZebra(cell *start) {
       return;
       }
     int q = 0;
-    cell *ctab[8];
+    cell *ctab[FULL_EDGE];
     for(int i=0; i<c->type; i++) {
       cell *c3 = c->move(i);
       if(c3 && c3 != c2 && c3->land == laZebra && c3->wall == waNone)
@@ -327,7 +327,6 @@ EX void wandering() {
   if(!canmove) return;
   if(!gen_wandering) return;
   if(racing::on) return;
-  if(dpgen::in) return;
   if(items[itOrbSafety]) return;
   pathdata pd(moYeti);
   int seepcount = getSeepcount();
@@ -797,7 +796,7 @@ EX void generateSnake(cell *c, int i, int snakecolor) {
     c2 = c3;
     c2->monst = moHexSnakeTail; c2->hitpoints = snakecolor;
     int t = c2->type;
-    if(hybri) t -= 2;
+    if(mhybrid) t -= 2;
     i = (j + (t%4 == 0 ? t/2 : (len%2 ? 2 : t - 2))) % t;
     createMov(c2, i);
     if(!inpair(c2->move(i), cpair)) {
