@@ -24,6 +24,7 @@ grow_vector<vector<cell*>> movesofgood;
 EX vector<pair<cell*, int> > butterflies;
 
 EX void addButterfly(cell *c) {
+  if(shmup::on) return;
   for(int i=0; i<isize(butterflies); i++)
     if(butterflies[i].first == c) {
       butterflies[i].second = 0;
@@ -817,6 +818,7 @@ EX void moveWorm(cell *c) {
     while(c2->mondir != NODIR) {
       allcells.push_back(c2);
       c2 = c2->move(c2->mondir);
+      if(!c2) { allcells.pop_back(); break; }
       }
     allcells.push_back(c2);
     for(int i=isize(allcells)-2; i>=0; i--) {
@@ -865,7 +867,7 @@ EX void moveWorm(cell *c) {
       }
     eItem loc = treasureType(c->land);
     bool spiceSeen = false;
-    while(c->monst == moWorm || c->monst == moWormtail || c->monst == moTentacle || c->monst == moTentacletail) {
+    while(c && (c->monst == moWorm || c->monst == moWormtail || c->monst == moTentacle || c->monst == moTentacletail)) {
       // if(!id) 
       explodeAround(c);
       drawParticles(c, minf[c->monst].color, 16);
@@ -981,7 +983,7 @@ EX void removeIvy(cell *c) {
   c->monst = moNone; // NEWYEARFIX
   for(int i=0; i<c->type; i++)
   // note that semi-vines don't count
-    if(c->move(i)->wall == waVinePlant) {
+    if(c->move(i) && c->move(i)->wall == waVinePlant) {
       destroyHalfvine(c);
       if (!do_not_touch_this_wall(c))
         c->wall = waVinePlant;
@@ -1134,7 +1136,7 @@ EX void groupmove2(const movei& mi, eMonster movtype, flagtype mf) {
     if((mf & MF_ONLYEAGLE) && bird_disruption(c) && markOrb(itOrbGravity)) return;
     // in the gravity lands, eagles cannot ascend in their second move
     if((mf & MF_ONLYEAGLE) && gravityLevelDiff(c, from) < 0) {
-      onpath(c, 0);
+      onpath_mark(c);
       return;
       }
     if((mf & MF_NOFRIEND) && isFriendly(c)) return;
@@ -1149,12 +1151,12 @@ EX void groupmove2(const movei& mi, eMonster movtype, flagtype mf) {
       if(c->move(j) && canAttack(c, c->monst, c->move(j), c->move(j)->monst, af)) {
         attackMonster(c->move(j), AF_NORMAL | AF_GETPLAYER | AF_MSG, c->monst);
         animateAttack(movei(c, j), LAYER_SMALL);
-        onpath(c, 0);
+        onpath_mark(c);
         // XLATC eagle
         return;
         }
     
-    if(from->cpdist == 0 || from->monst) { onpath(c, 0); return; }
+    if(from->cpdist == 0 || from->monst) { onpath_mark(c); return; }
     
     if(movtype == moDragonHead) {
       dragon::move(mi);
@@ -1163,16 +1165,16 @@ EX void groupmove2(const movei& mi, eMonster movtype, flagtype mf) {
     
     moveMonster(mi);
     
-    onpath(from, 0);
+    onpath_mark(from);
 
     if(isDie(mi.t->monst)) {
       /* other dice will not pathfind through the original cell */
       /* this makes it easier for the player to roll dice correctly */
-      onpath(c, 0);
+      onpath_mark(c);
       return;
       }
     }
-  onpath(c, 0);
+  onpath_mark(c);
   // MAXGCELL
   if(isize(gendfs) < 1000 || c->cpdist <= 6) gendfs.push_back(c);
   }
@@ -1231,7 +1233,7 @@ EX void groupmove(eMonster movtype, flagtype mf) {
     if((mf & MF_ONLYEAGLE) && c->monst != moEagle && c->monst != moBat) return;
     if(movegroup(c->monst) == movtype && c->pathdist != 0) {
       cell *c2 = moveNormal(c, mf);
-      if(c2) onpath(c2, 0);
+      if(c2) onpath_mark(c2);
       }
     }
   }
@@ -1322,7 +1324,7 @@ EX void hexvisit(cell *c, cell *from, int d, bool mounted, int colorpair) {
     moveHexSnake(movei(from, d).rev(), mounted);
     }
 
-  onpath(c, 0);
+  onpath_mark(c);
 
   // MAXGCELL
   if(isize(hexdfs) < 2000 || c->cpdist <= 6) 
@@ -1336,12 +1338,12 @@ EX void movehex(bool mounted, int colorpair) {
   if(mounted) { 
     if(dragon::target && dragon::target->monst != moHexSnake) {
       hexdfs.push_back(dragon::target); 
-      onpath(dragon::target, 0);
+      onpath_mark(dragon::target);
       }
     }
   else for(cell *c: targets) {
     hexdfs.push_back(c);
-    onpath(c, 0);
+    onpath_mark(c);
     }
   //hexdfs.push_back(cwt.at);
   
@@ -1528,7 +1530,7 @@ EX void moveghosts() {
       
       vector<int> mdir;
 
-      for(int j=0; j<c->type; j++) 
+      for(int p: {0, 1}) for(int j=0; j<c->type; j++) if(p == 1 || (c->move(j) && isPlayerOn(c->move(j))))
         if(c->move(j) && canAttack(c, c->monst, c->move(j), c->move(j)->monst, AF_GETPLAYER | AF_ONLY_FBUG)) {
           // XLATC ghost/greater shark
           
@@ -1685,8 +1687,7 @@ EX void movegolems(flagtype flags) {
       if(m == moGolem) qg++;
       if(m == moFriendlyGhost) markOrb(itOrbUndeath);
 
-      bool recorduse[ittypes];
-      for(int i=0; i<ittypes; i++) recorduse[i] = orbused[i];
+      auto recorduse = orbused;
 
       DEBB(DF_TURN, ("stayval"));
       int bestv = stayvalue(m, c);
@@ -1710,7 +1711,7 @@ EX void movegolems(flagtype flags) {
           }
         }
 
-      for(int i=0; i<ittypes; i++) orbused[i] = recorduse[i];
+      orbused = recorduse;
       
 //    printf("stayvalue = %d, result = %d, bq = %d\n", stayvalue(m,c), bestv, bq);
         
