@@ -393,8 +393,8 @@ struct shmup_configurer {
 
   void operator()() {
   #if CAP_SDL
-    cmode = sm::SHMUPCONFIG;
-    gamescreen(3);
+    cmode = sm::SHMUPCONFIG | sm::SIDE | sm::DARKEN;
+    gamescreen();
     dialog::init(XLAT("keyboard & joysticks"));
     
     bool haveconfig = shmup::on || players > 1 || multi::alwaysuse;
@@ -503,7 +503,8 @@ EX void configure() {
   }
 
 EX void showConfigureMultiplayer() {
-  gamescreen(1);
+  cmode = sm::SIDE | sm::MAYDARK;
+  gamescreen();
   dialog::init("multiplayer");
   
   for(int i=1; i <= MAXPLAYER; i++) {
@@ -752,6 +753,7 @@ EX void initConfig() {
   }
 
 EX void get_actions() {
+  #if !ISMOBILE
   const Uint8 *keystate = SDL12_GetKeyState(NULL);
 
   for(int i=0; i<NUMACT; i++) 
@@ -786,6 +788,7 @@ EX void get_actions() {
       axespressed[scfg.axeaction[j][b] % SHMUPAXES] += value;
       }
     }
+#endif
 #endif
   }
 
@@ -826,7 +829,7 @@ EX void handleInput(int delta) {
 #endif
   
   if(actionspressed[58] && !lactionpressed[58]) 
-    pushScreen(showMainMenu);
+    pushScreen(showGameMenu);
     
   panx *= d;
   pany *= d;
