@@ -18,8 +18,9 @@ constexpr flagtype GF_VARCOLOR = 2;
 constexpr flagtype GF_LIGHTFOG = 4;
 constexpr flagtype GF_LEVELS   = 8;
 constexpr flagtype GF_TEXTURE_SHADED  = 16;
+constexpr flagtype GF_NO_FOG   = 32;
 
-constexpr flagtype GF_which    = 31;
+constexpr flagtype GF_which    = 63;
 
 constexpr flagtype SF_PERS3        = 256;
 constexpr flagtype SF_BAND         = 512;
@@ -31,6 +32,8 @@ constexpr flagtype SF_ORIENT       = 16384;
 constexpr flagtype SF_BOX          = 32768;
 constexpr flagtype SF_ZFOG         = 65536;
 constexpr flagtype SF_ODSBOX       = (1<<17);
+
+constexpr flagtype SF_SEMIDIRECT   = (1<<18);
 #endif
 
 EX bool solv_all;
@@ -83,7 +86,7 @@ EX string stereo_shader() {
     "float s = t.z;\n"
     "float l = length(t.xyz);\n"
     "t /= max(l, 1e-2);\n"
-    "t.z += " + glhr::to_glsl(panini_alpha) + ";\n"
+    "t.z += " + glhr::to_glsl(stereo_alpha) + ";\n"
     "t *= l;\n"
     "t.w = 1.;\n";
   }
@@ -104,11 +107,57 @@ EX string shader_lie_log() {
   else if(hyperbolic) {
     return "vec4 lie_log(vec4 v) { v = deparabolic13(v); v[3] = 1.; /* if(abs(v[0]) > 1e-6) { float m = v[0] / (exp(v[0]) - 1.); v[1] *= m; v[2] *= m; } */ return v; }\n";
     }
+  else if(sl2) {
+    return shader_rel_log() + "vec4 lie_log(vec4 h) { return rel_log(h); }\n";
+    }
   else {
     return "vec4 lie_log(vec4 v) { return v; }\n";
     }
   }
 
+EX string shader_rel_log() {
+  if(sl2) return
+  "uniform mediump float uIndexSL;\n"
+  "vec4 rel_log(vec4 h) {\n"
+    "float shift = uIndexSL + atan2(h[2], h[3]); \n"
+    "float ca = cos(uIndexSL); float sa = -sin(uIndexSL);\n"
+    "vec4 h1 = h;\n"
+    "h[2] = h1[2] * ca - h1[3] * sa; h[3] = h1[3] * ca + h1[2] * sa;\n"
+    "h[0] = h1[0] * ca - h1[1] * sa; h[1] = h1[1] * ca + h1[0] * sa;\n"
+    "h1 = h;"
+
+    "if(h1[3] <= 1. && h1[3] >= -1.) {\n"
+    "float r = sqrt(h1[2]*h1[2] - h1[0]*h1[0] - h1[1]*h1[1]);\n"
+    "float z = asin_clamp(r);\n"
+    "if(h1[3] < 0.) z = PI - z;\n"
+    "z += floor(shift / 2. / PI + .5) * 2. * PI;\n"
+    "float scale = z/r;\n"
+    "h1 = h1 * scale; h1[3] = 1.;\n"
+    "} else if(shift > PI || shift < -PI || h1[3] < -1.) { return vec4(0,0,0,1); } else {\n"
+
+    "float r = sqrt(h1[0]*h1[0] + h1[1]*h1[1] - h1[2]*h1[2]);\n"
+    "float z = asinh(r);\n"
+    "float scale = z/r;\n"
+    "h1 = h1 * scale; h1[3] = 1.;\n"
+    "}\n"
+
+    "return h1;\n"
+    "}\n";
+
+  if(hyperbolic && GDIM == 3) return
+    "vec4 rel_log(vec4 h) {\n"
+    "  float choice = h[3] * h[3] - h[0] * h[0] - h[1] * h[1];\n"
+    "  float z, r;\n"
+    "  if(choice > 0.) { r = sqrt(choice); z = asinh(r); }\n"
+    "  else { r = sqrt(-choice); z = asin_clamp(r); if(h[2] < 0.) z = PI - z; }\n"
+    "  h = h * z / r; h[2] = h[3]; h[3] = 1.;\n"
+    "  return h;\n"
+    "  }\n";
+
+  println(hlog, "geometry is: ", geometry);
+  throw hr_exception("shader_rel_log in wrong geometry");
+  }
+
 shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
   string varying, vsh, fsh, vmain = "void main() {\n", fmain = "void main() {\n";
 
@@ -189,6 +238,11 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
     skip_t = true;
     shader_flags |= SF_DIRECT;
     }
+  else if(!vid.consider_shader_projection && semidirect_rendering && models::is_perspective(pmodel)) {
+    vmain += "// this\n";
+    distfun = "length(t.xyz)";
+    shader_flags |= SF_PERS3 | SF_SEMIDIRECT;
+    }
   else if(!vid.consider_shader_projection) {
     shader_flags |= SF_PIXELS;
     }        
@@ -221,18 +275,23 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
       "t.z = (sin(uCamera) * cos(zl) * cos(d) - sin(zl) * cos(uCamera)) / uz;\n"
       ;
     }
-  else if(pmodel == mdDisk && MDIM == 3 && !spherespecial && !prod) {
+  else if(pmodel == mdDisk && MDIM == 3 && !spherespecial && !gproduct) {
     shader_flags |= SF_DIRECT;
     }
   else if(glhr::noshaders) {
     shader_flags |= SF_PIXELS;
     }
-  else if(pmodel == mdDisk && GDIM == 3 && !spherespecial && !nonisotropic && !prod) {
+  else if(pmodel == mdDisk && GDIM == 3 && !spherespecial && !nonisotropic && !gproduct) {
     coordinator += "t /= (t[3] + uAlpha);\n";
-    vsh += "uniform mediump float uAlpha;";
+    vsh += "uniform mediump float uAlpha;\n";
     shader_flags |= SF_DIRECT | SF_BOX | SF_ZFOG;
     treset = true;
     }
+  else if(pmodel == mdConformalSquare && pconf.model_transition == 1) {
+    shader_flags |= SF_ORIENT | SF_DIRECT;
+    coordinator += "t = uPP * t;", vsh += "uniform mediump mat4 uPP;";
+    coordinator += "t = to_square(t);";
+    }
   else if(pmodel == mdBand && hyperbolic) {
     shader_flags |= SF_BAND | SF_ORIENT | SF_BOX | SF_DIRECT;
     coordinator += "t = uPP * t;", vsh += "uniform mediump mat4 uPP;";
@@ -269,6 +328,12 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
     distfun = "length(t.xyz)";
     vsh += shader_lie_log();
     }
+  else if(pmodel == mdRelPerspective) {
+    shader_flags |= SF_PERS3 | SF_DIRECT;
+    coordinator += "t = rel_log(t);\n";
+    distfun = "length(t.xyz)";
+    vsh += shader_rel_log();
+    }
   else if(pmodel == mdGeodesic) {
     shader_flags |= SF_PERS3 | SF_DIRECT;
     coordinator += "t = inverse_exp(t);\n";
@@ -291,28 +356,21 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
       distfun = "length(t.xyz)";
     switch(cgclass) {
       #if CAP_SOLV
-      case gcSolNIH:
-        switch(sn::geom()) {
-          case gSol:
-            if(solv_all) {
-              vsh += "\n#define SOLV_ALL\n";
-              }
-            vsh += sn::shader_symsol;
-            break;
-          case gNIH:
-            vsh += sn::shader_nsym;
-            break;
-          case gSolN:
-            vsh += sn::shader_nsymsol;
-            break;
-          default:
-            println(hlog, "error: unknown sn geometry");
-          }            
-        treset = true;
+      case gcSol:
+        if(solv_all) {
+          vsh += "\n#define SOLV_ALL\n";
+          }
+        vsh += sn::shader_symsol;
+        break;
+      case gcNIH:
+        vsh += sn::shader_nsym;
+        break;
+      case gcSolN:
+        vsh += sn::shader_nsymsol;
         break;
       #endif
       case gcNil:
-        vsh += nilv::nilshader;
+        vsh += nilv::nilshader();
         break;
       case gcSL2:      
         vsh += slr::slshader;
@@ -348,6 +406,14 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
     }
   else if(pmodel == mdPerspective) {
     shader_flags |= SF_PERS3 | SF_DIRECT;
+    if(sl2) {
+      vsh += "uniform mediump float uIndexSL;\n";
+      coordinator +=
+        "float ca = cos(uIndexSL); float sa = -sin(uIndexSL);\n"
+        "vec4 h1 = t;\n"
+        "t[2] = h1[2] * ca - h1[3] * sa; t[3] = 1.;\n"
+        "t[0] = h1[0] * ca - h1[1] * sa; t[1] = h1[1] * ca + h1[0] * sa;\n";
+      }
     #if CAP_VR
     if(vrhr::rendering() && hyperbolic && vrhr::eyes != vrhr::eEyes::truesim) {
       azi_hyperbolic = true;
@@ -387,13 +453,16 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
   if(nil && pmodel == mdPerspective)  {
     vsh += "uniform mediump float uRotCos, uRotSin, uRotNil;\n";
     coordinator +=
-      "t.z += (uRotCos * t.x + uRotSin * t.y) * (uRotCos * t.y - uRotSin * t.x) * uRotNil / 2. - t.x * t.y / 2.;\n";
+      "t.z += (uRotCos * t.x + uRotSin * t.y) * (uRotCos * t.y - uRotSin * t.x) * uRotNil / 2. - " + glhr::to_glsl(nilv::model_used) + " * t.x * t.y / 2.;\n";
     }
 
   if(!skip_t) {
     vmain += "mediump vec4 t = uMV * aPosition;\n";
     vmain += coordinator;
-    if(GDIM == 3 && WDIM == 2 && hyperbolic && context_fog && pmodel == mdPerspective) {
+    if(shader_flags & GF_NO_FOG) {
+      vmain += "// no fog used\n";
+      }
+    else if(GDIM == 3 && WDIM == 2 && hyperbolic && context_fog && cgi.emb->is_same_in_same() && pmodel == mdPerspective) {
       vsh += 
         "uniform mediump mat4 uRadarTransform;\n"
         "uniform mediump sampler2D tAirMap;\n"
@@ -451,7 +520,6 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
       shader_flags |= SF_ORIENT;
       }
     else if((shader_flags & SF_PERS3) && stereo_alpha && !vrhr::rendering_eye()) {
-      vmain += "t = uPP * t;", vsh += "uniform mediump mat4 uPP;";
       vmain += stereo_shader();
       }
       
@@ -509,7 +577,12 @@ shared_ptr<glhr::GLprogram> write_shader(flagtype shader_flags) {
 void display_data::set_projection(int ed, ld shift) {
   flagtype shader_flags = current_display->next_shader_flags;
   unsigned id;
-  id = geometry;
+  id = cgclass;
+  if(stretch::in()) id = 15;
+  id <<= 1; if(GDIM == 3) id |= 1;
+  id <<= 1; if(embedded_plane) id |= 1;
+  if(GDIM == 2 && hyperbolic && pconf.alpha < 0 && pconf.alpha > -1) id |= 1;
+  id <<= 3; id |= vid.stereo_mode;
   id <<= 6; id |= pmodel;
   if(levellines && pmodel != mdPixel) {
     shader_flags |= GF_LEVELS;
@@ -526,7 +599,7 @@ void display_data::set_projection(int ed, ld shift) {
   if(sol && solv_all) id |= 1;
   if(in_h2xe()) id |= 1;
   if(in_s2xe()) id |= 2;
-  if(WDIM == 2 && GDIM == 3 && hyperbolic && context_fog) id |= 1;
+  if(WDIM == 2 && GDIM == 3 && hyperbolic && context_fog && cgi.emb->is_same_in_same()) id |= 1;
   shared_ptr<glhr::GLprogram> selected;
 
   if(matched_programs.count(id)) selected = matched_programs[id];
@@ -570,7 +643,7 @@ void display_data::set_projection(int ed, ld shift) {
   if(selected->uIterations != -1) {
     glhr::set_index_sl(0);
     glhr::set_sv(stretch::not_squared());
-    glhr::set_sl_iterations(slr::steps);
+    glhr::set_sl_iterations(slr::shader_iterations);
     }
 
   glhr::new_projection();
@@ -597,17 +670,19 @@ void display_data::set_projection(int ed, ld shift) {
       eyewidth_translate(ed);
     }
   
+  auto bcolor = models::is_perspective(pmodel) ? backcolor : (modelcolor >> 8);
+
   auto ortho = [&] (ld x, ld y) {
     glhr::glmatrix M = glhr::ortho(x, y, 1);
     if(shader_flags & SF_ZFOG) {
       M[2][2] = 2 / (pconf.clip_max - pconf.clip_min);
       M[3][2] = (pconf.clip_min + pconf.clip_max) / (pconf.clip_max - pconf.clip_min);
-      auto cols = glhr::acolor(darkena(backcolor, 0, 0xFF));
+      auto cols = glhr::acolor(darkena(bcolor, 0, 0xFF));
       glUniform4f(selected->uFogColor, cols[0], cols[1], cols[2], cols[3]);
       }
     else M[2][2] /= 10000;
     glhr::projection_multiply(M);
-    if(nisot::local_perspective_used() && (shader_flags & SF_BOX))
+    if(nisot::local_perspective_used && (shader_flags & SF_BOX))
       glhr::projection_multiply(glhr::tmtogl_transpose(NLP));
     if(ed && vid.stereo_mode != sODS) {
       glhr::glmatrix m = glhr::id;
@@ -652,15 +727,15 @@ void display_data::set_projection(int ed, ld shift) {
     }
   else if(shader_flags & SF_ODSBOX) {
     ortho(M_PI, M_PI);
-    glhr::fog_max(1/sightranges[geometry], darkena(backcolor, 0, 0xFF));
+    glhr::fog_max(1/sightranges[geometry], darkena(bcolor, 0, 0xFF));
     }
   else if(shader_flags & SF_PERS3) {
     if(vrhr::rendering()) use_mv();
     else {
       glhr::projection_multiply(glhr::frustum(cd->tanfov, cd->tanfov * cd->ysize / cd->xsize));
       glhr::projection_multiply(glhr::scale(1, -1, -1));
-      if(nisot::local_perspective_used()) {
-        if(prod) {
+      if(nisot::local_perspective_used) {
+        if(gproduct) {
           for(int i=0; i<3; i++) NLP[3][i] = NLP[i][3] = 0;
           NLP[3][3] = 1;
           }
@@ -672,7 +747,7 @@ void display_data::set_projection(int ed, ld shift) {
         glhr::eyeshift = glhr::tmtogl(xpush(vid.ipd * ed/2));
         }
       }
-    glhr::fog_max(1/sightranges[geometry], darkena(backcolor, 0, 0xFF));
+    glhr::fog_max(1/sightranges[geometry], darkena(bcolor, 0, 0xFF));
     }
   else {
     if(pconf.alpha > -1) {
@@ -705,7 +780,7 @@ void display_data::set_projection(int ed, ld shift) {
 
     pp = pp * pp0;    
 
-    if(nisot::local_perspective_used()) 
+    if(nisot::local_perspective_used)
       pp = glhr::tmtogl_transpose(NLP) * pp;
 
     if(get_shader_flags() & SF_ORIENT) {
@@ -738,7 +813,7 @@ void display_data::set_projection(int ed, ld shift) {
     glhr::projection_multiply(model_orientation_gl());
 
   if(selected->shader_flags & SF_BAND)
-    glhr::projection_multiply(glhr::scale(2 / M_PI, 2 / M_PI, GDIM == 3 ? 2/M_PI : 1));
+    glhr::projection_multiply(glhr::scale(1 / 90._deg, 1 / 90._deg, GDIM == 3 ? 1/90._deg : 1));
 
   if(selected->shader_flags & SF_BAND) {
     glhr::projection_multiply(glhr::translate(shift, 0, 0));
@@ -785,6 +860,67 @@ EX void add_if(string& shader, const string& seek, const string& function) {
 EX void add_fixed_functions(string& shader) {
   /* from the most complex to the simplest */
 
+  add_if(shader, "to_square",
+    "mediump vec4 to_square(mediump vec4 h) {\n"
+    "float d = length(h.xy);\n"
+    "float x = d / (h.z + 1.);\n"
+
+    "float cos_phiosqrt2 = sqrt(2.) / (x + 1./x);\n"
+    "float cos_lambda = -h.y / d;\n"
+    "float sin_lambda = h.x / d;\n"
+    "float cos_a = cos_phiosqrt2 * (sin_lambda + cos_lambda);\n"
+    "float cos_b = cos_phiosqrt2 * (sin_lambda - cos_lambda);\n"
+    "float sin_a = sqrt(1. - cos_a * cos_a);\n"
+    "float sin_b = sqrt(1. - cos_b * cos_b);\n"
+    "float cos_a_cos_b = cos_a * cos_b;\n"
+    "float sin_a_sin_b = sin_a * sin_b;\n"
+    "float sin2_m = 1.0 + cos_a_cos_b - sin_a_sin_b;\n"
+    "float sin2_n = 1.0 - cos_a_cos_b - sin_a_sin_b;\n"
+    "float sin_m = sqrt_clamp(sin2_m);\n"
+    "float cos_m = sqrt_clamp(1. - sin2_m);\n"
+    "if(sin_lambda < 0.) sin_m = -sin_m;\n"
+    "float sin_n = sqrt_clamp(sin2_n);\n"
+    "float cos_n = sqrt_clamp(1.0 - sin2_n);\n"
+    "if(cos_lambda > 0.0) sin_n = -sin_n;\n"
+    "#define divby 0.53935260118837935472\n"
+    "vec4 res = vec4(ellFaux(cos_m,sin_m,sqrt(2.)/2.) * divby, ellFaux(cos_n,sin_n,sqrt(2.)/2.) * divby, 0, 1);\n"
+    "if(x > 1.) {\n"
+    "  if(abs(res[0]) > abs(res[1])) {\n"
+    "    if(res[0] > 0.) res[0] = 2. - res[0]; else res[0] = -2. - res[0];\n"
+    "    }\n"
+    "  else {\n"
+    "    if(res[1] > 0.) res[1] = 2. - res[1]; else res[1] = -2. - res[1];\n"
+    "    }\n"
+    "  }\n"
+    "return res;\n"
+    "}\n");
+
+  add_if(shader, "sqrt_clamp", "mediump float sqrt_clamp(mediump float x) { return x >= 0. ? sqrt(x) : 0.; }\n");
+  add_if(shader, "ellFaux", "mediump float ellFaux(mediump float cos_phi, mediump float sin_phi, mediump float k) {\n"
+    "return sin_phi * ellRF(cos_phi * cos_phi, 1. - k * k * sin_phi * sin_phi, 1.);\n"
+     "}\n");
+  add_if(shader, "ellRF", "mediump float ellRF(mediump float x, mediump float y, mediump float z) {\n"
+    "float delx = 1., dely = 1., delz = 1.;\n"
+    "const float eps = 0.0025;\n"
+    "float mean;\n"
+    "while(abs(delx) > eps || abs(dely) > eps || abs(delz) > eps) {\n"
+    "  float sx = sqrt(x);\n"
+    "  float sy = sqrt(y);\n"
+    "  float sz = sqrt(z);\n"
+    "  float len = sx * (sy+sz) + sy * sz;\n"
+    "  float x = .25 * (x+len);\n"
+    "  float y = .25 * (y+len);\n"
+    "  float z = .25 * (z+len);\n"
+    "  mean = (x+y+z)/3.;\n"
+    "  delx = (mean-x) / mean;\n"
+    "  dely = (mean-y) / mean;\n"
+    "  delz = (mean-z) / mean;\n"
+    "  }\n"
+    "float e2 = delx * dely - delz * delz;\n"
+    "float e3 = delx * dely * delz;\n"
+    "return ((1.0 + (e2 / 24.0 - 0.1 - 3.0 * e3 / 44.0) * e2+ e3 / 14.) / sqrt(mean));\n"
+    "}\n");
+
   add_if(shader, "tanh", "mediump float tanh(mediump float x) { return sinh(x) / cosh(x); }\n");
   add_if(shader, "sinh", "mediump float sinh(mediump float x) { return (exp(x) - exp(-x)) / 2.0; }\n");
   add_if(shader, "asin_clamp", "mediump float asin_clamp(mediump float x) { return x > 1. ? PI/2. : x < -1. ? -PI/2. : asin(x); }\n");
