@@ -5,11 +5,27 @@ namespace hr {
 EX bool context_fog = true;
 
 EX ld camera_level;
-EX bool draw_sky = true;
+EX bool camera_sign;
+
+#if HDR
+enum eSkyMode { skyNone, skyAutomatic, skySkybox, skyAlways };
+#endif
+
+EX eSkyMode draw_sky;
+
+EX bool auto_remove_roofs;
+EX bool simple_sky;
+
+EX bool camera_over(ld x) {
+  if(!auto_remove_roofs) return false;
+  if(camera_sign) return camera_level <= x;
+  return camera_level >= x;
+  }
 
 #if MAXMDIM >= 4 && CAP_GL
 
 EX int get_skybrightness(int mul IS(1)) {
+  if(simple_sky) return 255;
   ld s = 1 - mul * (camera_level - cgi.WALL) / -2;
   if(s > 1) return 255;
   if(s < 0) return 0;
@@ -35,28 +51,66 @@ struct dqi_sky : drawqueueitem {
   
 EX struct dqi_sky *sky;
 
+EX bool do_draw_skybox() {
+  if(no_wall_rendering) return false;
+  if(!euclid) return false;
+  if(draw_sky == skySkybox) return true;
+  if(!embedded_plane) return false;
+  if(draw_sky == skyAutomatic) return !cgi.emb->is_sph_in_low() && !cgi.emb->is_cylinder();
+  return false;
+  }
+
 EX void prepare_sky() {
   sky = NULL;
-  if(euclid) {
-    if(WDIM == 3 || GDIM == 2) return;
-    if(no_wall_rendering) return;
+  if(do_draw_skybox()) {
     shiftmatrix T = ggmatrix(currentmap->gamestart());
     T.T = gpushxto0(tC0(T.T)) * T.T;
     queuepoly(T, cgi.shEuclideanSky, 0x0044e4FF);
-    queuepolyat(T * zpush(cgi.SKY+0.5) * xpush(cgi.SKY+0.5), cgi.shSun, 0xFFFF00FF, PPR::SKY);
+    queuepolyat(T * zpush(cgi.STAR * 0.7) * xpush(cgi.STAR * 0.7), cgi.shSkyboxSun, 0xFFFF00FF, PPR::SKY);
     }
   else {
-    sky = &queuea<dqi_sky> (PPR::MISSILE);
+    sky = &queuea<dqi_sky> (euclid ? PPR::EUCLIDEAN_SKY : PPR::MISSILE);
     }
   }
 
-vector<glhr::colored_vertex> skyvertices;
-cell *sky_centerover;
-shiftmatrix sky_cview;
+EX vector<glhr::colored_vertex> skyvertices;
+EX cell *sky_centerover;
+EX shiftmatrix sky_cview;
+
+EX void delete_sky() {
+  sky_centerover = nullptr;
+  skyvertices.clear();
+  }
+
+EX bool do_draw_sky() {
+  if(!embedded_plane) return false;
+  if(draw_sky == skyAlways) return true;
+  if(draw_sky != skyAutomatic) return false;
+
+  if(vid.wall_height < 0 && cgi.emb->is_euc_in_hyp()) return false; /* just looks bad, hollow horospheres should not have sky */
+  if(vid.wall_height < 0 && meuclid && geom3::ggclass() == gcNIH) return false; /* same */
+  if(among(geom3::ggclass(), gcSol, gcSolN)) return false;  /* errors */
+  if(among(geom3::ggclass(), gcNil)) return false;  /* errors sometimes too */
+  if(cgi.emb->is_hyp_in_solnih()) return false;
+  if(cgi.emb->is_euc_in_product()) return false;
+  if(cgi.emb->is_euc_in_sl2()) return false;
+  if(cgi.emb->is_cylinder()) return false;
+
+  return true;
+  }
+
+EX bool do_draw_stars(bool rev) {
+  if(!do_draw_sky()) return false;
+  if(cgi.emb->is_sph_in_low()) {
+    if(cgi.SKY < 0) return false;
+    }
+  if(cgi.emb->is_euc_in_hyp() && (rev ? cgi.SKY > 0 : cgi.SKY < 0)) return false;
+  return true;
+  }
 
 void compute_skyvertices(const vector<sky_item>& sky) {
   skyvertices.clear();
-  if(!draw_sky) return;
+  if(!do_draw_sky()) return;
 
   int sk = get_skybrightness();
   
@@ -66,11 +120,11 @@ void compute_skyvertices(const vector<sky_item>& sky) {
         darkena(si.skycolor, 0, 0xFF)
         );
   
-  hyperpoint skypoint = cpush0(2, cgi.SKY);
-  hyperpoint hellpoint = cpush0(2, -cgi.SKY);
+  hyperpoint skypoint = cpush0(2, cgi.SKY + cgi.emb->center_z());
+  hyperpoint hellpoint = cpush0(2, cgi.HELL + cgi.emb->center_z());
   
   vector<glhr::colored_vertex> this_poly;
-  
+
   for(const sky_item& si: sky) {
     auto c = si.c;
     
@@ -184,6 +238,7 @@ void compute_skyvertices(const vector<sky_item>& sky) {
         }
 
       if(true) {
+        hyperpoint tctr = tile_center();
         cellwalker cw0(c, i);
         cellwalker cw = cw0;
         do {
@@ -197,7 +252,7 @@ void compute_skyvertices(const vector<sky_item>& sky) {
   
         transmatrix T1 = Id;
         do {
-          vertices.push_back(tC0(T1));
+          vertices.push_back(T1 * tctr);
           vcolors.push_back(colors[cw.at].first);
           T1 = T1 * currentmap->adj(cw.at, cw.spin);
           cw += wstep; cw++;
@@ -210,16 +265,17 @@ void compute_skyvertices(const vector<sky_item>& sky) {
         for(int i=0; i<k; i++) ccolor = gradient(ccolor, vcolors[i], 0, 1, i+1);
         
         hyperpoint ctr = Hypc;
+        for(auto& p: vertices) p = cgi.emb->normalize_flat(p);
         for(auto& p: vertices) ctr = ctr + p;
-        normalize(ctr);
-  
+        ctr = cgi.emb->normalize_flat(ctr);
+
         for(int j=0; j<k; j++) {
           int j1 = (j+1) % k;
           glhr::colored_vertex cv[prec+1][prec+1];
           for(int x=0; x<=prec; x++) for(int y=0; y<=prec; y++) if(x+y <= prec) {
             hyperpoint h = ctr * (prec-x-y) + vertices[j] * x + vertices[j1] * y;
-            h = normalize(h);
-            color_t co = gradient(ccolor, gradient(vcolors[j], vcolors[j1], 0, y, x+y), prec, x+y, 0);
+            h = cgi.emb->normalize_flat(h);
+            color_t co = gradient(ccolor, gradient(vcolors[j], vcolors[j1], 0, y, x+y), 0, x+y, prec);
             // co = (hrand(0x1000000)  << 8) | 0xFF;
             // co = minecolors[(x+2*y) % 7] << 8 | 0xFF;
             h = unshift(si.T) * orthogonal_move(h, cgi.SKY);
@@ -244,18 +300,21 @@ void compute_skyvertices(const vector<sky_item>& sky) {
       }
     }
 
-  for(auto& v: skyvertices) for(int i=0; i<3; i++) v.color[i] *= 2;
+  if(!simple_sky) for(auto& v: skyvertices) for(int i=0; i<3; i++) v.color[i] *= 2;
   }
 
 void dqi_sky::draw() {
   if(!vid.usingGL || sky.empty() || skyvertices.empty()) return;
-  
+  if(!do_draw_sky()) { delete_sky(); return; }
+
   #if CAP_VR
   transmatrix s = (vrhr::rendering() ? vrhr::master_cview : cview()).T * inverse(sky_cview.T);
   #else
   transmatrix s = cview().T * inverse(sky_cview.T);
   #endif
     
+  be_euclidean_infinity(s);
+
   for(int ed = current_display->stereo_active() ? -1 : 0; ed<2; ed+=2) {
     if(global_projection && global_projection != ed) continue;
     current_display->next_shader_flags = GF_VARCOLOR;
@@ -268,9 +327,18 @@ void dqi_sky::draw() {
       glapplymatrix(s);
       }
     glhr::prepare(skyvertices);
-    glhr::set_fogbase(1.0 + 5 / sightranges[geometry]);
-    glhr::set_depthtest(model_needs_depth() && prio < PPR::SUPERLINE);
-    glhr::set_depthwrite(model_needs_depth() && prio != PPR::TRANSPARENT_SHADOW && prio != PPR::EUCLIDEAN_SKY);
+    glhr::color2(0xFFFFFFFF);
+    if(simple_sky) {
+      glhr::set_fogbase(1);
+      glhr::set_depthtest(true);
+      glhr::set_depthwrite(true);
+      }
+    else {
+      glhr::set_fogbase(1.0 + abs(cgi.SKY - cgi.LOWSKY) / sightranges[geometry]);
+      glhr::set_depthtest(model_needs_depth() && prio < PPR::SUPERLINE);
+      glhr::set_depthwrite(model_needs_depth() && prio != PPR::TRANSPARENT_SHADOW && prio != PPR::EUCLIDEAN_SKY);
+      }
+
     glDrawArrays(GL_TRIANGLES, 0, isize(skyvertices));
     }
   }
@@ -282,28 +350,61 @@ color_t skycolor(cell *c) {
   return gradient(0x4040FF, 0xFFFFFF, 0, z, 63);
   }
 
-EX const ld star_val = 2;
+EX bool use_euclidean_infinity = true;
+
+/** move an Euclidean matrix to V(C0) == C0 */
+EX void be_euclidean_infinity(transmatrix& V) {
+  if(euclid && msphere && use_euclidean_infinity)
+    for(int i=0; i<3; i++) V[i][3] = 0;
+  }
+
+EX void draw_star(const shiftmatrix& V, const hpcshape& sh, color_t col, ld rev IS(false)) {
+  if(!do_draw_stars(rev)) return;
+
+  ld val = cgi.STAR;
+  if(rev) val = cgi.FLOOR * 2 - val;
+  val += cgi.emb->center_z();
+
+  auto V1 = V; be_euclidean_infinity(V1.T);
+  queuepolyat(V1 * lzpush(val), sh, col, PPR::SKY);
+  }
+
+EX ld star_prob = 0.33;
+
+/* the first star is supposed to appear as long as probability > 0 */
+EX vector<ld> stars = {1e-20};
+
+EX bool star_for(int i) {
+  i = i & ((1<<16)-1);
+  while(isize(stars) <= i) stars.push_back(randd());
+  return stars[i] < star_prob;
+  }
+
+EX hookset<bool(celldrawer *cd)> hooks_ceiling;
+
+EX void g_add_to_sky(cell *c, shiftmatrix& V, color_t col, color_t col2) {
+  if(sky) sky->sky.emplace_back(c, V, col, col2);
+  };
 
 void celldrawer::draw_ceiling() {
 
-  if(pmodel != mdPerspective || sphere) return;
+  if(!models::is_perspective(pmodel)) return;
   
+  if(callhandlers(false, hooks_ceiling, this)) return;
+
   auto add_to_sky = [this] (color_t col, color_t col2) {
     if(sky) sky->sky.emplace_back(c, V, col, col2);
     };
-  
+
   switch(ceiling_category(c)) {
     /* ceilingless levels */
     case 1: {
-      if(euclid) return;
-      if(fieldpattern::fieldval_uniq(c) % 3 == 0) {
-        queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shNightStar, 0xFFFFFFFF, PPR::SKY);
-        }
+      if(star_for(fieldpattern::fieldval_uniq(c)))
+        draw_star(V, cgi.shNightStar, 0xFFFFFFFF);
       add_to_sky(0x00000F, 0x00000F);
       if(c->land == laAsteroids) {
-        if(fieldpattern::fieldval_uniq(c) % 9 < 3) {
-          queuepolyat(V * zpush(-star_val-cgi.SKY), cgi.shNightStar, 0xFFFFFFFF, PPR::SKY);
-          }
+        if(star_for(fieldpattern::fieldval_uniq(c) ^ 0x5555))
+          draw_star(V, cgi.shNightStar, 0xFFFFFFFF, true);
         int sk = get_skybrightness(-1);
         auto sky = draw_shapevec(c, V * MirrorZ, cgi.shFullFloor.levels[SIDE_SKY], 0x000000FF + 0x100 * (sk/17), PPR::SKY);
         if(sky) sky->tinf = NULL, sky->flags |= POLY_INTENSE;
@@ -312,7 +413,6 @@ void celldrawer::draw_ceiling() {
       }
     
     case 2: {
-      if(euclid) return;
       color_t col;
       color_t skycol;
       
@@ -320,25 +420,19 @@ void celldrawer::draw_ceiling() {
         case laWineyard:
           col = 0x4040FF;
           skycol = 0x8080FF;
-          if(emeraldval(c) / 4 == 11) {
-            queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shSun, 0xFFFF00FF, PPR::SKY);
-            }
+          if(emeraldval(c) / 4 == 11) draw_star(V, cgi.shSun, 0xFFFF00FF);
           break;
 
         case laDesert:
           col = 0x2020C0;
           skycol = 0x8080FF;
-          if(emeraldval(c) / 4 == 11) {
-            queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shSun, 0xFFFF00FF, PPR::SKY);
-            }
+          if(emeraldval(c) / 4 == 11) draw_star(V, cgi.shSun, 0xFFFF00FF);
           break;
 
         case laFrog:
           col = 0x4040FF;
           skycol = 0x8080FF;
-          if(zebra40(c) / 4 == 1) {
-            queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shSun, 0xFFFF00FF, PPR::SKY);
-            }
+          if(zebra40(c) / 4 == 1) draw_star(V, cgi.shSun, 0xFFFF00FF);
           break;
 
         case laPower:
@@ -395,7 +489,7 @@ void celldrawer::draw_ceiling() {
     
     case 3: {
       add_to_sky(0, 0);
-      if(camera_level <= cgi.WALL) return;
+      if(camera_over(cgi.WALL)) return;
       if(c->land == laMercuryRiver) fcol = linf[laTerracotta].color, fd = 1;
       if(qfi.fshape) draw_shapevec(c, V, qfi.fshape->levels[SIDE_WALL], darkena(fcol, fd, 0xFF), PPR::WALL);
       forCellIdEx(c2, i, c) 
@@ -410,7 +504,7 @@ void celldrawer::draw_ceiling() {
     
     case 4: {
       add_to_sky(0x00000F, 0x00000F);
-      if(camera_level <= cgi.HIGH2) return;
+      if(camera_over(cgi.HIGH)) return;
       auto ispal = [&] (cell *c0) { return c0->land == laPalace && among(c0->wall, waPalace, waClosedGate, waOpenGate); };
       color_t wcol2 = 0xFFD500;
       if(ispal(c)) {
@@ -426,17 +520,14 @@ void celldrawer::draw_ceiling() {
             placeSidewall(c, i, SIDE_HIGH2, V, darkena(wcol2, fd, 0xFF));
         }
       if(among(c->wall, waClosedGate, waOpenGate) && qfi.fshape) draw_shapevec(c, V, qfi.fshape->levels[SIDE_WALL], 0x202020FF, PPR::WALL);
-      if(euclid) return;
 
-      if(true) {
-        queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shNightStar, 0xFFFFFFFF, PPR::SKY);
-        }
+      draw_star(V, cgi.shNightStar, 0xFFFFFFFF);
       break;
       }
     
     case 6: {
       add_to_sky(skycolor(c), 0x4040C0);
-      if(camera_level <= cgi.HIGH2) return;
+      if(camera_over(cgi.HIGH2)) return;
       color_t wcol2 = winf[waRuinWall].color;
       if(c->landparam == 1)
         forCellIdEx(c2, i, c) if(c2->landparam != 1)
@@ -453,10 +544,9 @@ void celldrawer::draw_ceiling() {
     
     case 7: {
       add_to_sky(0x00000F, 0x00000F);
-      if(fieldpattern::fieldval_uniq(c) % 5 < 2) {
-        queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shNightStar, 0xFFFFFFFF, PPR::SKY);
-        }
-      if(camera_level <= cgi.HIGH2) return;
+      if(fieldpattern::fieldval_uniq(c) % 5 < 2)
+        draw_star(V, cgi.shNightStar, 0xFFFFFFFF);
+      if(camera_over(cgi.HIGH2)) return;
       color_t wcol2 = winf[waColumn].color;
       if(c->landparam == 1)
         forCellIdEx(c2, i, c) if(c2->landparam != 1)
@@ -473,7 +563,7 @@ void celldrawer::draw_ceiling() {
     
     case 5: {
       add_to_sky(0x00000F, 0x00000F);
-      if(camera_level <= cgi.WALL) return;
+      if(camera_over(cgi.WALL)) return;
     
       if(pseudohept(c)) {
         forCellIdEx(c2, i, c) 
@@ -482,18 +572,28 @@ void celldrawer::draw_ceiling() {
       else if(qfi.fshape)
         draw_shapevec(c, V, qfi.fshape->levels[SIDE_WALL], darkena(fcol, fd, 0xFF), PPR::WALL);
 
-      if(euclid) return;
-      if(true) {
-        queuepolyat(V * zpush(cgi.SKY+star_val), cgi.shNightStar, 0xFFFFFFFF, PPR::SKY);
-        }
+      draw_star(V, cgi.shNightStar, 0xFFFFFFFF);
+      break;
       }
+
+    default:
+      add_to_sky(0, 0);
     }
   }
 
 EX struct renderbuffer *airbuf;
 
+EX void swap_if_missing(bool missing) {
+  if(!missing) return;
+  arb::swap_vertices();
+  #if CAP_IRR
+  irr::swap_vertices();
+  #endif
+  }
+
 EX void make_air() {
   if(!sky) return;
+  if(!embedded_plane) return;
 
   if(centerover != sky_centerover) {
     sky_centerover = centerover;
@@ -502,6 +602,7 @@ EX void make_air() {
     }
   
   if(!context_fog) return;
+  if(vrhr::rendering()) return;
 
   const int AIR_TEXTURE = 512;
   if(!airbuf) {
@@ -518,6 +619,10 @@ EX void make_air() {
   dynamicval<int> i(vrhr::state, 0);
   #endif
 
+  bool missing = false;
+
+  auto cgi1 = &cgi;
+
   if(1) {
     //shot::take("airtest.png", drawqueue); 
     dynamicval<videopar> v(vid, vid);    
@@ -545,8 +650,17 @@ EX void make_air() {
     pmodel = mdDisk;
 
     vid.always3 = false;
-    geom3::apply_always3();    
+    geom3::apply_always3();
     check_cgi();
+    missing = !(cgi.state & 2);
+    cgi.require_basics();
+    geom3::swap_direction = -1;
+    if(missing) {
+      swap(cgi.emb, cgi1->emb);
+      swap_if_missing(missing);
+      swap(cgi.emb, cgi1->emb);
+      }
+
     cgi.require_shapes();
     
     eGeometry orig = geometry;
@@ -562,10 +676,12 @@ EX void make_air() {
         S = g.T.T;
         S = current_display->radar_transform * S;
         geometry = orig;
-        swapmatrix(S);
+        S = cgi1->emb->actual_to_base(S);
         }
       
-      auto& h = cgi.shFullFloor.b[shvid(g.c)];
+      int id = shvid(g.c);
+      ensure_floorshape_generated(id, g.c);
+      auto& h = cgi.shFullFloor.b[id];
 
       dqi_poly p;
       p.V = shiftless(S);
@@ -595,6 +711,12 @@ EX void make_air() {
 
   GLERR("after draw");
   geom3::apply_always3();
+  geom3::swap_direction = +1;
+  if(missing) {
+    swap(cgi.emb, cgi1->emb);
+    swap_if_missing(missing);
+    swap(cgi.emb, cgi1->emb);
+    }
   check_cgi();
   calcparam();
   GLERR("after make_air");
