@@ -77,6 +77,8 @@ EX hookset<bool()> hooks_welcome_message;
 /** \brief Print the welcome message during the start of game. Depends on the current mode and other settings. */
 EX void welcomeMessage() {
   if(callhandlers(false, hooks_welcome_message)) return;
+  if(nohelp == 1) return;
+  if(embedded_plane) return IPF(welcomeMessage());
 #if CAP_TOUR
   else if(tour::on) return; // displayed by tour
 #endif
@@ -87,7 +89,9 @@ EX void welcomeMessage() {
       items[itGreenStone] = 99;
     addMessage(XLAT("Welcome to %the1 Challenge!", moPrincess));
     addMessage(XLAT("The more Hypersian Rugs you collect, the harder it is.", moPrincess));
-    }  
+    }
+  else if(intra::in)
+    addMessage(XLAT("Welcome to a portal map! Press 'o' to configure."));
   else if(randomPatternsMode)
     addMessage(XLAT("Welcome to the Random Pattern mode!"));
   else if(tactic::on)
@@ -128,6 +132,8 @@ EX void welcomeMessage() {
     addMessage(XLAT("Welcome to the Heptagonal Mode!"));
   else if(cheater || autocheat)
     addMessage(XLAT("Welcome to HyperRogue! (cheat mode on)"));
+  else if(hat::in())
+    addMessage(XLAT("Welcome to HatRogue!"));
   else
     addMessage(XLAT("Welcome to HyperRogue!"));
 
@@ -137,6 +143,7 @@ EX void welcomeMessage() {
       addMessage(XLAT(lv.msg));
     }
 
+  if(nohelp == 2) return;
 #if ISMAC
   addMessage(XLAT("Press F1 or right-shift-click things for help."));
 #elif !ISMOBILE
@@ -173,10 +180,6 @@ EX void initgame() {
     firstland = safetyland;
     }
 
-  #if CAP_RACING  
-  if(racing::on) racing::apply_seed();
-  #endif
-  
   if(!safety) {
     firstland = specialland;
     ineligible_starting_land = !landUnlocked(specialland);
@@ -213,6 +216,7 @@ EX void initgame() {
   clearing::direct = 0;
   clearing::imputed = 0;
   rosephase = 0;
+  shmup::count_pauses = 0;
 
   if(firstland == laElementalWall) cwt.at->land = randomElementalLand();
   
@@ -279,7 +283,7 @@ EX void initgame() {
 
   if(cwt.at->land == laCrossroads2) {
     cell *c = cwt.at;
-    if(hybri) { c = hybrid::get_where(c).first; PIU( c->cmove(0) ); }
+    if(mhybrid) { c = hybrid::get_where(c).first; PIU( c->cmove(0) ); }
     c->landparam = 12;
     c->cmove(0)->landparam = 44;
     c->cmove(0)->land = laCrossroads2;
@@ -380,7 +384,6 @@ EX void initgame() {
     truelotus = 0;
     asteroids_generated = 0;
     asteroid_orbs_generated = 0;
-    dpgen::in = false;
     survivalist = true;
     #if CAP_CRYSTAL
     crystal::used_compass_inside = false;
@@ -819,11 +822,11 @@ EX void applyBoxes() {
   addinv(itOrbMirror);
   addinv(itGreenStone);
   list_invorb();
-  applyBoxBool(inv::on, "inventory"); // 306
   #if CAP_INV
+  applyBoxBool(inv::on, "inventory"); // 306
   applyBoxNum(inv::rseed, "@inv-rseed");
   #else
-  { int u; applyBoxNum(u); }
+  { int u; applyBoxNum(u); applyBoxNum(u); }
   #endif
 
   // 10.1:
@@ -954,7 +957,9 @@ modecode_t fill_modecode() {
   dynamicval<eVariation> sp5(variation, (eVariation) save.box[186]);
   dynamicval<int> sp7(gp::param.first, save.box[342]);
   dynamicval<int> sp8(gp::param.second, save.box[343]);
+  #if CAP_INV
   dynamicval<bool> spinv(inv::on, save.box[306]);
+  #endif
 
   if(save.box[238]) geometry = gSphere;
   if(save.box[239]) geometry = gElliptic;
@@ -1006,9 +1011,10 @@ EX }
 long long saveposition = -1;
 
 EX void remove_emergency_save() {
+  if(scorefile == "") return;
 #if !ISWINDOWS
   if(saveposition >= 0) { 
-    if(truncate(scorefile, saveposition)) {}
+    if(truncate(scorefile.c_str(), saveposition)) {}
     saveposition = -1;
     }
 #endif
@@ -1020,15 +1026,16 @@ EX void saveStats(bool emergency IS(false)) {
   DEBBI(DF_INIT, ("saveStats [", scorefile, "]"));
 
   if(autocheat) return;
+  if(scorefile == "") return;
   #if CAP_TOUR
   if(tour::on) return;
   #endif
   if(randomPatternsMode) return;
   if(daily::on) return;
   if(peace::on) return;
-  if(dpgen::in) return;
   if(experimental) return;
-  if(!gold()) return;
+
+  if(!gold() && !racing::on) return;
   
   remove_emergency_save();
   
@@ -1036,7 +1043,7 @@ EX void saveStats(bool emergency IS(false)) {
 
   xcode = modecode();
   
-  FILE *f = fopen(scorefile, "at");
+  FILE *f = fopen(scorefile.c_str(), "at");
   
   if(!f) {
     // printf("Could not open the score file '%s'!\n", scorefile);
@@ -1084,6 +1091,20 @@ EX void saveStats(bool emergency IS(false)) {
     return;
     }
 
+  #if CAP_RACING
+  if(racing::on) {
+    auto& bs = racing::best_scores_to_save;
+    if(racing::official_race && !cheater && bs.count(specialland)) {
+      fprintf(f, "RACING %s %d %d date: %s\n", VER,
+        int(specialland), bs[specialland],
+        buf);
+      bs.erase(specialland);
+      }
+    fclose(f);
+    return;
+    }
+  #endif
+
   fprintf(f, "HyperRogue: game statistics (version " VER ")\n");
   if(cheater)
     fprintf(f, "CHEATER! (cheated %d times)\n", cheater);
@@ -1147,12 +1168,13 @@ bool tamper = false;
 // load the save
 EX void loadsave() {
   if(autocheat) return;
+  if(scorefile == "") return;
 #if CAP_TOUR
   if(tour::on) return;
 #endif
   DEBBI(DF_INIT, ("loadSave"));
 
-  FILE *f = fopen(scorefile, "rt");
+  FILE *f = fopen(scorefile.c_str(), "rt");
   havesave = f;
   if(!f) return;
   bool ok = false;
@@ -1250,6 +1272,20 @@ EX void loadsave() {
         }
       }
 
+  #if CAP_RACING
+  if(buf[0] == 'R' && buf[1] == 'A' && buf[2] == 'C') {
+    char buf1[80], ver[10];
+    int land, score;
+    sscanf(buf, "%70s%9s%d%d", buf1, ver, &land, &score);
+    /* score may equal 0 because of earlier bugs */
+    if(score) {
+      auto& res = racing::best_scores[eLand(land)];
+      if(score < res || res == 0) res = score;
+      }
+    println(hlog, "loaded the score for ", dnameof(eLand(land)), " of ", score);
+    }
+  #endif
+
     }
   fclose(f);
   if(ok && sc.box[65 + 4 + itOrbSafety - itOrbLightning]) 
@@ -1275,12 +1311,14 @@ EX void load_last_save() {
 
   loadBox();
 //  printf("boxid = %d\n", boxid);
+  #if CAP_COMPLEX2
   if(items[itHolyGrail]) {
     items[itHolyGrail]--;
     camelot::knighted = newRoundTableRadius();
     items[itHolyGrail]++;
     }
   else camelot::knighted = 0;
+  #endif
   safety = true;
   if(items[itSavedPrincess] < 0) items[itSavedPrincess] = 0;
   addMessage(XLAT("Game loaded."));
@@ -1334,7 +1372,8 @@ EX void stop_game() {
   }
 
 EX eModel default_model() {
-  if(nonisotropic) return mdGeodesic;
+  if(sl2) return mdGeodesic;
+  if(nonisotropic) return nisot::geodesic_movement ? mdGeodesic : mdLiePerspective;
   if(GDIM == 3) return mdPerspective;
   return mdDisk;
   }
@@ -1372,14 +1411,14 @@ EX void set_geometry(eGeometry target) {
         gp::param.second = 0;
       }
     #endif
-    if(DUAL && geometry != gArchimedean && !hybri) 
+    if(DUAL && geometry != gArchimedean && !mhybrid)
       variation = ginf[geometry].default_variation;
     #if CAP_BT
-    if(bt::in() || WDIM == 3 || kite::in() || arb::in()) if(!hybri) variation = eVariation::pure;
+    if(bt::in() || WDIM == 3 || aperiodic || arb::in()) if(!mhybrid) variation = eVariation::pure;
     #endif
     if(S3 >= OINF) variation = eVariation::pure;
-    if(INVERSE && !hybri) variation = gp::variation_for(gp::param);
-    if(ginf[target].default_variation == eVariation::pure && geometry != gArchimedean && !hybri)
+    if(INVERSE && !mhybrid) variation = gp::variation_for(gp::param);
+    if(ginf[target].default_variation == eVariation::pure && geometry != gArchimedean && !mhybrid)
       variation = eVariation::pure;
     geometry_settings(was_default);
     
@@ -1409,7 +1448,7 @@ EX void set_variation(eVariation target) {
       return;
       }
     if(target != eVariation::pure) {
-      if(bt::in() || sol || kite::in() || WDIM == 3) if(!prod) geometry = gNormal;
+      if(bt::in() || sol || aperiodic || WDIM == 3) if(!mproduct) geometry = gNormal;
       }
     auto& cd = ginf[gCrystal];
     if(target == eVariation::bitruncated && cryst && cd.sides == 8 && cd.vertex == 4) {
@@ -1500,6 +1539,7 @@ EX void switch_game_mode(char switchWhat) {
       peace::on = false;
       tour::on = false;
       inv::on = false;
+      yendor::on = false;
       land_structure = racing::on ? lsSingle : lsNiceWalls;
       princess::challenge = false;
       dual::disable();
@@ -1605,6 +1645,13 @@ EX void start_game() {
 // popAllScreens + popAllGames + stop_game + switch_game_mode + start_game
 EX void restart_game(char switchWhat IS(rg::nothing)) {
   popScreenAll();  
+  #if CAP_RACING
+  if(switchWhat == rg::nothing && racing::on) {
+    racing::restore_goals();
+    racing::reset_race();
+    return;
+    }
+  #endif
   stop_game();
   switch_game_mode(switchWhat);
   start_game();
@@ -1629,6 +1676,33 @@ EX eLand firstland0;
 
 EX purehookset hooks_initialize;
 
+EX bool savefile_selection = false;
+
+EX void select_savefile() {
+  if(!savefile_selection) return;
+  start_game();
+  bool canceled = true;
+  if(scorefile == "") scorefile = "hyperrogue.log";
+  pushScreen([] { quitmainloop = true; });
+  dialog::openFileDialog(scorefile, XLAT("choose your score/save file"), ".log", [&] {
+    println(hlog, "hook called");
+    canceled = false;
+    quitmainloop = true;
+    return true;
+    });
+  clearMessages();
+  mainloop();
+  quitmainloop = false;
+  stop_game();
+  popScreenAll();
+  if(canceled) scorefile = "";
+  }
+
+EX void progress_warning() {
+  if(scorefile == "" && savefile_selection)
+    addMessage(XLAT("Your progress will not be saved."));
+  }
+
 EX void initAll() {
   callhooks(hooks_initialize);
   init_floorcolors();
@@ -1646,7 +1720,9 @@ EX void initAll() {
   
   // initlanguage();
   initialize_all();
+  
 #if CAP_SAVE
+  select_savefile();
   loadsave();
   if(IRREGULAR) irr::auto_creator();
 #endif
