@@ -188,6 +188,8 @@ vector<pair<ld, ld>> exp_parser::parse_with_reps() {
   return vals;
   }
 
+ld bmousexs, bmouseys;
+
 cld exp_parser::parse(int prio) {
   cld res;
   skip_white();
@@ -230,7 +232,7 @@ cld exp_parser::parse(int prio) {
     force_eat(",");
     ld b = rparse(0);
     force_eat(")");
-    res = edge_of_triangle_with_angles(2*M_PI/a, M_PI/b, M_PI/b);
+    res = edge_of_triangle_with_angles(TAU/a, M_PI/b, M_PI/b);
     }
   else if(eat("edge_angles(")) {
     cld a = rparse(0);
@@ -253,7 +255,7 @@ cld exp_parser::parse(int prio) {
     force_eat(",");
     ld b = rparse(0);
     force_eat(")");
-    res = edge_of_triangle_with_angles(M_PI/2, M_PI/a, M_PI/b);
+    res = edge_of_triangle_with_angles(90._deg, M_PI/a, M_PI/b);
     }
   #if CAP_ARCM
   else if(eat("arcmedge(")) {
@@ -307,8 +309,7 @@ cld exp_parser::parse(int prio) {
       hyperpoint h = xpush(c) * spin(M_PI - 2*alpha) * xpush0(c);
       ld result = 2 * atan2(h);
       if(result < 0) result = -result;
-      while(result > 2 * M_PI) result -= 2 * M_PI;
-      if(result > M_PI) result = 2 * M_PI - result;
+      cyclefix(result, 0);
       res = result;
       }
 
@@ -389,8 +390,16 @@ cld exp_parser::parse(int prio) {
     else if(number == "psl_steps") res = cgi.psl_steps;
     else if(number == "single_step") res = cgi.single_step;
     else if(number == "step") res = hdist0(tC0(currentmap->adj(cwt.at, 0)));
-    else if(number == "edgelen") res = hdist(get_corner_position(cwt.at, 0), get_corner_position(cwt.at, 1));
+    else if(number == "edgelen") { start_game(); res = hdist(get_corner_position(cwt.at, 0), get_corner_position(cwt.at, 1)); }
     else if(number == "mousey") res = mousey;
+    else if(number == "mousexs") {
+      if(!inHighQual) bmousexs = (1. * mousex - current_display->xcenter) / current_display->radius;
+      res = bmousexs;
+      }
+    else if(number == "mouseys") {
+      if(!inHighQual) bmouseys = (1. * mousey - current_display->ycenter) / current_display->radius;
+      res = bmouseys;
+      }
     else if(number == "random") res = randd();
     else if(number == "mousez") res = cld(mousex - current_display->xcenter, mousey - current_display->ycenter) / cld(current_display->radius, 0);
     else if(number == "shot") res = inHighQual ? 1 : 0;
@@ -737,6 +746,7 @@ EX string compress_string(string s) {
   if(deflate(&strm, Z_FINISH) != Z_STREAM_END) throw hr_exception("z-error-2");
   println(hlog, "deflate ok");
   string out(&buf[0], (char*)(strm.next_out) - &buf[0]);
+  deflateEnd(&strm);
   println(hlog, isize(s), " -> ", isize(out));
   return out;
   }
@@ -753,8 +763,9 @@ EX string decompress_string(string s) {
   vector<char> buf(10000000, 0);
   strm.avail_out = 10000000;
   strm.next_out = (Bytef*) &buf[0];
-  if(inflate(&strm, Z_FINISH) != Z_STREAM_END) throw hr_exception("z-error-2");
+  if(inflate(&strm, Z_FINISH) != Z_STREAM_END) throw hr_exception("z-error-2");  
   string out(&buf[0], (char*)(strm.next_out) - &buf[0]);
+  inflateEnd(&strm);
   println(hlog, isize(s), " -> ", isize(out));
   return out;
   }
@@ -764,6 +775,19 @@ EX bool file_exists(string fname) {
   return access(fname.c_str(), F_OK) != -1;
   }
 
+/** find a file named s, possibly in HYPERPATH */
+EX string find_file(string s) {
+  string s1;
+  if(file_exists(s)) return s;
+  char *p = getenv("HYPERPATH");
+  if(p && file_exists(s1 = s0 + p + s)) return s1;
+  if(file_exists(s1 = HYPERPATH + s)) return s1;
+#ifdef FHS
+  if(file_exists(s1 = "/usr/share/hyperrogue/" + s)) return s1;
+#endif
+  return s;
+  }
+
 EX void open_url(string s) {
   #if ISWEB
   EM_ASM_({
@@ -805,6 +829,21 @@ EX void open_wiki(const char *title) {
   open_url(url);
 }
 
+EX string read_file_as_string(string fname) {
+  string buf;
+  #if ISANDROID || ISIOS
+  buf = get_asset(fname);
+  #else
+  FILE *f = fopen(fname.c_str(), "rb");
+  if(!f) f = fopen((rsrcdir + fname).c_str(), "rb");
+  buf.resize(1000000);
+  int qty = fread(&buf[0], 1, 1000000, f);
+  buf.resize(qty);
+  fclose(f);
+  #endif
+  return buf;
+  }
+
 EX void floyd_warshall(vector<vector<char>>& v) {
   int N = isize(v);
   for(int k=0; k<N; k++)
